Package: dqcheckr
Type: Package
Title: Automated Data Quality Checks for Recurring Dataset Deliveries
Version: 0.2.1
Date: 2026-06-03
Authors@R: person("Mick", "Mioduszewski",
                  email = "mick@mioduszewski.net",
                  role  = c("aut", "cre"))
Description: 
    Automates quality verification of recurring external dataset deliveries.
    For each new file arrival, it runs single-snapshot quality checks, compares
    the file to the previous delivery, writes a self-contained 'HTML' report, and
    records summary statistics in a local 'SQLite' database for long-term trend
    tracking. Supports 'CSV' and fixed-width formats. Custom organisation-specific
    checks can be supplied as plain R files.
License: MIT + file LICENSE
URL: https://github.com/mickmioduszewski/dqcheckr
BugReports: https://github.com/mickmioduszewski/dqcheckr/issues
Encoding: UTF-8
Language: en-GB
Depends: R (>= 4.2)
Imports: readr, DBI, RSQLite, quarto, knitr, kableExtra, ggplot2,
        gridExtra, dplyr, tidyr, yaml, rlang
Suggests: testthat (>= 3.1.0), withr, rmarkdown
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-08 00:13:34 UTC; mick
Author: Mick Mioduszewski [aut, cre]
Maintainer: Mick Mioduszewski <mick@mioduszewski.net>
Repository: CRAN
Date/Publication: 2026-06-08 05:00:02 UTC
Built: R 4.6.0; ; 2026-06-08 08:32:17 UTC; unix

# Generated by roxygen2: do not edit by hand

export(check_allowed_values)
export(check_col_count)
export(check_distinct_counts)
export(check_duplicate_rows)
export(check_empty_column)
export(check_inferred_types)
export(check_key_uniqueness)
export(check_min_row_count)
export(check_missing_rate)
export(check_non_numeric)
export(check_numeric_bounds)
export(check_numeric_stats)
export(check_outliers)
export(check_pattern)
export(check_row_count)
export(check_schema_contract)
export(compare_snapshots)
export(detect_files)
export(dq_result)
export(infer_col_type)
export(list_snapshots)
export(load_config)
export(overall_status)
export(read_dataset)
export(read_recent_snapshots)
export(resolve_col_type)
export(run_comparison_checks)
export(run_custom_checks)
export(run_dq_check)
export(run_qc_checks)
importFrom(dplyr,"%>%")
importFrom(ggplot2,aes)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,theme_minimal)
importFrom(gridExtra,grid.arrange)
importFrom(kableExtra,cell_spec)
importFrom(kableExtra,kable_styling)
importFrom(kableExtra,kbl)
importFrom(knitr,raw_html)
importFrom(stats,IQR)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(tidyr,pivot_longer)
importFrom(utils,head)

# dqcheckr demonstrations

Two runnable demonstrations using the Star Wars dataset from `dplyr`.

## Setup

Install dqcheckr, then copy this folder to a local working directory:

```r
# Install from GitHub
devtools::install_github("mickmioduszewski/dqcheckr")

# Copy demonstrations to ~/dqcheckr_demo
dest <- file.path(path.expand("~"), "dqcheckr_demo")
file.copy(system.file("demonstrations", package = "dqcheckr"),
          dirname(dest), recursive = TRUE)
file.rename(file.path(dirname(dest), "demonstrations"), dest)
```

Open the copied folder in RStudio and set the working directory to it:
**Session → Set Working Directory → To Source File Location**

---

## demo.R — Named-file mode

Checks the Star Wars dataset in both CSV and fixed-width format. Uses explicit
`current_file:` paths in the config — no folder scan.

```
data/
  starwars.csv        current file (CSV)
  starwars.fwf        current file (fixed-width)
config/
  dqcheckr.yml        global rules
  starwars_csv.yml    CSV dataset config
  starwars_fwf.yml    FWF dataset config
output/
  reports/            HTML reports written here
  snapshots.sqlite    run history written here
```

Run: open `demo.R` and source it. Two HTML reports open in the browser.

Expected result: **FAIL** on both — `vehicles` and `starships` columns
are genuinely sparse (87% and 77% missing) so they exceed the missing-rate
threshold. All other checks pass.

---

## demo2.R — Folder-scan mode with version comparison and custom checks

Compares two deliveries of the Star Wars dataset using automatic folder-scan
file detection. Applies custom human-specific checks.

```
data2/
  starwars_v1.csv     original data  (previous delivery)
  starwars_v2.csv     perturbed data (current delivery)
config2/
  dqcheckr.yml        global rules
  starwars_folder.yml folder-scan dataset config
custom2/
  starwars_custom.R   custom checks for human characters
output2/
  reports/            HTML reports written here
  snapshots.sqlite    run history written here
```

**Perturbations in v2** (all within configured thresholds — QC checks pass):

| Column | Change | Effect |
|--------|--------|--------|
| `height` | +15 to all values | ~8.6% mean shift — triggers CP-04 |
| `mass` | ×1.10 | ~10% mean shift — triggers CP-04 |
| `birth_year` | −5 to all values | ~5.7% mean shift — triggers CP-04 |
| `homeworld` | 5 characters → "Exegol" | new distinct value — triggers CP-05 |
| `species` | 1 character → "Chiss" | new distinct value — triggers CP-05 |

**Custom checks** (`custom2/starwars_custom.R`):

| Check | Condition | Status |
|-------|-----------|--------|
| CC-01 | Human with height > 210 | FAIL |
| CC-02 | Human with height 200–210 | WARN |
| CC-03 | Human with mass > 140 | WARN |

Run: open `demo2.R` and source it. One HTML report opens in the browser
showing QC checks, custom checks, and version comparison.

Expected result: **FAIL** — `vehicles`/`starships` missing rate (standard)
and CC-01 Human height (custom). Version comparison shows distribution
shifts across height, mass, and birth_year, plus new distinct values in
homeworld and species.

# dqcheckr

Automated data quality checks for recurring dataset deliveries.

For each new file arrival, `dqcheckr` runs a battery of quality checks,
compares the file to the previous delivery, writes a self-contained HTML
report, and records summary statistics in a local SQLite database so that
quality trends can be tracked over time. Supports CSV and fixed-width formats.
Custom organisation-specific checks can be supplied as plain R files.

This is a CLI/API package — no UI. If you'd rather configure and run checks
without writing R code, see
[dqcheckrGUI](https://github.com/mickmioduszewski/dqcheckrGUI), a Shiny
front-end built on top of this package.

## What it does

- Runs single-snapshot quality checks (schema, missing rates, types, patterns,
  bounds, and more) against an incoming file
- Compares the file to the previous delivery (row/column differences, drift)
- Writes a self-contained HTML report for each run
- Records summary statistics in a SQLite snapshot database for long-term trend
  tracking and historical drift comparison
- Lets you supply your own organisation-specific checks as plain R files

## Installation

```r
install.packages("dqcheckr")

# or, the development version from GitHub
devtools::install_github("mickmioduszewski/dqcheckr")
```

## Usage

A data officer runs a single command for each arriving dataset:

```r
library(dqcheckr)

run_dq_check("customer_accounts", config_dir = "path/to/configs")
```

This prints a one-line console summary, writes an HTML report, and returns
`list(status, report_path, snapshot_id)` invisibly.

Two YAML files control every run: a global `dqcheckr.yml` (default thresholds
shared across datasets) and a per-dataset `<dataset_name>.yml` (file location,
expected columns, column-level rules and overrides).

## Learn more

See `vignette("dqcheckr")` for a full walkthrough of configuration and the
available checks, or the [package documentation
site](https://mickmioduszewski.github.io/dqcheckr/).

## License

MIT

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)

## ----one-call-----------------------------------------------------------------
# run_dq_check("customer_accounts", config_dir = "path/to/configs")

## ----install------------------------------------------------------------------
# devtools::install("path/to/dqr/dqcheckr")
# library(dqcheckr)

## ----type-inference-----------------------------------------------------------
# infer_col_type(c("2024-01-15", "2024-06-30"))  # "date"
# infer_col_type(c("100", "200", "N/A"))          # "character" (only 67% numeric)
# infer_col_type(c("100", "200", "N/A", rep("50", 17)))  # "numeric" (95% numeric)
# infer_col_type(c(NA, "", NA))                   # "unknown"
# # with a custom threshold:
# infer_col_type(c(rep("1", 17), "a", "b", "c"), threshold = 0.80)  # "numeric" (85%)

## ----run-check----------------------------------------------------------------
# result <- run_dq_check(
#   dataset_name = "customer_accounts",
#   config_dir   = "path/to/configs",   # contains dqcheckr.yml + customer_accounts.yml
#   open_report  = TRUE                 # open HTML in browser when interactive
# )
# 
# # Console output (always printed):
# # [dqcheckr] customer_accounts: FAIL - 1 warning(s), 2 failure(s). Report: reports/...html
# 
# result$status       # "PASS", "WARN", or "FAIL"
# result$report_path  # full path to the HTML file
# result$snapshot_id  # integer row ID in snapshots table

## ----individual-checks--------------------------------------------------------
# library(dqcheckr)
# 
# # Build a minimal config programmatically
# cfg <- list(
#   rules = list(
#     max_missing_rate     = 0.05,
#     max_non_numeric_rate = 0.01,
#     min_row_count        = 0
#   ),
#   column_rules     = list(
#     status = list(allowed_values = c("ACTIVE", "CLOSED"))
#   ),
#   key_columns      = "id",
#   expected_columns = NULL
# )
# 
# df <- read.csv("data/customer_accounts.csv", colClasses = "character")
# df[] <- lapply(df, trimws)
# 
# # Run any individual check
# check_missing_rate(df, cfg)       # QC-01
# check_empty_column(df, cfg)       # QC-02
# check_duplicate_rows(df, cfg)     # QC-03
# check_row_count(df, cfg)          # QC-04  (INFO)
# check_col_count(df, cfg)          # QC-05  (INFO)
# check_inferred_types(df, cfg)     # QC-06  (INFO)
# check_numeric_stats(df, cfg)      # QC-07  (INFO)
# check_distinct_counts(df, cfg)    # QC-08  (INFO)
# check_allowed_values(df, cfg)     # QC-09
# check_numeric_bounds(df, cfg)     # QC-10
# check_non_numeric(df, cfg)        # QC-11
# check_key_uniqueness(df, cfg)     # QC-12
# check_pattern(df, cfg)            # QC-13
# check_min_row_count(df, cfg)      # QC-14
# check_outliers(df, cfg)           # QC-15
# check_schema_contract(df, cfg)    # SC-01 / SC-02
# 
# # Inspect a result
# r <- check_missing_rate(df, cfg)[[1]]
# r$status    # "PASS" or "FAIL"
# r$observed  # e.g. "3.2% missing (4 of 125)"
# 
# # Run all single-snapshot checks at once
# all_results <- run_qc_checks(df, cfg)
# 
# # Overall status across any result list
# overall_status(all_results)  # "FAIL" > "WARN" > "PASS" > "INFO"

## ----custom-file--------------------------------------------------------------
# # File: custom/customer_accounts_checks.R
# 
# custom_checks <- function(df) {
#   results <- list()
# 
#   # Rule: ACTIVE accounts must not have a zero balance
#   active_zero <- df[df$account_status == "ACTIVE" &
#                     !is.na(df$account_balance) &
#                     df$account_balance == "0", ]
#   n <- nrow(active_zero)
#   results <- c(results, list(dq_result(
#     check_id   = "CUST-01",
#     check_name = "No zero-balance active accounts",
#     column     = "account_balance",   # enables per-column storage in SQLite
#     status     = if (n > 0) "FAIL" else "PASS",
#     observed   = sprintf("%d ACTIVE account(s) with balance 0", n),
#     message    = if (n > 0)
#       sprintf("%d ACTIVE account(s) have a zero balance.", n)
#     else
#       "No ACTIVE accounts have a zero balance."
#   )))
# 
#   results
# }

## ----snapshot-query-----------------------------------------------------------
# library(DBI)
# library(RSQLite)
# 
# con <- dbConnect(SQLite(), "data/snapshots.sqlite")
# 
# # Recent runs for one dataset
# dbGetQuery(con,
#   "SELECT id, file_name, overall_status, check_fail_count, run_timestamp
#    FROM snapshots
#    WHERE dataset_name = 'customer_accounts'
#    ORDER BY id DESC
#    LIMIT 10")
# 
# # Column-level stats for the most recent run
# dbGetQuery(con,
#   "SELECT column_name, dq_check, value, threshold
#    FROM column_snapshots
#    WHERE snapshot_id = (
#      SELECT MAX(id) FROM snapshots WHERE dataset_name = 'customer_accounts'
#    )
#    ORDER BY column_name, dq_check")
# 
# dbDisconnect(con)

## ----cross-dataset------------------------------------------------------------
# dbGetQuery(con,
#   "SELECT dataset_name, COUNT(*) AS runs,
#           SUM(check_fail_count) AS total_failures
#    FROM snapshots
#    GROUP BY dataset_name")

## ----starwars-setup-----------------------------------------------------------
# # Config directory contains:
# #   dqcheckr.yml         — global thresholds
# #   starwars_csv.yml     — CSV dataset config
# #   starwars_fwf.yml     — FWF dataset config
# 
# # Run checks on both formats
# result_csv <- run_dq_check("starwars_csv", config_dir = "config", open_report = TRUE)
# result_fwf <- run_dq_check("starwars_fwf", config_dir = "config", open_report = TRUE)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)

## -----------------------------------------------------------------------------
# run_dq_check("customer_accounts", config_dir = "config")

## -----------------------------------------------------------------------------
# dq_result(check_id, check_name, column = NA, status, observed, threshold = NA, message)
# # status: "PASS" | "WARN" | "FAIL" | "INFO"

## -----------------------------------------------------------------------------
# run_dq_check(
#   dataset_name,        # matches <dataset_name>.yml in config_dir
#   config_dir = ".",    # contains dqcheckr.yml and <dataset_name>.yml
#   open_report = TRUE   # open in browser when interactive
# )
# # Returns invisibly: list(status, report_path, snapshot_id)

## -----------------------------------------------------------------------------
# compare_snapshots(
#   dataset_name,
#   snapshot_id_prev = NULL,   # defaults to second-most-recent
#   snapshot_id_curr = NULL,   # defaults to most-recent
#   db_path          = NULL,   # defaults to snapshot_db from dqcheckr.yml
#   config_dir       = ".",
#   report           = TRUE,   # render HTML drift report
#   open_report      = interactive()
# )
# # Returns invisibly: named list with dataset_name, snap_prev, snap_curr,
# #   table_drift, schema_changes, missing_rate_changes, non_numeric_changes,
# #   mean_shifts, distinct_changes
# 
# list_snapshots(dataset_name = NULL, db_path)
# # Returns a data frame of snapshot records; dataset_name = NULL returns all datasets

#' Run organisation-specific custom checks
#'
#' Sources the R file specified by \code{config$custom_checks_file}, which must
#' define a function \code{custom_checks(df)} returning a list of
#' \code{\link{dq_result}} objects. Returns an empty list if
#' \code{custom_checks_file} is not set in the config.
#'
#' @param df A data frame. The current delivery.
#' @param config Named list. Merged configuration as returned by
#'   \code{\link{load_config}}.
#'
#' @return A list of \code{\link{dq_result}} objects (may be empty).
#'
#' @examples
#' cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg     <- load_config("starwars_csv", config_dir = cfg_dir)
#' path    <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
#' df      <- read_dataset(path, cfg)
#' results <- run_custom_checks(df, cfg)
#'
#' @export
run_custom_checks <- function(df, config) {
  path <- config$custom_checks_file
  if (is.null(path)) return(list())

  if (!file.exists(path)) {
    rlang::abort(paste0("Custom checks file not found: ", path))
  }

  env <- new.env(parent = baseenv())
  env$dq_result <- dq_result
  tryCatch(
    source(path, local = env),
    error = function(e)
      rlang::abort(paste0("Failed to source custom checks file '", path, "': ",
                          conditionMessage(e)))
  )

  if (!exists("custom_checks", envir = env, inherits = FALSE)) {
    rlang::abort(paste0("custom_checks() function not defined in: ", path))
  }

  results <- tryCatch(
    env$custom_checks(df),
    error = function(e)
      rlang::abort(paste0("custom_checks() runtime error: ", conditionMessage(e)))
  )

  if (!is.list(results)) {
    rlang::abort(paste0("custom_checks() must return a list of dq_result objects, got: ",
                        class(results)))
  }

  results
}

#' dqcheckr: Automated Data Quality Checks for Recurring Dataset Deliveries
#'
#' Automates quality verification of recurring external dataset deliveries.
#' For each new file arrival, it runs single-snapshot quality checks (QC-01 to
#' QC-14, SC-01/SC-02), compares the file to the previous delivery (CP-01 to
#' CP-08), writes a self-contained 'HTML' report, and records summary statistics
#' in a local 'SQLite' database for long-term trend tracking. Supports 'CSV' and
#' fixed-width formats. Custom organisation-specific checks can be supplied as
#' plain R files.
#'
#' The main entry point is \code{\link{run_dq_check}}. Configuration is driven
#' by two 'YAML' files: a global \code{dqcheckr.yml} and a per-dataset
#' \code{<dataset_name>.yml}.
#'
#' @keywords internal
#' @importFrom kableExtra kbl kable_styling cell_spec
#' @importFrom ggplot2 ggplot aes geom_line geom_point geom_col labs
#' @importFrom ggplot2 scale_fill_manual theme_minimal
#' @importFrom gridExtra grid.arrange
#' @importFrom dplyr %>%
#' @importFrom tidyr pivot_longer
#' @importFrom knitr raw_html
"_PACKAGE"

# =============================================================================
# demo.R — Named-file mode: CSV and FWF data quality checks
# =============================================================================
#
# Runs dqcheckr against the Star Wars dataset in two formats:
#   data/starwars.csv   — comma-separated file
#   data/starwars.fwf   — fixed-width file (same data)
#
# Both use explicit current_file: paths in their config (no folder scan).
# Results are written to output/reports/ and output/snapshots.sqlite.
#
# HOW TO RUN
#   1. Copy this demonstrations/ folder to a local directory:
#        dest <- file.path(path.expand("~"), "dqcheckr_demo")
#        file.copy(system.file("demonstrations", package = "dqcheckr"),
#                  dirname(dest), recursive = TRUE)
#        file.rename(file.path(dirname(dest), "demonstrations"), dest)
#   2. Open demo.R in RStudio and set the working directory to this folder:
#        Session -> Set Working Directory -> To Source File Location
#   3. Run the script (Ctrl+Enter or Source).
# =============================================================================

library(dqcheckr)

result_csv <- run_dq_check("starwars_csv", config_dir = "config", open_report = TRUE)
result_fwf <- run_dq_check("starwars_fwf", config_dir = "config", open_report = TRUE)

cat("\nCSV: status =", result_csv$status, "| report:", basename(result_csv$report_path), "\n")
cat("FWF: status =", result_fwf$status, "| report:", basename(result_fwf$report_path), "\n")

#' Null-coalescing operator
#' @keywords internal
#' @noRd
`%||%` <- function(a, b) if (!is.null(a)) a else b

#' Construct a data quality result object
#'
#' Creates the atomic result unit returned by every check function.
#'
#' @param check_id Character. Short identifier for the check (e.g. \code{"QC-01"}).
#' @param check_name Character. Human-readable name of the check.
#' @param column Character. Column the check applies to, or \code{NA_character_}
#'   for row-level or file-level checks.
#' @param status Character. One of \code{"PASS"}, \code{"WARN"}, \code{"FAIL"},
#'   or \code{"INFO"}.
#' @param observed Character. What was observed (e.g. \code{"5.2\% missing"}).
#' @param threshold Character. The configured threshold, or \code{NA_character_}
#'   if not applicable.
#' @param message Character. Human-readable description of the result.
#'
#' @return A named list with seven elements: \code{check_id}, \code{check_name},
#'   \code{column}, \code{status}, \code{observed}, \code{threshold},
#'   \code{message}.
#'
#' @examples
#' dq_result("QC-01", "Missing rate", column = "age",
#'           status = "PASS", observed = "0% missing",
#'           message = "No missing values.")
#'
#' @export
dq_result <- function(check_id, check_name, column = NA_character_,
                      status, observed, threshold = NA_character_, message) {
  valid_statuses <- c("PASS", "WARN", "FAIL", "INFO")
  if (is.null(status) || !status %in% valid_statuses) {
    rlang::abort(sprintf("status must be one of: %s (got: %s)",
                         paste(valid_statuses, collapse = ", "), status),
                 .internal = FALSE)
  }
  list(
    check_id   = check_id,
    check_name = check_name,
    column     = column,
    status     = status,
    observed   = as.character(observed),
    threshold  = if (is.na(threshold)) NA_character_ else as.character(threshold),
    message    = as.character(message)
  )
}

#' Load and merge dataset configuration
#'
#' Reads the global \code{dqcheckr.yml} and the dataset-specific YAML, merging
#' \code{rule_overrides} from the dataset config on top of \code{default_rules}
#' from the global config.
#'
#' @param dataset_name Character. Dataset name; must match
#'   \code{<dataset_name>.yml} in \code{config_dir}.
#' @param config_dir Character. Path to the directory containing both YAML
#'   files.
#'
#' @return A named list representing the merged configuration.
#'
#' @examples
#' cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg <- load_config("starwars_csv", config_dir = cfg_dir)
#' cfg$format
#'
#' @export
load_config <- function(dataset_name, config_dir) {
  global_path  <- file.path(config_dir, "dqcheckr.yml")
  dataset_path <- file.path(config_dir, paste0(dataset_name, ".yml"))

  if (!file.exists(global_path)) {
    rlang::abort(paste0("Global config not found: ", global_path))
  }
  if (!file.exists(dataset_path)) {
    rlang::abort(paste0("Dataset config not found: ", dataset_path))
  }

  global_cfg  <- yaml::read_yaml(global_path)
  dataset_cfg <- yaml::read_yaml(dataset_path)

  rules <- global_cfg$default_rules %||% list()

  if (!is.null(dataset_cfg$rule_overrides)) {
    for (key in names(dataset_cfg$rule_overrides)) {
      rules[[key]] <- dataset_cfg$rule_overrides[[key]]
    }
  }

  dataset_cfg$rules <- rules
  dataset_cfg
}

#' Infer the logical type of a character column
#'
#' Classifies a character vector as \code{"date"}, \code{"numeric"},
#' \code{"character"}, or \code{"unknown"} by applying rules in priority order.
#'
#' @param x Character vector to classify (as read from a CSV or FWF file).
#' @param threshold Numeric. Minimum proportion of non-empty values that must
#'   parse as numeric for the column to be classified as \code{"numeric"}.
#'   Defaults to \code{0.90}. Configurable via \code{type_inference_threshold}
#'   in \code{rule_overrides}.
#'
#' @return A single character string: \code{"date"}, \code{"numeric"},
#'   \code{"character"}, or \code{"unknown"}.
#'
#' @examples
#' infer_col_type(c("2024-01-01", "2024-06-15"))   # "date"
#' infer_col_type(c("1.5", "2.0", "3.1"))          # "numeric"
#' infer_col_type(c("high", "low", "medium"))       # "character"
#' infer_col_type(c(NA, "", NA))                    # "unknown"
#' infer_col_type(c(rep("1", 17), "a", "b", "c"), threshold = 0.80)  # "numeric"
#'
#' @export
infer_col_type <- function(x, threshold = 0.90) {
  non_empty <- x[!is.na(x) & x != ""]

  if (length(non_empty) == 0) return("unknown")

  date_formats <- c("%Y-%m-%d", "%d/%m/%Y", "%m/%d/%Y", "%Y%m%d", "%d-%m-%Y")
  for (fmt in date_formats) {
    parsed <- suppressWarnings(as.Date(non_empty, format = fmt))
    if (!any(is.na(parsed))) return("date")
  }

  numeric_vals <- suppressWarnings(as.numeric(non_empty))
  if (mean(!is.na(numeric_vals)) >= threshold) return("numeric")

  "character"
}

#' Compute the worst status across a list of dq_result objects
#'
#' Returns the single worst status in precedence order:
#' \code{"FAIL"} > \code{"WARN"} > \code{"PASS"} > \code{"INFO"}.
#'
#' @param results A list of \code{\link{dq_result}} objects.
#'
#' @return A single character string: \code{"FAIL"}, \code{"WARN"},
#'   \code{"PASS"}, or \code{"INFO"}.
#'
#' @examples
#' r1 <- dq_result("QC-01", "test", status = "PASS", observed = "ok", message = "ok")
#' r2 <- dq_result("QC-02", "test", status = "WARN", observed = "ok", message = "ok")
#' overall_status(list(r1, r2))  # "WARN"
#'
#' @export
overall_status <- function(results) {
  if (length(results) == 0) return("INFO")
  statuses <- vapply(results, `[[`, character(1), "status")
  if ("FAIL" %in% statuses) return("FAIL")
  if ("WARN" %in% statuses) return("WARN")
  if ("PASS" %in% statuses) return("PASS")
  "INFO"
}

library(testthat)
library(dqcheckr)

# -- dq_result() ---------------------------------------------------------------

test_that("dq_result() returns a named list with all seven fields", {
  r <- dq_result("QC-01", "Missing rate", column = "name",
                 status = "PASS", observed = "0%", message = "All good")
  expect_type(r, "list")
  expect_named(r, c("check_id", "check_name", "column", "status",
                    "observed", "threshold", "message"), ignore.order = FALSE)
})

test_that("dq_result() defaults column and threshold to NA_character_", {
  r <- dq_result("QC-04", "Row count", status = "INFO",
                 observed = "100", message = "Info only")
  expect_identical(r$column,    NA_character_)
  expect_identical(r$threshold, NA_character_)
})

test_that("dq_result() stops when status is not one of PASS/WARN/FAIL/INFO", {
  expect_error(
    dq_result("QC-01", "Missing rate", status = "ERROR",
              observed = "x", message = "x"),
    regexp = "status"
  )
})

test_that("dq_result() stops on NULL status", {
  expect_error(
    dq_result("QC-01", "Missing rate", status = NULL,
              observed = "x", message = "x")
  )
})

test_that("dq_result() accepts all four valid status values", {
  for (s in c("PASS", "WARN", "FAIL", "INFO")) {
    r <- dq_result("X-01", "test", status = s, observed = "x", message = "x")
    expect_equal(r$status, s)
  }
})

# -- overall_status() ----------------------------------------------------------

test_that("overall_status() returns FAIL when any result is FAIL", {
  results <- list(
    dq_result("A", "a", status = "PASS", observed = "x", message = "x"),
    dq_result("B", "b", status = "FAIL", observed = "x", message = "x"),
    dq_result("C", "c", status = "WARN", observed = "x", message = "x")
  )
  expect_equal(overall_status(results), "FAIL")
})

test_that("overall_status() returns WARN when no FAIL but some WARN", {
  results <- list(
    dq_result("A", "a", status = "PASS", observed = "x", message = "x"),
    dq_result("B", "b", status = "WARN", observed = "x", message = "x"),
    dq_result("C", "c", status = "INFO", observed = "x", message = "x")
  )
  expect_equal(overall_status(results), "WARN")
})

test_that("overall_status() returns PASS when only PASS and INFO", {
  results <- list(
    dq_result("A", "a", status = "PASS", observed = "x", message = "x"),
    dq_result("B", "b", status = "INFO", observed = "x", message = "x")
  )
  expect_equal(overall_status(results), "PASS")
})

test_that("overall_status() returns INFO on empty list", {
  expect_equal(overall_status(list()), "INFO")
})

# -- infer_col_type() ----------------------------------------------------------

test_that("infer_col_type() returns 'date' for ISO date strings", {
  x <- c("2024-01-01", "2024-06-15", "2023-12-31")
  expect_equal(infer_col_type(x), "date")
})

test_that("infer_col_type() returns 'date' for dd/mm/yyyy format", {
  x <- c("01/01/2024", "15/06/2024", "31/12/2023")
  expect_equal(infer_col_type(x), "date")
})

test_that("infer_col_type() returns 'numeric' when >=90% are numeric", {
  x <- c("1.5", "2.0", "3.1", "abc", rep("4.0", 16))  # 19/20 = 95% numeric
  expect_equal(infer_col_type(x), "numeric")
})

test_that("infer_col_type() returns 'character' when <90% are numeric", {
  x <- c("high", "low", "medium", "1.0", "2.0")  # 40% numeric
  expect_equal(infer_col_type(x), "character")
})

test_that("infer_col_type() returns 'unknown' for all empty/NA", {
  x <- c(NA, "", NA, "")
  expect_equal(infer_col_type(x), "unknown")
})

test_that("infer_col_type() returns 'numeric' for all-NA-except-numeric", {
  x <- c(NA, "1.0", "2.5", NA)
  expect_equal(infer_col_type(x), "numeric")
})

test_that("infer_col_type() respects a custom threshold", {
  # 85% numeric - below default 90%, above custom 80%
  x <- c(rep("1.0", 17), "abc", "def", "ghi")  # 17/20 = 85%
  expect_equal(infer_col_type(x, threshold = 0.90), "character")
  expect_equal(infer_col_type(x, threshold = 0.80), "numeric")
})

test_that("infer_col_type() threshold applies via config in check_inferred_types()", {
  df  <- data.frame(score = c(rep("1.0", 17), "bad", "bad", "bad"),
                    stringsAsFactors = FALSE)
  cfg_strict <- list(rules = list(type_inference_threshold = 0.90))
  cfg_lenient <- list(rules = list(type_inference_threshold = 0.80))
  strict  <- check_inferred_types(df, cfg_strict)
  lenient <- check_inferred_types(df, cfg_lenient)
  expect_equal(strict[[1]]$observed,  "character")
  expect_equal(lenient[[1]]$observed, "numeric")
})

# -- load_config() -------------------------------------------------------------

test_that("load_config() merges rule_overrides over defaults", {
  tmp <- tempdir()
  sub <- file.path(tmp, "cfg_merge_test")
  dir.create(sub, showWarnings = FALSE)
  writeLines(c(
    "default_rules:",
    "  max_missing_rate: 0.05",
    "  min_row_count: 0"
  ), file.path(sub, "dqcheckr.yml"))
  writeLines(c(
    "dataset_name: test_ds",
    "folder: data/",
    "format: csv",
    "rule_overrides:",
    "  max_missing_rate: 0.02"
  ), file.path(sub, "test_ds.yml"))

  cfg <- load_config("test_ds", sub)
  expect_equal(cfg$rules$max_missing_rate, 0.02)
  expect_equal(cfg$rules$min_row_count, 0)
})

test_that("load_config() stops when dqcheckr.yml is missing", {
  tmp <- tempdir()
  sub <- file.path(tmp, "no_global_yml")
  dir.create(sub, showWarnings = FALSE)
  expect_error(load_config("any", sub), regexp = "dqcheckr.yml")
})

test_that("load_config() stops when dataset yml is missing", {
  tmp <- tempdir()
  sub <- file.path(tmp, "no_dataset_yml")
  dir.create(sub, showWarnings = FALSE)
  writeLines(c("default_rules:", "  max_missing_rate: 0.05"),
             file.path(sub, "dqcheckr.yml"))
  expect_error(load_config("nonexistent_ds", sub))
})

# =============================================================================
# demo2.R — Folder-scan mode with version comparison and custom checks
# =============================================================================
#
# Demonstrates dqcheckr's folder-scan file detection, version comparison,
# and custom checks.
#
# data2/ contains two deliveries of the Star Wars dataset:
#   starwars_v1.csv  — original data  (used as "previous" delivery)
#   starwars_v2.csv  — perturbed data (used as "current" delivery)
#
# The perturbed version passes all single-snapshot quality checks but shows
# measurable shifts in column distributions that the CP comparison checks
# will flag:
#   height     +15 to all values    (~8.6% mean shift)
#   mass       x1.10                (~10%  mean shift)
#   birth_year -5 to all values     (~5.7% mean shift)
#   homeworld  5 characters moved to "Exegol"  (new distinct value)
#   species    1 character changed to "Chiss"   (new distinct value)
#
# Custom checks (custom2/starwars_custom.R) add three human-specific rules:
#   CC-01 FAIL  Human height > 210
#   CC-02 WARN  Human height 200-210
#   CC-03 WARN  Human mass > 140
#
# HOW TO RUN
#   1. Copy this demonstrations/ folder to a local directory:
#        dest <- file.path(path.expand("~"), "dqcheckr_demo")
#        file.copy(system.file("demonstrations", package = "dqcheckr"),
#                  dirname(dest), recursive = TRUE)
#        file.rename(file.path(dirname(dest), "demonstrations"), dest)
#   2. Open demo2.R in RStudio and set the working directory to this folder:
#        Session -> Set Working Directory -> To Source File Location
#   3. Run the script (Ctrl+Enter or Source).
# =============================================================================

library(dqcheckr)

# Ensure v2 is newer than v1 so folder scan picks v2 as current.
# Stamping both guarantees correct order regardless of when files were copied.
Sys.setFileTime("data2/starwars_v2.csv", Sys.time())
Sys.setFileTime("data2/starwars_v1.csv", Sys.time() - 3600)

result <- run_dq_check("starwars_folder", config_dir = "config2", open_report = TRUE)

cat("\nStatus     :", result$status, "\n")
cat("Report     :", result$report_path, "\n")
cat("Snapshot ID:", result$snapshot_id, "\n")

#' Compute missing rate for a vector
#' @keywords internal
.missing_rate_vec <- function(x) mean(is.na(x) | x == "")

#' CP-01: Compare row count between deliveries
#' @keywords internal
compare_row_count <- function(df_current, df_previous, config) {
  threshold  <- config$rules$max_row_count_change_pct %||% 0.10
  n_curr     <- nrow(df_current)
  n_prev     <- nrow(df_previous)
  pct_change <- (n_curr - n_prev) / n_prev
  status     <- if (abs(pct_change) > threshold) "WARN" else "PASS"
  list(dq_result(
    check_id   = "CP-01",
    check_name = "Row count change",
    status     = status,
    observed   = sprintf("%d rows (previous: %d; change: %+.1f%%)",
                         n_curr, n_prev, pct_change * 100),
    threshold  = sprintf("+/-%.0f%%", threshold * 100),
    message    = if (status == "WARN")
      sprintf("Row count changed by %+.1f%%, exceeding the %+.0f%% threshold.",
              pct_change * 100, threshold * 100)
    else
      sprintf("Row count change of %+.1f%% is within threshold.", pct_change * 100)
  ))
}

#' CP-02: Detect schema differences between deliveries
#' @keywords internal
compare_schema <- function(df_current, df_previous, config) {
  type_threshold <- config$rules$type_inference_threshold %||% 0.90
  flag_new   <- isTRUE(config$rules$flag_new_columns     %||% TRUE)
  flag_drop  <- isTRUE(config$rules$flag_dropped_columns %||% TRUE)
  flag_type  <- isTRUE(config$rules$flag_type_changes    %||% TRUE)

  new_cols     <- setdiff(names(df_current), names(df_previous))
  dropped_cols <- setdiff(names(df_previous), names(df_current))
  common_cols  <- intersect(names(df_current), names(df_previous))

  type_changes <- character(0)
  for (col in common_cols) {
    t_curr <- infer_col_type(df_current[[col]], type_threshold)
    t_prev <- infer_col_type(df_previous[[col]], type_threshold)
    if (t_curr != t_prev) {
      type_changes <- c(type_changes,
                        sprintf("%s (%s -> %s)", col, t_prev, t_curr))
    }
  }

  # Flags control what is reported; actual changes are always tracked for SQLite
  reported_new  <- if (flag_new)  new_cols     else character(0)
  reported_drop <- if (flag_drop) dropped_cols else character(0)
  reported_type <- if (flag_type) type_changes else character(0)

  has_change <- length(reported_new) > 0 || length(reported_drop) > 0 ||
    length(reported_type) > 0
  status <- if (has_change) "WARN" else "PASS"

  parts <- c(
    if (length(reported_new) > 0)
      paste("New columns:", paste(reported_new, collapse = ", ")),
    if (length(reported_drop) > 0)
      paste("Dropped columns:", paste(reported_drop, collapse = ", ")),
    if (length(reported_type) > 0)
      paste("Type changes:", paste(reported_type, collapse = "; "))
  )
  observed <- if (length(parts) > 0) paste(parts, collapse = ". ") else "No schema changes."

  result_list <- list(dq_result(
    check_id   = "CP-02",
    check_name = "Schema diff",
    status     = status,
    observed   = observed,
    message    = if (has_change)
      "Schema differences detected vs previous delivery."
    else
      "Schema is identical to previous delivery."
  ))

  attr(result_list, "new_cols")     <- new_cols
  attr(result_list, "dropped_cols") <- dropped_cols
  result_list
}

#' CP-03: Compare per-column missing rate between deliveries
#' @keywords internal
compare_missing_rate <- function(df_current, df_previous, config) {
  max_change_pp <- config$rules$max_missing_rate_change_pp %||% 2.0
  common_cols   <- intersect(names(df_current), names(df_previous))
  lapply(common_cols, function(col) {
    rate_curr <- .missing_rate_vec(df_current[[col]])
    rate_prev <- .missing_rate_vec(df_previous[[col]])
    change_pp <- (rate_curr - rate_prev) * 100
    status    <- if (change_pp > max_change_pp) "WARN" else "PASS"
    dq_result(
      check_id   = "CP-03",
      check_name = "Missing rate change",
      column     = col,
      status     = status,
      observed   = sprintf("%.1f%% (previous: %.1f%%; change: %+.1f pp)",
                           rate_curr * 100, rate_prev * 100, change_pp),
      threshold  = sprintf("+%.1f pp", max_change_pp),
      message    = if (status == "WARN")
        sprintf("Column '%s' missing rate increased by %.1f pp, exceeding threshold.", col, change_pp)
      else
        sprintf("Column '%s' missing rate change is within threshold.", col)
    )
  })
}

#' CP-04: Compare numeric column means between deliveries
#' @keywords internal
compare_numeric_mean <- function(df_current, df_previous, config) {
  threshold      <- config$rules$max_numeric_mean_shift_pct %||% 0.20
  type_threshold <- config$rules$type_inference_threshold %||% 0.90
  common_cols <- intersect(names(df_current), names(df_previous))
  results     <- list()
  for (col in common_cols) {
    if (infer_col_type(df_current[[col]],  type_threshold) != "numeric") next
    if (infer_col_type(df_previous[[col]], type_threshold) != "numeric") next
    mean_curr <- mean(suppressWarnings(as.numeric(df_current[[col]])),  na.rm = TRUE)
    mean_prev <- mean(suppressWarnings(as.numeric(df_previous[[col]])), na.rm = TRUE)
    if (is.nan(mean_prev) || mean_prev == 0) next
    shift_pct <- abs((mean_curr - mean_prev) / mean_prev)
    status    <- if (shift_pct > threshold) "WARN" else "PASS"
    results <- c(results, list(dq_result(
      check_id   = "CP-04",
      check_name = "Numeric mean shift",
      column     = col,
      status     = status,
      observed   = sprintf("mean: %.4g (previous: %.4g; change: %+.1f%%)",
                           mean_curr, mean_prev,
                           (mean_curr - mean_prev) / abs(mean_prev) * 100),
      threshold  = sprintf("+/-%.0f%%", threshold * 100),
      message    = if (status == "WARN")
        sprintf("Column '%s' mean shifted by %.1f%%, exceeding threshold.",
                col, shift_pct * 100)
      else
        sprintf("Column '%s' mean shift is within threshold.", col)
    )))
  }
  results
}

#' CP-05: Detect new distinct values in character columns
#' @keywords internal
compare_new_values <- function(df_current, df_previous, config) {
  type_threshold <- config$rules$type_inference_threshold %||% 0.90
  common_cols <- intersect(names(df_current), names(df_previous))
  results     <- list()
  for (col in common_cols) {
    if (infer_col_type(df_current[[col]], type_threshold) != "character") next
    curr_vals <- unique(df_current[[col]][!is.na(df_current[[col]]) &
                                          df_current[[col]] != ""])
    prev_vals <- unique(df_previous[[col]][!is.na(df_previous[[col]]) &
                                           df_previous[[col]] != ""])
    new_vals  <- setdiff(curr_vals, prev_vals)
    results <- c(results, list(dq_result(
      check_id   = "CP-05",
      check_name = "New distinct values",
      column     = col,
      status     = "INFO",
      observed   = if (length(new_vals) > 0)
        paste("New values:", paste(new_vals, collapse = ", "))
      else
        "No new values.",
      message    = sprintf("Column '%s': %d new distinct value(s) vs previous.",
                           col, length(new_vals))
    )))
  }
  results
}

#' CP-06: Detect dropped distinct values in character columns
#' @keywords internal
compare_dropped_values <- function(df_current, df_previous, config) {
  type_threshold <- config$rules$type_inference_threshold %||% 0.90
  common_cols <- intersect(names(df_current), names(df_previous))
  results     <- list()
  for (col in common_cols) {
    if (infer_col_type(df_current[[col]], type_threshold) != "character") next
    curr_vals     <- unique(df_current[[col]][!is.na(df_current[[col]]) &
                                              df_current[[col]] != ""])
    prev_vals     <- unique(df_previous[[col]][!is.na(df_previous[[col]]) &
                                               df_previous[[col]] != ""])
    dropped_vals  <- setdiff(prev_vals, curr_vals)
    results <- c(results, list(dq_result(
      check_id   = "CP-06",
      check_name = "Dropped distinct values",
      column     = col,
      status     = "INFO",
      observed   = if (length(dropped_vals) > 0)
        paste("Dropped values:", paste(dropped_vals, collapse = ", "))
      else
        "No dropped values.",
      message    = sprintf("Column '%s': %d value(s) from previous not in current.",
                           col, length(dropped_vals))
    )))
  }
  results
}

#' CP-07: Compare non-numeric rate in numeric columns between deliveries
#' @keywords internal
compare_non_numeric_rate <- function(df_current, df_previous, config) {
  threshold      <- config$rules$max_non_numeric_rate_change_pp %||% 1.0
  type_threshold <- config$rules$type_inference_threshold %||% 0.90
  common_cols <- intersect(names(df_current), names(df_previous))
  results     <- list()
  for (col in common_cols) {
    t_curr <- infer_col_type(df_current[[col]], type_threshold)
    t_prev <- infer_col_type(df_previous[[col]], type_threshold)
    if (!t_curr %in% c("numeric", "character") &&
        !t_prev %in% c("numeric", "character")) next
    if (t_curr == "character" && t_prev == "character") next

    .nn_rate <- function(x) {
      ne <- x[!is.na(x) & x != ""]
      if (length(ne) == 0) return(0)
      length(ne[is.na(suppressWarnings(as.numeric(ne)))]) / length(ne)
    }

    rate_curr <- .nn_rate(df_current[[col]])
    rate_prev <- .nn_rate(df_previous[[col]])
    change_pp <- (rate_curr - rate_prev) * 100
    if (change_pp <= 0) next

    status <- if (change_pp > threshold) "WARN" else "PASS"
    results <- c(results, list(dq_result(
      check_id   = "CP-07",
      check_name = "Non-numeric rate change",
      column     = col,
      status     = status,
      observed   = sprintf("%.2f%% (previous: %.2f%%; change: %+.2f pp)",
                           rate_curr * 100, rate_prev * 100, change_pp),
      threshold  = sprintf("+%.1f pp", threshold),
      message    = if (status == "WARN")
        sprintf("Column '%s' non-numeric rate increased by %.2f pp.", col, change_pp)
      else
        sprintf("Column '%s' non-numeric rate change is within threshold.", col)
    )))
  }
  results
}

#' CP-08: Check column order consistency between deliveries
#' @keywords internal
compare_column_order <- function(df_current, df_previous, config) {
  if (!isTRUE(config$rules$flag_column_order_change %||% TRUE)) return(list())
  fmt        <- tolower(config$format %||% "csv")
  curr_names <- names(df_current)
  prev_names <- names(df_previous)

  if (!identical(curr_names, prev_names)) {
    status <- if (fmt == "fwf") "FAIL" else "WARN"
    return(list(dq_result(
      check_id   = "CP-08",
      check_name = "Column order change",
      status     = status,
      observed   = sprintf("Current: [%s] | Previous: [%s]",
                           paste(curr_names, collapse = ", "),
                           paste(prev_names, collapse = ", ")),
      message    = if (fmt == "fwf")
        "Column order has changed. This is an error for fixed-width files."
      else
        "Column order has changed vs previous delivery."
    )))
  }

  list(dq_result(
    check_id   = "CP-08",
    check_name = "Column order change",
    status     = "PASS",
    observed   = "Column order is unchanged.",
    message    = "Column order matches previous delivery."
  ))
}

#' Run all version comparison checks between two dataset snapshots
#'
#' Runs CP-01 to CP-08 comparing a current delivery against the previous one.
#'
#' @param df_current A data frame. The current delivery.
#' @param df_previous A data frame. The previous delivery.
#' @param config Named list. Merged configuration as returned by
#'   \code{\link{load_config}}.
#'
#' @return A list of \code{\link{dq_result}} objects. The list carries
#'   attributes \code{new_cols} and \code{dropped_cols} (character vectors)
#'   for use by the snapshot writer.
#'
#' @examples
#' cfg_dir   <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg       <- load_config("starwars_csv", config_dir = cfg_dir)
#' curr_path <- system.file("demonstrations/data2/starwars_v2.csv", package = "dqcheckr")
#' prev_path <- system.file("demonstrations/data2/starwars_v1.csv", package = "dqcheckr")
#' curr      <- read_dataset(curr_path, cfg)
#' prev      <- read_dataset(prev_path, cfg)
#' results   <- run_comparison_checks(curr, prev, cfg)
#'
#' @export
run_comparison_checks <- function(df_current, df_previous, config) {
  schema_res <- compare_schema(df_current, df_previous, config)
  new_cols     <- attr(schema_res, "new_cols")
  dropped_cols <- attr(schema_res, "dropped_cols")

  results <- c(
    compare_row_count(df_current, df_previous, config),
    schema_res,
    compare_missing_rate(df_current, df_previous, config),
    compare_numeric_mean(df_current, df_previous, config),
    compare_new_values(df_current, df_previous, config),
    compare_dropped_values(df_current, df_previous, config),
    compare_non_numeric_rate(df_current, df_previous, config),
    compare_column_order(df_current, df_previous, config)
  )

  attr(results, "new_cols")     <- new_cols
  attr(results, "dropped_cols") <- dropped_cols
  results
}

library(testthat)
library(dqcheckr)

base_config <- function() {
  list(
    format = "csv",
    rules = list(
      max_missing_rate               = 0.05,
      max_non_numeric_rate           = 0.01,
      min_row_count                  = 0,
      max_row_count_change_pct       = 0.10,
      max_numeric_mean_shift_pct     = 0.20,
      max_missing_rate_change_pp     = 2.0,
      max_non_numeric_rate_change_pp = 1.0
    ),
    column_rules     = list(),
    key_columns      = NULL,
    expected_columns = NULL
  )
}

make_curr <- function() {
  data.frame(
    id              = paste0("ID00", 1:5),
    name            = c("Alice", "Bob", "Clara", "David", "Emma"),
    country_code    = c("GB", "US", "DE", "FR", "GB"),
    account_status  = c("ACTIVE", "ACTIVE", "CLOSED", "ACTIVE", "SUSPENDED"),
    account_balance = c("15000", "8500", "0", "250000", "1200"),
    created_date    = c("2024-01-15", "2024-02-20", "2023-11-01",
                        "2024-03-10", "2024-04-05"),
    stringsAsFactors = FALSE
  )
}

make_prev <- function() {
  data.frame(
    id              = paste0("ID00", 1:5),
    name            = c("Alice", "Bob", "Clara", "David", "Emma"),
    country_code    = c("GB", "US", "DE", "FR", "GB"),
    account_status  = c("ACTIVE", "ACTIVE", "CLOSED", "ACTIVE", "SUSPENDED"),
    account_balance = c("14500", "8200", "0", "248000", "1100"),
    created_date    = c("2024-01-15", "2024-02-20", "2023-11-01",
                        "2024-03-10", "2024-04-05"),
    stringsAsFactors = FALSE
  )
}

# -- CP-01 Row count change ----------------------------------------------------

test_that("compare_row_count() returns PASS when change is within threshold", {
  res <- compare_row_count(make_curr(), make_prev(), base_config())
  expect_equal(res[[1]]$status, "PASS")
})

test_that("compare_row_count() returns WARN when change exceeds threshold", {
  curr <- make_curr()[1:2, ]
  res  <- compare_row_count(curr, make_prev(), base_config())
  expect_equal(res[[1]]$status, "WARN")
})

# -- CP-02 Schema diff ---------------------------------------------------------

test_that("compare_schema() returns PASS when schemas are identical", {
  res <- compare_schema(make_curr(), make_prev(), base_config())
  expect_equal(res[[1]]$status, "PASS")
})

test_that("compare_schema() returns WARN when current has new column", {
  curr       <- make_curr()
  curr$extra <- "x"
  res <- compare_schema(curr, make_prev(), base_config())
  expect_equal(res[[1]]$status, "WARN")
  expect_true("extra" %in% attr(res, "new_cols"))
})

test_that("compare_schema() returns WARN when column is dropped", {
  curr <- make_curr()[, -3]
  res  <- compare_schema(curr, make_prev(), base_config())
  expect_equal(res[[1]]$status, "WARN")
  expect_true("country_code" %in% attr(res, "dropped_cols"))
})

test_that("compare_schema() returns WARN on type change", {
  curr                  <- make_curr()
  curr$account_balance  <- c("high", "low", "medium", "high", "low")
  res <- compare_schema(curr, make_prev(), base_config())
  expect_equal(res[[1]]$status, "WARN")
})

test_that("compare_schema() attaches new_cols and dropped_cols attributes", {
  curr       <- make_curr()
  curr$new1  <- "x"
  res <- compare_schema(curr, make_prev(), base_config())
  expect_false(is.null(attr(res, "new_cols")))
  expect_false(is.null(attr(res, "dropped_cols")))
})

# -- CP-03 Missing rate change -------------------------------------------------

test_that("compare_missing_rate() returns PASS when missing rate unchanged", {
  res <- compare_missing_rate(make_curr(), make_prev(), base_config())
  expect_true(all(vapply(res, \(r) r$status == "PASS", logical(1))))
})

test_that("compare_missing_rate() returns WARN when missing rate increases significantly", {
  curr       <- make_curr()
  curr$name  <- NA_character_
  res        <- compare_missing_rate(curr, make_prev(), base_config())
  name_r     <- Filter(\(r) r$column == "name", res)
  expect_equal(name_r[[1]]$status, "WARN")
})

# -- CP-04 Numeric mean shift --------------------------------------------------

test_that("compare_numeric_mean() returns PASS when mean shift is small", {
  res <- compare_numeric_mean(make_curr(), make_prev(), base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "PASS")
})

test_that("compare_numeric_mean() returns WARN when mean shift exceeds threshold", {
  curr                 <- make_curr()
  curr$account_balance <- as.character(
    as.numeric(make_curr()$account_balance) * 10
  )
  res <- compare_numeric_mean(curr, make_prev(), base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "WARN")
})

# -- CP-05 New distinct values -------------------------------------------------

test_that("compare_new_values() returns INFO always", {
  res <- compare_new_values(make_curr(), make_prev(), base_config())
  expect_true(all(vapply(res, \(r) r$status == "INFO", logical(1))))
})

test_that("compare_new_values() reports new values in observed field", {
  curr                 <- make_curr()
  curr$country_code[1] <- "JP"
  res <- compare_new_values(curr, make_prev(), base_config())
  cc  <- Filter(\(r) r$column == "country_code", res)
  expect_match(cc[[1]]$observed, "JP")
})

# -- CP-06 Dropped distinct values --------------------------------------------

test_that("compare_dropped_values() returns INFO always", {
  res <- compare_dropped_values(make_curr(), make_prev(), base_config())
  expect_true(all(vapply(res, \(r) r$status == "INFO", logical(1))))
})

# -- CP-07 Non-numeric rate change ---------------------------------------------

test_that("compare_non_numeric_rate() returns PASS when rate unchanged", {
  res <- compare_non_numeric_rate(make_curr(), make_prev(), base_config())
  statuses <- vapply(res, `[[`, character(1), "status")
  expect_true(all(statuses %in% c("PASS", "INFO")))
})

test_that("compare_non_numeric_rate() returns WARN when rate increases significantly", {
  curr                 <- make_curr()
  curr$account_balance <- c("N/A", "unknown", "bad", "1000", "2000")
  res <- compare_non_numeric_rate(curr, make_prev(), base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "WARN")
})

# -- CP-08 Column order --------------------------------------------------------

test_that("compare_column_order() returns PASS when column order unchanged", {
  res <- compare_column_order(make_curr(), make_prev(), base_config())
  expect_equal(res[[1]]$status, "PASS")
})

test_that("compare_column_order() returns WARN for CSV when order changes", {
  curr     <- make_curr()[, rev(names(make_curr()))]
  cfg      <- base_config()
  cfg$format <- "csv"
  res <- compare_column_order(curr, make_prev(), cfg)
  expect_equal(res[[1]]$status, "WARN")
})

test_that("compare_column_order() returns FAIL for FWF when order changes", {
  curr     <- make_curr()[, rev(names(make_curr()))]
  cfg      <- base_config()
  cfg$format <- "fwf"
  res <- compare_column_order(curr, make_prev(), cfg)
  expect_equal(res[[1]]$status, "FAIL")
})

# -- flag_* config keys --------------------------------------------------------

test_that("compare_schema() suppresses new columns when flag_new_columns=FALSE", {
  curr <- make_curr()
  curr$extra_col <- "x"
  cfg <- base_config()
  cfg$rules$flag_new_columns <- FALSE
  res <- compare_schema(curr, make_prev(), cfg)
  expect_equal(res[[1]]$status, "PASS")
  expect_false(grepl("New columns", res[[1]]$observed))
  expect_equal(attr(res, "new_cols"), "extra_col")  # still tracked for SQLite
})

test_that("compare_schema() suppresses dropped columns when flag_dropped_columns=FALSE", {
  prev <- make_prev()
  prev$ghost_col <- "x"
  cfg <- base_config()
  cfg$rules$flag_dropped_columns <- FALSE
  res <- compare_schema(make_curr(), prev, cfg)
  expect_equal(res[[1]]$status, "PASS")
  expect_false(grepl("Dropped", res[[1]]$observed))
  expect_equal(attr(res, "dropped_cols"), "ghost_col")
})

test_that("compare_schema() suppresses type changes when flag_type_changes=FALSE", {
  prev <- make_prev()
  prev$account_balance <- c("high", "low", "mid", "zero", "neg")
  cfg <- base_config()
  cfg$rules$flag_type_changes <- FALSE
  res <- compare_schema(make_curr(), prev, cfg)
  expect_equal(res[[1]]$status, "PASS")
  expect_false(grepl("Type changes", res[[1]]$observed))
})

test_that("compare_column_order() returns empty list when flag_column_order_change=FALSE", {
  curr <- make_curr()[, rev(names(make_curr()))]
  cfg  <- base_config()
  cfg$rules$flag_column_order_change <- FALSE
  res  <- compare_column_order(curr, make_prev(), cfg)
  expect_length(res, 0)
})

test_that("run_comparison_checks() preserves new_cols and dropped_cols attributes", {
  curr <- make_curr()
  curr$extra_col <- "x"
  prev <- make_prev()
  prev$ghost_col <- "y"
  res <- run_comparison_checks(curr, prev, base_config())
  expect_equal(attr(res, "new_cols"),     "extra_col")
  expect_equal(attr(res, "dropped_cols"), "ghost_col")
})

test_that("run_comparison_checks() attributes are NULL when no schema changes", {
  res <- run_comparison_checks(make_curr(), make_prev(), base_config())
  expect_length(attr(res, "new_cols"),     0)
  expect_length(attr(res, "dropped_cols"), 0)
})

#' Render the HTML data quality report
#' @keywords internal
render_report <- function(dataset_name, file_name, file_path, df,
                          qc_results, cp_results, custom_results,
                          snapshot_history, config, col_stats = NULL, output_dir,
                          open_report = TRUE) {
  template <- system.file("templates", "report.Rmd", package = "dqcheckr")
  if (!nzchar(template)) {
    rlang::abort("Report template not found in package installation.")
  }

  # Normalise to absolute path: rmarkdown changes wd to the template dir,
  # which breaks any relative path supplied as output_file.
  dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
  output_dir <- normalizePath(output_dir, mustWork = FALSE)
  file_path  <- normalizePath(file_path,  mustWork = FALSE)

  ts    <- format(Sys.time(), "%Y%m%d_%H%M%S")
  fname <- sprintf("%s_%s.html", dataset_name, ts)
  out   <- file.path(output_dir, fname)

  if (is.null(col_stats)) col_stats <- compute_col_stats(df, config, qc_results)

  rmarkdown::render(
    input       = template,
    output_file = out,
    params      = list(
      dataset_name     = dataset_name,
      file_name        = file_name,
      file_path        = file_path,
      run_timestamp    = format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
      df               = df,
      qc_results       = qc_results,
      cp_results       = cp_results,
      custom_results   = custom_results,
      snapshot_history = snapshot_history,
      config           = config,
      col_stats        = col_stats,
      overall_status   = overall_status(c(qc_results, cp_results,
                                          custom_results))
    ),
    quiet = TRUE
  )

  if (open_report && interactive()) utils::browseURL(out)

  invisible(out)
}

#' Initialise the SQLite snapshot database
#' @keywords internal
init_snapshot_db <- function(db_path) {
  dir.create(dirname(db_path), recursive = TRUE, showWarnings = FALSE)
  con <- DBI::dbConnect(RSQLite::SQLite(), db_path)
  on.exit(DBI::dbDisconnect(con))

  DBI::dbExecute(con, "
    CREATE TABLE IF NOT EXISTS snapshots (
      id                       INTEGER PRIMARY KEY AUTOINCREMENT,
      dataset_name             TEXT    NOT NULL,
      run_timestamp            TEXT    NOT NULL,
      file_name                TEXT    NOT NULL,
      row_count                INTEGER NOT NULL,
      col_count                INTEGER NOT NULL,
      check_pass_count         INTEGER NOT NULL DEFAULT 0,
      check_warn_count         INTEGER NOT NULL DEFAULT 0,
      check_fail_count         INTEGER NOT NULL DEFAULT 0,
      check_info_count         INTEGER NOT NULL DEFAULT 0,
      overall_status           TEXT    NOT NULL,
      new_cols_vs_previous     TEXT,
      missing_cols_vs_previous TEXT,
      new_cols_vs_schema       TEXT,
      missing_cols_vs_schema   TEXT
    )
  ")

  DBI::dbExecute(con, "
    CREATE TABLE IF NOT EXISTS column_snapshots (
      id                 INTEGER PRIMARY KEY AUTOINCREMENT,
      snapshot_id        INTEGER NOT NULL REFERENCES snapshots(id),
      column_name        TEXT    NOT NULL,
      dq_check           TEXT    NOT NULL,
      value              TEXT,
      threshold          TEXT,
      severity_on_breach TEXT
    )
  ")

  invisible(db_path)
}

#' Compute per-column statistics for snapshot storage
#' @keywords internal
#' @importFrom stats sd
compute_col_stats <- function(df, config, qc_results) {
  type_threshold <- config$rules$type_inference_threshold %||% 0.90
  rows <- list()

  for (col in names(df)) {
    x         <- df[[col]]
    col_type  <- infer_col_type(x, type_threshold)
    non_empty <- x[!is.na(x) & x != ""]
    miss_count <- sum(is.na(x) | x == "")
    miss_rate  <- miss_count / nrow(df)
    dist_count <- length(unique(non_empty))

    rows <- c(rows, list(
      data.frame(column_name = col, dq_check = "inferred_type",
                 value = col_type, threshold = NA_character_,
                 severity_on_breach = NA_character_, stringsAsFactors = FALSE),
      data.frame(column_name = col, dq_check = "missing_count",
                 value = as.character(miss_count), threshold = NA_character_,
                 severity_on_breach = NA_character_, stringsAsFactors = FALSE),
      data.frame(column_name = col, dq_check = "missing_rate",
                 value = as.character(miss_rate),
                 threshold = as.character(config$rules$max_missing_rate %||% 0.05),
                 severity_on_breach = "FAIL", stringsAsFactors = FALSE),
      data.frame(column_name = col, dq_check = "distinct_count",
                 value = as.character(dist_count), threshold = NA_character_,
                 severity_on_breach = NA_character_, stringsAsFactors = FALSE)
    ))

    if (col_type == "numeric") {
      vals <- suppressWarnings(as.numeric(x))
      nn   <- vals[!is.na(vals)]
      nn_count <- sum(!is.na(non_empty) &
                      is.na(suppressWarnings(as.numeric(non_empty))))
      nn_rate  <- if (length(non_empty) > 0) nn_count / length(non_empty) else 0

      rows <- c(rows, list(
        data.frame(column_name = col, dq_check = "numeric_mean",
                   value = if (length(nn) > 0) as.character(mean(nn)) else NA_character_,
                   threshold = NA_character_, severity_on_breach = NA_character_,
                   stringsAsFactors = FALSE),
        data.frame(column_name = col, dq_check = "numeric_sd",
                   value = if (length(nn) > 1) as.character(sd(nn)) else NA_character_,
                   threshold = NA_character_, severity_on_breach = NA_character_,
                   stringsAsFactors = FALSE),
        data.frame(column_name = col, dq_check = "numeric_min",
                   value = if (length(nn) > 0) as.character(min(nn)) else NA_character_,
                   threshold = NA_character_, severity_on_breach = NA_character_,
                   stringsAsFactors = FALSE),
        data.frame(column_name = col, dq_check = "numeric_max",
                   value = if (length(nn) > 0) as.character(max(nn)) else NA_character_,
                   threshold = NA_character_, severity_on_breach = NA_character_,
                   stringsAsFactors = FALSE),
        data.frame(column_name = col, dq_check = "non_numeric_count",
                   value = as.character(nn_count), threshold = NA_character_,
                   severity_on_breach = NA_character_, stringsAsFactors = FALSE),
        data.frame(column_name = col, dq_check = "non_numeric_rate",
                   value = as.character(nn_rate),
                   threshold = as.character(config$rules$max_non_numeric_rate %||% 0.01),
                   severity_on_breach = "FAIL", stringsAsFactors = FALSE)
      ))
    }
  }

  do.call(rbind, rows)
}

#' Write a run snapshot to the SQLite database
#' @keywords internal
write_snapshot <- function(db_path, dataset_name, file_name, df,
                           qc_results, cp_results, custom_results, config,
                           col_stats = NULL) {
  tryCatch({
    init_snapshot_db(db_path)
    con <- DBI::dbConnect(RSQLite::SQLite(), db_path)
    on.exit(DBI::dbDisconnect(con))

    all_results <- c(qc_results, cp_results, custom_results)
    statuses    <- vapply(all_results, `[[`, character(1), "status")

    pass_count <- sum(statuses == "PASS")
    warn_count <- sum(statuses == "WARN")
    fail_count <- sum(statuses == "FAIL")
    info_count <- sum(statuses == "INFO")
    o_status   <- overall_status(all_results)

    new_cols_prev  <- attr(cp_results, "new_cols")
    drop_cols_prev <- attr(cp_results, "dropped_cols")
    new_cols_prev_str  <- if (length(new_cols_prev) > 0)
      paste(new_cols_prev, collapse = ",") else NA_character_
    drop_cols_prev_str <- if (length(drop_cols_prev) > 0)
      paste(drop_cols_prev, collapse = ",") else NA_character_

    expected <- config$expected_columns
    if (!is.null(expected)) {
      sc01_fail <- Filter(\(r) r$check_id == "SC-01" && r$status == "FAIL",
                          qc_results)
      sc02_fail <- Filter(\(r) r$check_id == "SC-02" && r$status == "FAIL",
                          qc_results)
      new_schema  <- if (length(sc01_fail) > 0)
        paste(vapply(sc01_fail, \(r) r$column, character(1)), collapse = ",")
      else NA_character_
      miss_schema <- if (length(sc02_fail) > 0)
        paste(vapply(sc02_fail, \(r) r$column, character(1)), collapse = ",")
      else NA_character_
    } else {
      new_schema  <- NA_character_
      miss_schema <- NA_character_
    }

    DBI::dbExecute(con,
      "INSERT INTO snapshots
       (dataset_name, run_timestamp, file_name, row_count, col_count,
        check_pass_count, check_warn_count, check_fail_count, check_info_count,
        overall_status, new_cols_vs_previous, missing_cols_vs_previous,
        new_cols_vs_schema, missing_cols_vs_schema)
       VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
      list(dataset_name,
           format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
           file_name,
           nrow(df), ncol(df),
           pass_count, warn_count, fail_count, info_count,
           o_status,
           new_cols_prev_str, drop_cols_prev_str,
           new_schema, miss_schema)
    )

    snapshot_id <- DBI::dbGetQuery(con,
      "SELECT last_insert_rowid() AS id")$id[[1]]

    if (is.null(col_stats)) col_stats <- compute_col_stats(df, config, qc_results)
    col_stats$snapshot_id <- snapshot_id

    DBI::dbAppendTable(con, "column_snapshots",
      col_stats[, c("snapshot_id", "column_name", "dq_check",
                    "value", "threshold", "severity_on_breach")])

    for (r in custom_results) {
      if (!is.na(r$column)) {
        sev <- if (r$status %in% c("WARN", "FAIL")) r$status else NA_character_
        DBI::dbExecute(con,
          "INSERT INTO column_snapshots
           (snapshot_id, column_name, dq_check, value, threshold, severity_on_breach)
           VALUES (?, ?, ?, ?, ?, ?)",
          list(snapshot_id, r$column, r$check_id,
               r$observed, r$threshold, sev))
      }
    }

    snapshot_id
  },
  error = function(e) {
    warning("SQLite write failed: ", conditionMessage(e))
    NULL
  })
}

#' Read recent snapshot history from the SQLite database
#'
#' Retrieves the \code{n} most recent run records for a given dataset from the
#' snapshot database, ordered newest-first.
#'
#' @param db_path Character. Path to the SQLite database file.
#' @param dataset_name Character. Dataset name to filter on.
#' @param n Integer. Maximum number of records to return. Defaults to 10.
#'
#' @return A data frame with one row per run and columns including
#'   \code{id}, \code{run_timestamp}, \code{file_name}, \code{row_count},
#'   \code{overall_status}, \code{check_pass_count}, \code{check_warn_count},
#'   \code{check_fail_count}. Returns an empty data frame if the database does
#'   not exist or contains no records for the dataset.
#'
#' @examples
#' history <- read_recent_snapshots(tempfile(fileext = ".sqlite"), "starwars_csv")
#'
#' @export
read_recent_snapshots <- function(db_path, dataset_name, n = 10) {
  empty <- data.frame(
    id = integer(), dataset_name = character(), run_timestamp = character(),
    file_name = character(), row_count = integer(), col_count = integer(),
    check_pass_count = integer(), check_warn_count = integer(),
    check_fail_count = integer(), check_info_count = integer(),
    overall_status = character(), new_cols_vs_previous = character(),
    missing_cols_vs_previous = character(), new_cols_vs_schema = character(),
    missing_cols_vs_schema = character(),
    stringsAsFactors = FALSE
  )

  if (!file.exists(db_path)) return(empty)

  tryCatch({
    con <- DBI::dbConnect(RSQLite::SQLite(), db_path)
    on.exit(DBI::dbDisconnect(con))
    if (!"snapshots" %in% DBI::dbListTables(con)) return(empty)
    DBI::dbGetQuery(con,
      "SELECT * FROM snapshots
       WHERE dataset_name = ?
       ORDER BY id DESC
       LIMIT ?",
      list(dataset_name, as.integer(n)))
  },
  error = function(e) empty)
}

library(testthat)
library(dqcheckr)

make_df <- function() {
  data.frame(
    id              = paste0("ID00", 1:5),
    account_balance = c("15000", "8500", "0", "250000", "1200"),
    stringsAsFactors = FALSE
  )
}

base_config <- function() {
  list(
    format       = "csv",
    rules        = list(
      max_missing_rate     = 0.05,
      max_non_numeric_rate = 0.01,
      min_row_count        = 0
    ),
    column_rules     = list(),
    key_columns      = NULL,
    expected_columns = NULL
  )
}

make_results <- function() {
  list(
    dq_result("QC-01", "Missing rate", column = "id",
              status = "PASS", observed = "0%", message = "OK"),
    dq_result("QC-04", "Row count",
              status = "INFO", observed = "5", message = "5 rows")
  )
}

# -- init_snapshot_db() --------------------------------------------------------

test_that("init_snapshot_db() creates both tables on a new database", {
  db  <- tempfile(fileext = ".sqlite")
  init_snapshot_db(db)
  con <- DBI::dbConnect(RSQLite::SQLite(), db)
  tables <- DBI::dbListTables(con)
  DBI::dbDisconnect(con)
  expect_true("snapshots"        %in% tables)
  expect_true("column_snapshots" %in% tables)
  unlink(db)
})

test_that("init_snapshot_db() is idempotent (calling twice does not error)", {
  db <- tempfile(fileext = ".sqlite")
  expect_no_error(init_snapshot_db(db))
  expect_no_error(init_snapshot_db(db))
  unlink(db)
})

test_that("init_snapshot_db() creates parent directory if missing", {
  tmp <- file.path(tempdir(), "dqtest_newsubdir", "snap.sqlite")
  init_snapshot_db(tmp)
  expect_true(file.exists(tmp))
  unlink(dirname(tmp), recursive = TRUE)
})

# -- write_snapshot() ----------------------------------------------------------

test_that("write_snapshot() returns a positive integer snapshot_id", {
  db  <- tempfile(fileext = ".sqlite")
  sid <- write_snapshot(db, "test_ds", "file.csv",
                        make_df(), make_results(), list(), list(),
                        base_config())
  expect_true(is.numeric(sid) || is.integer(sid))
  expect_true(sid >= 1)
  unlink(db)
})

test_that("write_snapshot() inserts one row into snapshots table", {
  db <- tempfile(fileext = ".sqlite")
  write_snapshot(db, "test_ds", "file.csv",
                 make_df(), make_results(), list(), list(),
                 base_config())
  con <- DBI::dbConnect(RSQLite::SQLite(), db)
  n   <- DBI::dbGetQuery(con, "SELECT COUNT(*) AS n FROM snapshots")$n
  DBI::dbDisconnect(con)
  expect_equal(n, 1L)
  unlink(db)
})

test_that("write_snapshot() inserts column_snapshots rows", {
  db <- tempfile(fileext = ".sqlite")
  write_snapshot(db, "test_ds", "file.csv",
                 make_df(), make_results(), list(), list(),
                 base_config())
  con <- DBI::dbConnect(RSQLite::SQLite(), db)
  n   <- DBI::dbGetQuery(con, "SELECT COUNT(*) AS n FROM column_snapshots")$n
  DBI::dbDisconnect(con)
  expect_true(n > 0L)
  unlink(db)
})

test_that("write_snapshot() emits warning but does not stop on write error", {
  fake_dir <- tempfile()
  file.create(fake_dir)
  on.exit(unlink(fake_dir))
  impossible <- file.path(fake_dir, "impossible.sqlite")
  expect_warning(
    write_snapshot(impossible, "ds", "f.csv",
                   make_df(), make_results(), list(), list(),
                   base_config()),
    regexp = "SQLite"
  )
})

# -- read_recent_snapshots() ---------------------------------------------------

test_that("read_recent_snapshots() returns empty data frame when no runs exist", {
  db  <- tempfile(fileext = ".sqlite")
  init_snapshot_db(db)
  res <- read_recent_snapshots(db, "no_such_dataset")
  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0L)
  unlink(db)
})

test_that("read_recent_snapshots() returns empty data frame when db does not exist", {
  res <- read_recent_snapshots(tempfile(fileext = ".sqlite"), "ds")
  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0L)
})

test_that("read_recent_snapshots() returns only rows for the requested dataset", {
  db <- tempfile(fileext = ".sqlite")
  write_snapshot(db, "ds_a", "a.csv", make_df(),
                 make_results(), list(), list(), base_config())
  write_snapshot(db, "ds_b", "b.csv", make_df(),
                 make_results(), list(), list(), base_config())
  res <- read_recent_snapshots(db, "ds_a")
  expect_equal(nrow(res), 1L)
  expect_true(all(res$dataset_name == "ds_a"))
  unlink(db)
})

test_that("read_recent_snapshots() returns at most n rows", {
  db <- tempfile(fileext = ".sqlite")
  for (i in 1:5) {
    write_snapshot(db, "ds_a", paste0("f", i, ".csv"), make_df(),
                   make_results(), list(), list(), base_config())
  }
  res <- read_recent_snapshots(db, "ds_a", n = 3)
  expect_equal(nrow(res), 3L)
  unlink(db)
})

# -- compute_col_stats() column contract ---------------------------------------

test_that("compute_col_stats() returns a data frame with the correct 5 columns in order", {
  df  <- data.frame(
    id    = c("1", "2", "3"),
    score = c("10.5", "20.0", "bad"),
    stringsAsFactors = FALSE
  )
  cfg <- list(rules = list(
    max_missing_rate    = 0.05,
    max_non_numeric_rate = 0.01,
    type_inference_threshold = 0.90
  ))
  qc <- run_qc_checks(df, cfg)
  cs <- compute_col_stats(df, cfg, qc)

  expect_s3_class(cs, "data.frame")
  expect_equal(names(cs),
               c("column_name", "dq_check", "value", "threshold", "severity_on_breach"))
})

test_that("compute_col_stats() column names map correctly to report display names", {
  df  <- data.frame(x = c("1", "2"), stringsAsFactors = FALSE)
  cfg <- list(rules = list(max_missing_rate = 0.05,
                           max_non_numeric_rate = 0.01,
                           type_inference_threshold = 0.90))
  qc  <- run_qc_checks(df, cfg)
  cs  <- compute_col_stats(df, cfg, qc)
  names(cs) <- c("Column", "Stat", "Value", "Threshold", "Severity")
  expect_equal(names(cs), c("Column", "Stat", "Value", "Threshold", "Severity"))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)

## ----one-call-----------------------------------------------------------------
# run_dq_check("customer_accounts", config_dir = "path/to/configs")

## ----install------------------------------------------------------------------
# devtools::install("path/to/dqr/dqcheckr")
# library(dqcheckr)

## ----type-inference-----------------------------------------------------------
# infer_col_type(c("2024-01-15", "2024-06-30"))  # "date"
# infer_col_type(c("100", "200", "N/A"))          # "character" (only 67% numeric)
# infer_col_type(c("100", "200", "N/A", rep("50", 17)))  # "numeric" (95% numeric)
# infer_col_type(c(NA, "", NA))                   # "unknown"
# # with a custom threshold:
# infer_col_type(c(rep("1", 17), "a", "b", "c"), threshold = 0.80)  # "numeric" (85%)

## ----run-check----------------------------------------------------------------
# result <- run_dq_check(
#   dataset_name = "customer_accounts",
#   config_dir   = "path/to/configs",   # contains dqcheckr.yml + customer_accounts.yml
#   open_report  = TRUE                 # open HTML in browser when interactive
# )
# 
# # Console output (always printed):
# # [dqcheckr] customer_accounts: FAIL - 1 warning(s), 2 failure(s). Report: reports/...html
# 
# result$status       # "PASS", "WARN", or "FAIL"
# result$report_path  # full path to the HTML file
# result$snapshot_id  # integer row ID in snapshots table

## ----individual-checks--------------------------------------------------------
# library(dqcheckr)
# 
# # Build a minimal config programmatically
# cfg <- list(
#   rules = list(
#     max_missing_rate     = 0.05,
#     max_non_numeric_rate = 0.01,
#     min_row_count        = 0
#   ),
#   column_rules     = list(
#     status = list(allowed_values = c("ACTIVE", "CLOSED"))
#   ),
#   key_columns      = "id",
#   expected_columns = NULL
# )
# 
# df <- read.csv("data/customer_accounts.csv", colClasses = "character")
# df[] <- lapply(df, trimws)
# 
# # Run a single check
# results <- check_missing_rate(df, cfg)
# 
# # Inspect
# results[[1]]$status    # "PASS" or "FAIL"
# results[[1]]$observed  # e.g. "3.2% missing (4 of 125)"
# 
# # Run all single-snapshot checks at once
# all_results <- run_qc_checks(df, cfg)
# 
# # Overall status
# overall_status(all_results)  # "FAIL" > "WARN" > "PASS" > "INFO"

## ----custom-file--------------------------------------------------------------
# # File: custom/customer_accounts_checks.R
# 
# custom_checks <- function(df) {
#   results <- list()
# 
#   # Rule: ACTIVE accounts must not have a zero balance
#   active_zero <- df[df$account_status == "ACTIVE" &
#                     !is.na(df$account_balance) &
#                     df$account_balance == "0", ]
#   n <- nrow(active_zero)
#   results <- c(results, list(dq_result(
#     check_id   = "CUST-01",
#     check_name = "No zero-balance active accounts",
#     column     = "account_balance",   # enables per-column storage in SQLite
#     status     = if (n > 0) "FAIL" else "PASS",
#     observed   = sprintf("%d ACTIVE account(s) with balance 0", n),
#     message    = if (n > 0)
#       sprintf("%d ACTIVE account(s) have a zero balance.", n)
#     else
#       "No ACTIVE accounts have a zero balance."
#   )))
# 
#   results
# }

## ----snapshot-query-----------------------------------------------------------
# library(DBI)
# library(RSQLite)
# 
# con <- dbConnect(SQLite(), "data/snapshots.sqlite")
# 
# # Recent runs for one dataset
# dbGetQuery(con,
#   "SELECT id, file_name, overall_status, check_fail_count, run_timestamp
#    FROM snapshots
#    WHERE dataset_name = 'customer_accounts'
#    ORDER BY id DESC
#    LIMIT 10")
# 
# # Column-level stats for the most recent run
# dbGetQuery(con,
#   "SELECT column_name, dq_check, value, threshold
#    FROM column_snapshots
#    WHERE snapshot_id = (
#      SELECT MAX(id) FROM snapshots WHERE dataset_name = 'customer_accounts'
#    )
#    ORDER BY column_name, dq_check")
# 
# dbDisconnect(con)

## ----cross-dataset------------------------------------------------------------
# dbGetQuery(con,
#   "SELECT dataset_name, COUNT(*) AS runs,
#           SUM(check_fail_count) AS total_failures
#    FROM snapshots
#    GROUP BY dataset_name")

## ----starwars-setup-----------------------------------------------------------
# # Config directory contains:
# #   dqcheckr.yml         — global thresholds
# #   starwars_csv.yml     — CSV dataset config
# #   starwars_fwf.yml     — FWF dataset config
# 
# # Run checks on both formats
# result_csv <- run_dq_check("starwars_csv", config_dir = "config", open_report = TRUE)
# result_fwf <- run_dq_check("starwars_fwf", config_dir = "config", open_report = TRUE)

#' Run a full data quality check pipeline
#'
#' Orchestrates the complete dqcheckr pipeline: loads configuration, detects
#' files, runs QC and comparison checks, writes a snapshot to SQLite, and
#' renders an HTML report.
#'
#' @param dataset_name Character. Name of the dataset; must match a YAML config
#'   file \code{<dataset_name>.yml} in \code{config_dir}.
#' @param config_dir Character. Path to the directory containing
#'   \code{dqcheckr.yml} and the dataset YAML file. Defaults to \code{"."}.
#' @param open_report Logical. Whether to open the HTML report in the browser
#'   after rendering (only takes effect in interactive sessions).
#'
#' @return Invisibly, a named list with:
#'   \describe{
#'     \item{status}{Overall status string: \code{"PASS"}, \code{"WARN"},
#'       \code{"FAIL"}, or \code{"INFO"}.}
#'     \item{report_path}{Absolute path to the rendered HTML report.}
#'     \item{snapshot_id}{Integer row ID of the snapshot written to SQLite,
#'       or \code{NULL} if the write failed.}
#'   }
#'
#' @examples
#' \donttest{
#' tmp <- gsub("\\\\", "/", tempdir())
#' dat <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
#' writeLines(c(
#'   paste0('snapshot_db: "',       tmp, '/snap.sqlite"'),
#'   paste0('report_output_dir: "', tmp, '"'),
#'   'default_rules:',
#'   '  max_missing_rate: 0.60',
#'   '  min_row_count: 80'
#' ), file.path(tmp, "dqcheckr.yml"))
#' writeLines(c(
#'   'dataset_name: "starwars_csv"',
#'   paste0('current_file: "', dat, '"'),
#'   'format: csv',
#'   'encoding: "UTF-8"',
#'   'delimiter: ","'
#' ), file.path(tmp, "starwars_csv.yml"))
#' result <- run_dq_check("starwars_csv", config_dir = tmp, open_report = FALSE)
#' result$status
#' }
#'
#' @export
run_dq_check <- function(dataset_name,
                         config_dir   = ".",
                         open_report  = TRUE) {
  config <- load_config(dataset_name, config_dir)

  files   <- detect_files(config)
  df_curr <- read_dataset(files$current, config)
  df_prev <- if (!is.null(files$previous))
    read_dataset(files$previous, config)
  else
    NULL

  qc_results     <- run_qc_checks(df_curr, config)
  cp_results     <- if (!is.null(df_prev))
    run_comparison_checks(df_curr, df_prev, config)
  else
    list()
  custom_results <- run_custom_checks(df_curr, config)

  col_stats <- compute_col_stats(df_curr, config, qc_results)

  db_path     <- normalizePath(config$snapshot_db %||% "data/snapshots.sqlite",
                               mustWork = FALSE)
  snapshot_id <- write_snapshot(
    db_path, dataset_name,
    basename(files$current),
    df_curr, qc_results, cp_results, custom_results, config,
    col_stats = col_stats
  )

  snapshot_history <- read_recent_snapshots(db_path, dataset_name, n = 10)

  report_path <- render_report(
    dataset_name     = dataset_name,
    file_name        = basename(files$current),
    file_path        = files$current,
    df               = df_curr,
    qc_results       = qc_results,
    cp_results       = cp_results,
    custom_results   = custom_results,
    snapshot_history = snapshot_history,
    config           = config,
    col_stats        = col_stats,
    output_dir       = config$report_output_dir %||% "reports/",
    open_report      = open_report
  )

  status <- overall_status(c(qc_results, cp_results, custom_results))
  all_r  <- c(qc_results, cp_results, custom_results)
  n_warn <- sum(vapply(all_r, \(r) r$status == "WARN", logical(1)))
  n_fail <- sum(vapply(all_r, \(r) r$status == "FAIL", logical(1)))

  message(sprintf("[dqcheckr] %s: %s - %d warning(s), %d failure(s). Report: %s",
                  dataset_name, status, n_warn, n_fail, report_path))

  invisible(list(
    status      = status,
    report_path = report_path,
    snapshot_id = snapshot_id
  ))
}

library(testthat)
library(dqcheckr)

# End-to-end: run_dq_check() against fixture files in a temp config directory.
# This test wires up real YAML configs and calls the full pipeline.

setup_integration_env <- function() {
  tmp <- tempdir()
  cfg <- file.path(tmp, "dq_integration")
  dir.create(cfg, showWarnings = FALSE)

  writeLines(c(
    "snapshot_db: 'data/snapshots.sqlite'",
    "report_output_dir: 'reports/'",
    "default_rules:",
    "  max_missing_rate: 0.05",
    "  max_non_numeric_rate: 0.01",
    "  min_row_count: 0",
    "  max_row_count_change_pct: 0.10",
    "  max_numeric_mean_shift_pct: 0.20",
    "  max_missing_rate_change_pp: 2.0",
    "  max_non_numeric_rate_change_pp: 1.0",
    "  flag_new_columns: true",
    "  flag_dropped_columns: true",
    "  flag_type_changes: true",
    "  flag_column_order_change: true"
  ), file.path(cfg, "dqcheckr.yml"))

  fix_curr <- testthat::test_path("fixtures", "valid_accounts_current.csv")
  fix_prev <- testthat::test_path("fixtures", "valid_accounts_previous.csv")

  writeLines(c(
    "dataset_name: 'integ_ds'",
    sprintf("current_file: '%s'", fix_curr),
    sprintf("previous_file: '%s'", fix_prev),
    "format: csv",
    "encoding: UTF-8",
    sprintf("snapshot_db: '%s'", file.path(cfg, "snapshots.sqlite")),
    sprintf("report_output_dir: '%s'", file.path(cfg, "reports"))
  ), file.path(cfg, "integ_ds.yml"))

  cfg
}

test_that("run_dq_check() returns a list with status, report_path, snapshot_id", {
  cfg_dir <- setup_integration_env()
  result  <- run_dq_check("integ_ds", config_dir = cfg_dir, open_report = FALSE)
  expect_type(result, "list")
  expect_named(result, c("status", "report_path", "snapshot_id"), ignore.order = FALSE)
  expect_true(result$status %in% c("PASS", "WARN", "FAIL", "INFO"))
})

test_that("run_dq_check() writes an HTML report file to disk", {
  cfg_dir <- setup_integration_env()
  result  <- run_dq_check("integ_ds", config_dir = cfg_dir, open_report = FALSE)
  expect_true(file.exists(result$report_path))
  expect_match(result$report_path, "\\.html$")
})

test_that("run_dq_check() writes a snapshot to the SQLite database", {
  cfg_dir <- setup_integration_env()
  result  <- run_dq_check("integ_ds", config_dir = cfg_dir, open_report = FALSE)
  expect_false(is.null(result$snapshot_id))
  expect_true(result$snapshot_id >= 1)
})

test_that("run_dq_check() returns PASS status for the clean fixture pair", {
  cfg_dir <- setup_integration_env()
  result  <- run_dq_check("integ_ds", config_dir = cfg_dir, open_report = FALSE)
  expect_equal(result$status, "PASS")
})

custom_checks <- function(df) {
  results <- list()

  is_human <- if ("species" %in% names(df))
    !is.na(df$species) & tolower(df$species) == "human"
  else
    logical(nrow(df))

  # CC-01: Human height error — above 210
  if ("height" %in% names(df) && any(is_human)) {
    heights   <- suppressWarnings(as.numeric(df$height[is_human]))
    char_names <- if ("name" %in% names(df)) df$name[is_human] else rep(NA, sum(is_human))

    fail_names <- char_names[!is.na(heights) & heights > 210]
    n_fail     <- length(fail_names)

    results <- c(results, list(dq_result(
      check_id   = "CC-01",
      check_name = "Human height — above maximum",
      column     = "height",
      status     = if (n_fail > 0) "FAIL" else "PASS",
      observed   = if (n_fail > 0)
        sprintf("%d Human(s) with height > 210: %s",
                n_fail, paste(fail_names, collapse = ", "))
      else
        "No Human characters exceed height 210.",
      threshold  = "210",
      message    = if (n_fail > 0)
        sprintf("%d Human character(s) have height above 210.", n_fail)
      else
        "All Human characters are within the height maximum of 210."
    )))
  }

  # CC-02: Human height warning — elevated range 200-210
  if ("height" %in% names(df) && any(is_human)) {
    heights    <- suppressWarnings(as.numeric(df$height[is_human]))
    char_names <- if ("name" %in% names(df)) df$name[is_human] else rep(NA, sum(is_human))

    warn_names <- char_names[!is.na(heights) & heights > 199 & heights <= 210]
    n_warn     <- length(warn_names)

    results <- c(results, list(dq_result(
      check_id   = "CC-02",
      check_name = "Human height — elevated range",
      column     = "height",
      status     = if (n_warn > 0) "WARN" else "PASS",
      observed   = if (n_warn > 0)
        sprintf("%d Human(s) with height 200-210: %s",
                n_warn, paste(warn_names, collapse = ", "))
      else
        "No Human characters in the elevated height range (200-210).",
      threshold  = "200-210",
      message    = if (n_warn > 0)
        sprintf("%d Human character(s) have height between 200 and 210.", n_warn)
      else
        "No Human characters in the elevated height range."
    )))
  }

  # CC-03: Human mass warning — above 140
  if ("mass" %in% names(df) && any(is_human)) {
    masses     <- suppressWarnings(as.numeric(df$mass[is_human]))
    char_names <- if ("name" %in% names(df)) df$name[is_human] else rep(NA, sum(is_human))

    warn_names <- char_names[!is.na(masses) & masses > 140]
    n_warn     <- length(warn_names)

    results <- c(results, list(dq_result(
      check_id   = "CC-03",
      check_name = "Human mass — elevated",
      column     = "mass",
      status     = if (n_warn > 0) "WARN" else "PASS",
      observed   = if (n_warn > 0)
        sprintf("%d Human(s) with mass > 140: %s",
                n_warn, paste(warn_names, collapse = ", "))
      else
        "No Human characters exceed mass 140.",
      threshold  = "140",
      message    = if (n_warn > 0)
        sprintf("%d Human character(s) have mass above 140.", n_warn)
      else
        "All Human characters are within the mass threshold of 140."
    )))
  }

  results
}

#' Detect current and previous dataset files
#'
#' Resolves the current and previous file paths from the configuration. If
#' \code{current_file} is set explicitly, it is used directly. Otherwise the
#' two most recently modified files in \code{folder} are used.
#'
#' @param config Named list. Merged configuration as returned by
#'   \code{\link{load_config}}.
#'
#' @return A named list with elements \code{current} (character path) and
#'   \code{previous} (character path or \code{NULL}).
#'
#' @examples
#' cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg <- load_config("starwars_csv", config_dir = cfg_dir)
#' cfg$current_file <- system.file("demonstrations/data/starwars.csv",
#'                                  package = "dqcheckr")
#' files <- detect_files(cfg)
#' files$current
#'
#' @export
detect_files <- function(config) {
  if (!is.null(config$current_file)) {
    if (!file.exists(config$current_file)) {
      rlang::abort(paste0("current_file not found: ", config$current_file))
    }
    previous <- NULL
    if (!is.null(config$previous_file)) {
      if (!file.exists(config$previous_file)) {
        rlang::abort(paste0("previous_file not found: ", config$previous_file))
      }
      previous <- config$previous_file
    }
    return(list(current = config$current_file, previous = previous))
  }

  folder <- config$folder
  if (is.null(folder) || !dir.exists(folder)) {
    rlang::abort(paste0("Folder not found: ", folder %||% "(NULL)"))
  }

  files <- list.files(folder, full.names = TRUE)
  if (length(files) == 0) {
    rlang::abort(paste0("No files found in folder: ", folder))
  }

  files <- files[order(file.mtime(files), decreasing = TRUE)]
  list(
    current  = files[1],
    previous = if (length(files) >= 2) files[2] else NULL
  )
}

#' Read a dataset file into a data frame
#'
#' Reads a CSV or fixed-width file, coercing all columns to character and
#' trimming whitespace. Encoding and delimiter are taken from \code{config}.
#'
#' @param path Character. Path to the file to read.
#' @param config Named list. Merged configuration as returned by
#'   \code{\link{load_config}}. Must include \code{format} (\code{"csv"} or
#'   \code{"fwf"}). For FWF files, \code{fwf_widths} is required and
#'   \code{fwf_col_names} and \code{fwf_skip} are optional.
#'
#' @return A data frame with all columns as character vectors.
#'
#' @examples
#' cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg  <- load_config("starwars_csv", config_dir = cfg_dir)
#' path <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
#' df   <- read_dataset(path, cfg)
#'
#' @export
read_dataset <- function(path, config) {
  fmt <- tolower(config$format %||% "csv")
  enc <- config$encoding %||% "UTF-8"

  if (fmt == "csv") {
    delim <- config$delimiter %||% ","
    df <- tryCatch(
      readr::read_delim(
        path,
        delim      = delim,
        col_types  = readr::cols(.default = "c"),
        locale     = readr::locale(encoding = enc),
        show_col_types = FALSE
      ),
      error = function(e) rlang::abort(paste0("Failed to parse file '", path, "': ", conditionMessage(e)))
    )
  } else if (fmt == "fwf") {
    if (is.null(config$fwf_widths)) {
      rlang::abort("fwf_widths must be set in config for fixed-width files")
    }
    df <- tryCatch(
      readr::read_fwf(
        path,
        col_positions = readr::fwf_widths(
          config$fwf_widths,
          col_names = config$fwf_col_names
        ),
        col_types  = readr::cols(.default = "c"),
        locale     = readr::locale(encoding = enc),
        skip       = config$fwf_skip %||% 0L,
        show_col_types = FALSE
      ),
      error = function(e) rlang::abort(paste0("Failed to parse file '", path, "': ", conditionMessage(e)))
    )
  } else {
    rlang::abort(paste0("Unsupported format: '", config$format, "'. Must be 'csv' or 'fwf'."))
  }

  df <- as.data.frame(df, stringsAsFactors = FALSE)
  for (col in names(df)) {
    df[[col]] <- trimws(df[[col]])
  }
  df
}

library(testthat)
library(dqcheckr)

fix <- function(f) testthat::test_path("fixtures", f)

# -- read_dataset() ------------------------------------------------------------

test_that("read_dataset() reads CSV and returns all-character data frame", {
  cfg <- list(format = "csv", encoding = "UTF-8", delimiter = ",")
  df  <- read_dataset(fix("valid_accounts_current.csv"), cfg)
  expect_s3_class(df, "data.frame")
  expect_true(all(vapply(df, is.character, logical(1))))
})

test_that("read_dataset() has the expected columns", {
  cfg <- list(format = "csv", encoding = "UTF-8", delimiter = ",")
  df  <- read_dataset(fix("valid_accounts_current.csv"), cfg)
  expect_equal(ncol(df), 6)
  expect_equal(names(df),
               c("id", "name", "country_code", "account_status",
                 "account_balance", "created_date"))
})

test_that("read_dataset() trims leading/trailing whitespace from all cells", {
  tmp <- tempfile(fileext = ".csv")
  writeLines(c("a,b", "  hello  ,  world  "), tmp)
  cfg <- list(format = "csv", encoding = "UTF-8", delimiter = ",")
  df  <- read_dataset(tmp, cfg)
  expect_equal(df$a[1], "hello")
  expect_equal(df$b[1], "world")
  unlink(tmp)
})

test_that("read_dataset() preserves NA as NA after trim", {
  tmp <- tempfile(fileext = ".csv")
  writeLines(c("a,b", "1,", "2,2"), tmp)
  cfg <- list(format = "csv", encoding = "UTF-8", delimiter = ",")
  df  <- read_dataset(tmp, cfg)
  expect_true(is.na(df$b[1]) || df$b[1] == "")
})

test_that("read_dataset() stops on nonexistent file", {
  cfg <- list(format = "csv", encoding = "UTF-8", delimiter = ",")
  expect_error(read_dataset("/nonexistent/path/file.csv", cfg))
})

test_that("read_dataset() stops on unknown format", {
  cfg <- list(format = "parquet")
  expect_error(
    read_dataset(fix("valid_accounts_current.csv"), cfg),
    regexp = "format"
  )
})

test_that("read_dataset() reads FWF file correctly", {
  cfg <- list(
    format       = "fwf",
    encoding     = "UTF-8",
    fwf_widths   = c(6, 16, 3, 11, 13, 10),
    fwf_col_names = c("id", "name", "country_code",
                      "account_status", "account_balance", "created_date")
  )
  df <- read_dataset(fix("accounts_fwf_current.txt"), cfg)
  expect_s3_class(df, "data.frame")
  expect_equal(ncol(df), 6)
  expect_equal(nrow(df), 10)
})

# -- detect_files() ------------------------------------------------------------

test_that("detect_files() uses current_file when set; previous NULL when absent (D-01)", {
  cfg <- list(current_file = fix("valid_accounts_current.csv"))
  result <- detect_files(cfg)
  expect_equal(result$current, fix("valid_accounts_current.csv"))
  expect_null(result$previous)
})

test_that("detect_files() uses both files when both explicit paths set", {
  cfg <- list(
    current_file  = fix("valid_accounts_current.csv"),
    previous_file = fix("valid_accounts_previous.csv")
  )
  result <- detect_files(cfg)
  expect_equal(result$current,  fix("valid_accounts_current.csv"))
  expect_equal(result$previous, fix("valid_accounts_previous.csv"))
})

test_that("detect_files() stops when current_file does not exist", {
  cfg <- list(current_file = "/no/such/file.csv")
  expect_error(detect_files(cfg), regexp = "current_file")
})

test_that("detect_files() stops when folder does not exist", {
  cfg <- list(folder = "/no/such/folder/")
  expect_error(detect_files(cfg), regexp = "Folder")
})

test_that("detect_files() stops when folder is empty", {
  tmp   <- tempdir()
  empty <- file.path(tmp, "dqtest_emptydir")
  dir.create(empty, showWarnings = FALSE)
  for (f in list.files(empty, full.names = TRUE)) file.remove(f)
  cfg <- list(folder = empty)
  expect_error(detect_files(cfg), regexp = "No files")
})

test_that("detect_files() returns previous=NULL for single-file folder", {
  tmp     <- tempdir()
  onefile <- file.path(tmp, "dqtest_onefile")
  dir.create(onefile, showWarnings = FALSE)
  f <- file.path(onefile, "data.csv")
  writeLines("a,b\n1,2", f)
  cfg    <- list(folder = onefile)
  result <- detect_files(cfg)
  expect_equal(result$current, f)
  expect_null(result$previous)
  unlink(f)
})

#' Test for missing or empty values
#' @keywords internal
.missing_vals <- function(x) is.na(x) | x == ""

#' QC-01: Check missing rate per column
#'
#' Returns a \code{\link{dq_result}} per column flagging columns whose
#' proportion of missing or empty values exceeds \code{max_missing_rate}.
#'
#' @param df A data frame with all columns as character vectors.
#' @param config Named list as returned by \code{\link{load_config}}.
#'
#' @return A list of \code{\link{dq_result}} objects, one per column.
#'
#' @examples
#' cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg  <- load_config("starwars_csv", config_dir = cfg_dir)
#' path <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
#' df   <- read_dataset(path, cfg)
#' check_missing_rate(df, cfg)
#'
#' @export
check_missing_rate <- function(df, config) {
  threshold <- config$rules$max_missing_rate %||% 0.05
  lapply(names(df), function(col) {
    missing_count <- sum(.missing_vals(df[[col]]))
    missing_rate  <- missing_count / nrow(df)
    status <- if (missing_rate > threshold) "FAIL" else "PASS"
    dq_result(
      check_id   = "QC-01",
      check_name = "Missing rate",
      column     = col,
      status     = status,
      observed   = sprintf("%.1f%% missing (%d of %d)",
                           missing_rate * 100, missing_count, nrow(df)),
      threshold  = sprintf("%.1f%%", threshold * 100),
      message    = if (status == "FAIL")
        sprintf("Column '%s' missing rate %.1f%% exceeds threshold %.1f%%.",
                col, missing_rate * 100, threshold * 100)
      else
        sprintf("Column '%s' missing rate is within threshold.", col)
    )
  })
}

#' QC-02: Check for entirely empty columns
#' @keywords internal
check_empty_column <- function(df, config) {
  lapply(names(df), function(col) {
    is_empty <- all(.missing_vals(df[[col]]))
    dq_result(
      check_id   = "QC-02",
      check_name = "Empty column",
      column     = col,
      status     = if (is_empty) "FAIL" else "PASS",
      observed   = if (is_empty) "100% empty" else "Not empty",
      message    = if (is_empty)
        sprintf("Column '%s' is entirely empty.", col)
      else
        sprintf("Column '%s' has at least one non-empty value.", col)
    )
  })
}

#' QC-03: Check for fully-duplicate rows
#' @keywords internal
check_duplicate_rows <- function(df, config) {
  n_dups <- sum(duplicated(df))
  status <- if (n_dups > 0) "WARN" else "PASS"
  list(dq_result(
    check_id   = "QC-03",
    check_name = "Duplicate rows",
    status     = status,
    observed   = sprintf("%d fully-duplicate row(s)", n_dups),
    message    = if (n_dups > 0)
      sprintf("%d row(s) are exact duplicates of another row.", n_dups)
    else
      "No fully-duplicate rows found."
  ))
}

#' QC-04: Report row count
#' @keywords internal
check_row_count <- function(df, config) {
  list(dq_result(
    check_id   = "QC-04",
    check_name = "Row count",
    status     = "INFO",
    observed   = as.character(nrow(df)),
    message    = sprintf("File contains %d rows.", nrow(df))
  ))
}

#' QC-05: Report column count
#' @keywords internal
check_col_count <- function(df, config) {
  list(dq_result(
    check_id   = "QC-05",
    check_name = "Column count",
    status     = "INFO",
    observed   = as.character(ncol(df)),
    message    = sprintf("File contains %d columns.", ncol(df))
  ))
}

#' QC-06: Report inferred column types
#' @keywords internal
check_inferred_types <- function(df, config) {
  threshold <- config$rules$type_inference_threshold %||% 0.90
  lapply(names(df), function(col) {
    typ <- infer_col_type(df[[col]], threshold)
    dq_result(
      check_id   = "QC-06",
      check_name = "Inferred type",
      column     = col,
      status     = "INFO",
      observed   = typ,
      message    = sprintf("Column '%s' inferred as %s.", col, typ)
    )
  })
}

#' QC-07: Report numeric summary statistics
#' @keywords internal
#' @importFrom stats sd
check_numeric_stats <- function(df, config) {
  threshold <- config$rules$type_inference_threshold %||% 0.90
  results <- list()
  for (col in names(df)) {
    if (infer_col_type(df[[col]], threshold) != "numeric") next
    vals <- suppressWarnings(as.numeric(df[[col]]))
    vals <- vals[!is.na(vals)]
    if (length(vals) == 0) next
    results <- c(results, list(dq_result(
      check_id   = "QC-07",
      check_name = "Numeric stats",
      column     = col,
      status     = "INFO",
      observed   = sprintf("min=%.4g, max=%.4g, mean=%.4g, sd=%.4g",
                           min(vals), max(vals), mean(vals),
                           if (length(vals) > 1) sd(vals) else NA_real_),
      message    = sprintf("Summary statistics for numeric column '%s'.", col)
    )))
  }
  results
}

#' QC-08: Report distinct value counts for character columns
#' @keywords internal
check_distinct_counts <- function(df, config) {
  threshold <- config$rules$type_inference_threshold %||% 0.90
  results <- list()
  for (col in names(df)) {
    if (infer_col_type(df[[col]], threshold) != "character") next
    n_distinct <- length(unique(df[[col]][!.missing_vals(df[[col]])]))
    results <- c(results, list(dq_result(
      check_id   = "QC-08",
      check_name = "Distinct value count",
      column     = col,
      status     = "INFO",
      observed   = as.character(n_distinct),
      message    = sprintf("Column '%s' has %d distinct non-empty value(s).",
                           col, n_distinct)
    )))
  }
  results
}

#' QC-09: Check for values outside the allowed set
#' @keywords internal
check_allowed_values <- function(df, config) {
  results <- list()
  col_rules <- config$column_rules %||% list()
  for (col in names(col_rules)) {
    allowed <- col_rules[[col]]$allowed_values
    if (is.null(allowed) || !col %in% names(df)) next
    vals <- df[[col]][!.missing_vals(df[[col]])]
    bad  <- setdiff(unique(vals), allowed)
    status <- if (length(bad) > 0) "FAIL" else "PASS"
    results <- c(results, list(dq_result(
      check_id   = "QC-09",
      check_name = "Allowed values",
      column     = col,
      status     = status,
      observed   = if (length(bad) > 0)
        paste("Unexpected values:", paste(bad, collapse = ", "))
      else
        "All values are in the allowed list.",
      threshold  = paste("Allowed:", paste(allowed, collapse = ", ")),
      message    = if (length(bad) > 0)
        sprintf("Column '%s' contains %d unexpected value(s): %s.",
                col, length(bad), paste(bad, collapse = ", "))
      else
        sprintf("Column '%s' contains only allowed values.", col)
    )))
  }
  results
}

#' QC-10: Check for out-of-range numeric values
#' @keywords internal
#' @importFrom utils head
check_numeric_bounds <- function(df, config) {
  results <- list()
  col_rules <- config$column_rules %||% list()
  for (col in names(col_rules)) {
    min_val <- col_rules[[col]]$min_value
    max_val <- col_rules[[col]]$max_value
    if (is.null(min_val) && is.null(max_val)) next
    if (!col %in% names(df)) next
    vals    <- suppressWarnings(as.numeric(df[[col]]))
    violate <- character(0)
    if (!is.null(min_val)) {
      below <- df[[col]][!is.na(vals) & vals < min_val]
      violate <- c(violate, unique(below))
    }
    if (!is.null(max_val)) {
      above <- df[[col]][!is.na(vals) & vals > max_val]
      violate <- c(violate, unique(above))
    }
    violate <- unique(violate)
    status  <- if (length(violate) > 0) "FAIL" else "PASS"
    thr_parts <- c(
      if (!is.null(min_val)) paste("min:", min_val),
      if (!is.null(max_val)) paste("max:", max_val)
    )
    results <- c(results, list(dq_result(
      check_id   = "QC-10",
      check_name = "Numeric bounds",
      column     = col,
      status     = status,
      observed   = if (length(violate) > 0)
        paste("Out-of-range values:", paste(head(violate, 5), collapse = ", "))
      else
        "All values are within bounds.",
      threshold  = paste(thr_parts, collapse = "; "),
      message    = if (length(violate) > 0)
        sprintf("Column '%s' has %d out-of-range value(s).", col, length(violate))
      else
        sprintf("Column '%s' values are all within bounds.", col)
    )))
  }
  results
}

#' QC-11: Check non-numeric rate in numeric columns
#' @keywords internal
check_non_numeric <- function(df, config) {
  threshold       <- config$rules$max_non_numeric_rate %||% 0.01
  type_threshold  <- config$rules$type_inference_threshold %||% 0.90
  results   <- list()
  for (col in names(df)) {
    if (infer_col_type(df[[col]], type_threshold) != "numeric") next
    non_empty <- df[[col]][!.missing_vals(df[[col]])]
    if (length(non_empty) == 0) next
    bad      <- non_empty[is.na(suppressWarnings(as.numeric(non_empty)))]
    rate     <- length(bad) / length(non_empty)
    status   <- if (rate > threshold) "FAIL" else if (length(bad) > 0) "WARN" else "PASS"
    results <- c(results, list(dq_result(
      check_id   = "QC-11",
      check_name = "Non-numeric values",
      column     = col,
      status     = status,
      observed   = sprintf("%d non-numeric value(s) (%.2f%%)",
                           length(bad), rate * 100),
      threshold  = sprintf("%.2f%%", threshold * 100),
      message    = switch(status,
        FAIL = sprintf("Column '%s' non-numeric rate %.2f%% exceeds threshold %.2f%%.",
                       col, rate * 100, threshold * 100),
        WARN = sprintf("Column '%s' has %d non-numeric value(s) below threshold.",
                       col, length(bad)),
        PASS = sprintf("Column '%s' has no non-numeric values.", col)
      )
    )))
  }
  results
}

#' QC-12: Check uniqueness of key columns
#' @keywords internal
check_key_uniqueness <- function(df, config) {
  keys <- config$key_columns
  if (is.null(keys) || length(keys) == 0) return(list())
  lapply(keys, function(col) {
    if (!col %in% names(df)) {
      return(dq_result(
        check_id   = "QC-12",
        check_name = "Key uniqueness",
        column     = col,
        status     = "FAIL",
        observed   = "Column not found in file",
        message    = sprintf("Key column '%s' is not present in the file.", col)
      ))
    }
    n_dups <- sum(duplicated(df[[col]]))
    status <- if (n_dups > 0) "FAIL" else "PASS"
    dq_result(
      check_id   = "QC-12",
      check_name = "Key uniqueness",
      column     = col,
      status     = status,
      observed   = sprintf("%d duplicate value(s) found", n_dups),
      message    = if (n_dups > 0)
        sprintf("Key column '%s' has %d duplicate value(s).", col, n_dups)
      else
        sprintf("Key column '%s' has all unique values.", col)
    )
  })
}

#' QC-13: Check values against a regex pattern
#' @keywords internal
check_pattern <- function(df, config) {
  results   <- list()
  col_rules <- config$column_rules %||% list()
  for (col in names(col_rules)) {
    pattern <- col_rules[[col]]$pattern
    if (is.null(pattern) || !col %in% names(df)) next
    non_empty  <- df[[col]][!.missing_vals(df[[col]])]
    bad_count  <- sum(!grepl(pattern, non_empty, perl = TRUE))
    status     <- if (bad_count > 0) "FAIL" else "PASS"
    results <- c(results, list(dq_result(
      check_id   = "QC-13",
      check_name = "Pattern / regex",
      column     = col,
      status     = status,
      observed   = sprintf("%d value(s) do not match pattern", bad_count),
      threshold  = pattern,
      message    = if (bad_count > 0)
        sprintf("Column '%s': %d value(s) violate pattern '%s'.",
                col, bad_count, pattern)
      else
        sprintf("Column '%s': all values match pattern '%s'.", col, pattern)
    )))
  }
  results
}

#' QC-14: Check minimum row count threshold
#' @keywords internal
check_min_row_count <- function(df, config) {
  min_rc <- config$rules$min_row_count %||% 0
  if (min_rc <= 0) {
    return(list(dq_result(
      check_id   = "QC-14",
      check_name = "Minimum row count",
      status     = "PASS",
      observed   = as.character(nrow(df)),
      message    = "Minimum row count check is disabled (min_row_count = 0)."
    )))
  }
  status <- if (nrow(df) < min_rc) "FAIL" else "PASS"
  list(dq_result(
    check_id   = "QC-14",
    check_name = "Minimum row count",
    status     = status,
    observed   = sprintf("%d rows", nrow(df)),
    threshold  = sprintf("%d rows minimum", min_rc),
    message    = if (status == "FAIL")
      sprintf("File has %d rows, below the minimum of %d.", nrow(df), min_rc)
    else
      sprintf("File has %d rows, meeting the minimum of %d.", nrow(df), min_rc)
  ))
}

#' SC-01/SC-02: Check columns against expected schema contract
#' @keywords internal
check_schema_contract <- function(df, config) {
  expected <- config$expected_columns
  if (is.null(expected)) return(list())

  results <- list()

  extra <- setdiff(names(df), expected)
  if (length(extra) > 0) {
    for (col in extra) {
      results <- c(results, list(dq_result(
        check_id   = "SC-01",
        check_name = "Unexpected column",
        column     = col,
        status     = "FAIL",
        observed   = sprintf("Column '%s' is not in the expected schema.", col),
        message    = sprintf("Column '%s' is present in the file but not in expected_columns.", col)
      )))
    }
  } else {
    results <- c(results, list(dq_result(
      check_id   = "SC-01",
      check_name = "Unexpected column",
      status     = "PASS",
      observed   = "No unexpected columns.",
      message    = "All file columns are in the expected schema."
    )))
  }

  missing_cols <- setdiff(expected, names(df))
  if (length(missing_cols) > 0) {
    for (col in missing_cols) {
      results <- c(results, list(dq_result(
        check_id   = "SC-02",
        check_name = "Missing expected column",
        column     = col,
        status     = "FAIL",
        observed   = sprintf("Expected column '%s' is absent.", col),
        message    = sprintf("Column '%s' is in expected_columns but absent from the file.", col)
      )))
    }
  } else {
    results <- c(results, list(dq_result(
      check_id   = "SC-02",
      check_name = "Missing expected column",
      status     = "PASS",
      observed   = "All expected columns are present.",
      message    = "No expected columns are missing from the file."
    )))
  }

  results
}

#' Run all generic quality checks on a dataset
#'
#' Runs the full QC check suite (QC-01 to QC-14, SC-01, SC-02) against a
#' single data frame snapshot.
#'
#' @param df A data frame with all columns as character vectors (as returned by
#'   \code{\link{read_dataset}}).
#' @param config Named list. Merged configuration as returned by
#'   \code{\link{load_config}}.
#'
#' @return A list of \code{\link{dq_result}} objects.
#'
#' @examples
#' cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
#' cfg     <- load_config("starwars_csv", config_dir = cfg_dir)
#' path    <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
#' df      <- read_dataset(path, cfg)
#' results <- run_qc_checks(df, cfg)
#'
#' @export
run_qc_checks <- function(df, config) {
  c(
    check_missing_rate(df, config),
    check_empty_column(df, config),
    check_duplicate_rows(df, config),
    check_row_count(df, config),
    check_col_count(df, config),
    check_inferred_types(df, config),
    check_numeric_stats(df, config),
    check_distinct_counts(df, config),
    check_allowed_values(df, config),
    check_numeric_bounds(df, config),
    check_non_numeric(df, config),
    check_key_uniqueness(df, config),
    check_pattern(df, config),
    check_min_row_count(df, config),
    check_schema_contract(df, config)
  )
}

library(testthat)
library(dqcheckr)

base_config <- function(overrides = list()) {
  cfg <- list(
    format = "csv",
    rules = list(
      max_missing_rate               = 0.05,
      max_non_numeric_rate           = 0.01,
      min_row_count                  = 0,
      max_row_count_change_pct       = 0.10,
      max_numeric_mean_shift_pct     = 0.20,
      max_missing_rate_change_pp     = 2.0,
      max_non_numeric_rate_change_pp = 1.0
    ),
    column_rules     = list(),
    key_columns      = NULL,
    expected_columns = NULL
  )
  utils::modifyList(cfg, overrides)
}

make_df <- function() {
  data.frame(
    id              = c("ID001", "ID002", "ID003", "ID004", "ID005"),
    name            = c("Alice", "Bob", "Clara", "David", "Emma"),
    country_code    = c("GB", "US", "DE", "FR", "GB"),
    account_status  = c("ACTIVE", "ACTIVE", "CLOSED", "ACTIVE", "SUSPENDED"),
    account_balance = c("15000", "8500", "0", "250000", "1200"),
    created_date    = c("2024-01-15", "2024-02-20", "2023-11-01",
                        "2024-03-10", "2024-04-05"),
    stringsAsFactors = FALSE
  )
}

# -- QC-01 Missing rate --------------------------------------------------------

test_that("check_missing_rate() returns PASS when missing rate is within threshold", {
  res      <- check_missing_rate(make_df(), base_config())
  statuses <- vapply(res, `[[`, character(1), "status")
  expect_true(all(statuses == "PASS"))
})

test_that("check_missing_rate() returns FAIL when column exceeds threshold", {
  df       <- make_df()
  df$name[2:5] <- NA
  res      <- check_missing_rate(df, base_config())
  name_res <- Filter(\(r) r$column == "name", res)
  expect_equal(name_res[[1]]$status, "FAIL")
})

# -- QC-02 Empty column --------------------------------------------------------

test_that("check_empty_column() returns PASS when no column is entirely empty", {
  res <- check_empty_column(make_df(), base_config())
  expect_true(all(vapply(res, \(r) r$status == "PASS", logical(1))))
})

test_that("check_empty_column() returns FAIL for entirely empty column", {
  df                   <- make_df()
  df$account_balance   <- NA_character_
  res                  <- check_empty_column(df, base_config())
  bal                  <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "FAIL")
})

# -- QC-03 Duplicate rows ------------------------------------------------------

test_that("check_duplicate_rows() returns PASS when no duplicate rows", {
  res <- check_duplicate_rows(make_df(), base_config())
  expect_equal(res[[1]]$status, "PASS")
})

test_that("check_duplicate_rows() returns WARN when duplicate rows exist", {
  df  <- rbind(make_df(), make_df()[2, ])
  res <- check_duplicate_rows(df, base_config())
  expect_equal(res[[1]]$status, "WARN")
})

# -- QC-04 Row count -----------------------------------------------------------

test_that("check_row_count() returns INFO with correct count", {
  res <- check_row_count(make_df(), base_config())
  expect_equal(res[[1]]$status, "INFO")
  expect_equal(res[[1]]$observed, "5")
})

# -- QC-05 Column count --------------------------------------------------------

test_that("check_col_count() returns INFO with correct count", {
  df  <- make_df()
  res <- check_col_count(df, base_config())
  expect_equal(res[[1]]$status, "INFO")
  expect_equal(res[[1]]$observed, as.character(ncol(df)))
})

# -- QC-06 Inferred types ------------------------------------------------------

test_that("check_inferred_types() returns INFO for each column", {
  df  <- make_df()
  res <- check_inferred_types(df, base_config())
  expect_equal(length(res), ncol(df))
  expect_true(all(vapply(res, \(r) r$status == "INFO", logical(1))))
})

test_that("check_inferred_types() infers 'numeric' for account_balance", {
  res <- check_inferred_types(make_df(), base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$observed, "numeric")
})

test_that("check_inferred_types() infers 'date' for created_date", {
  res  <- check_inferred_types(make_df(), base_config())
  date <- Filter(\(r) r$column == "created_date", res)
  expect_equal(date[[1]]$observed, "date")
})

# -- QC-07 Numeric stats -------------------------------------------------------

test_that("check_numeric_stats() returns INFO for numeric columns", {
  res <- check_numeric_stats(make_df(), base_config())
  expect_true(length(res) > 0)
  expect_true(all(vapply(res, \(r) r$status == "INFO", logical(1))))
})

test_that("check_numeric_stats() observed contains min/max/mean/sd", {
  res <- check_numeric_stats(make_df(), base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_match(bal[[1]]$observed, "min=")
  expect_match(bal[[1]]$observed, "max=")
  expect_match(bal[[1]]$observed, "mean=")
})

# -- QC-08 Distinct counts -----------------------------------------------------

test_that("check_distinct_counts() returns INFO for character columns", {
  res <- check_distinct_counts(make_df(), base_config())
  expect_true(length(res) > 0)
  expect_true(all(vapply(res, \(r) r$status == "INFO", logical(1))))
})

# -- QC-09 Allowed values ------------------------------------------------------

test_that("check_allowed_values() returns PASS when all values are allowed", {
  cfg <- base_config(list(column_rules = list(
    country_code = list(allowed_values = c("GB", "US", "DE", "FR"))
  )))
  res <- check_allowed_values(make_df(), cfg)
  cc  <- Filter(\(r) r$column == "country_code", res)
  expect_equal(cc[[1]]$status, "PASS")
})

test_that("check_allowed_values() returns FAIL when disallowed values present", {
  df               <- make_df()
  df$country_code[1] <- "ZZ"
  cfg <- base_config(list(column_rules = list(
    country_code = list(allowed_values = c("GB", "US", "DE", "FR"))
  )))
  res <- check_allowed_values(df, cfg)
  cc  <- Filter(\(r) r$column == "country_code", res)
  expect_equal(cc[[1]]$status, "FAIL")
  expect_match(cc[[1]]$observed, "ZZ")
})

test_that("check_allowed_values() returns empty list when no column_rules", {
  res <- check_allowed_values(make_df(), base_config())
  expect_equal(length(res), 0)
})

# -- QC-10 Numeric bounds ------------------------------------------------------

test_that("check_numeric_bounds() returns PASS when all values within range", {
  cfg <- base_config(list(column_rules = list(
    account_balance = list(min_value = 0, max_value = 1000000)
  )))
  res <- check_numeric_bounds(make_df(), cfg)
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "PASS")
})

test_that("check_numeric_bounds() returns FAIL when value below min", {
  df                    <- make_df()
  df$account_balance[1] <- "-500"
  cfg <- base_config(list(column_rules = list(
    account_balance = list(min_value = 0, max_value = 1000000)
  )))
  res <- check_numeric_bounds(df, cfg)
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "FAIL")
})

test_that("check_numeric_bounds() returns FAIL when value above max", {
  df                    <- make_df()
  df$account_balance[1] <- "2000000"
  cfg <- base_config(list(column_rules = list(
    account_balance = list(min_value = 0, max_value = 1000000)
  )))
  res <- check_numeric_bounds(df, cfg)
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "FAIL")
})

# -- QC-11 Non-numeric values --------------------------------------------------

test_that("check_non_numeric() returns PASS when all numeric column values are numeric", {
  res <- check_non_numeric(make_df(), base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "PASS")
})

test_that("check_non_numeric() returns FAIL when non-numeric rate exceeds threshold", {
  # 10 rows: 1 bad = 10% non-numeric. infer_col_type still returns 'numeric'
  # (90% numeric >= 90% threshold). 10% > 1% max_non_numeric_rate -> FAIL.
  df <- make_df()[rep(seq_len(nrow(make_df())), 2), ]
  df$account_balance[1] <- "N/A"
  res <- check_non_numeric(df, base_config())
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "FAIL")
})

test_that("check_non_numeric() returns WARN when non-numeric present but below threshold", {
  df100               <- make_df()[rep(1:5, 20), ]
  df100$account_balance[1] <- "N/A"
  cfg <- base_config(list(rules = list(
    max_missing_rate = 0.05, max_non_numeric_rate = 0.05,
    min_row_count = 0, max_row_count_change_pct = 0.10,
    max_numeric_mean_shift_pct = 0.20,
    max_missing_rate_change_pp = 2.0,
    max_non_numeric_rate_change_pp = 1.0
  )))
  res <- check_non_numeric(df100, cfg)
  bal <- Filter(\(r) r$column == "account_balance", res)
  expect_equal(bal[[1]]$status, "WARN")
})

# -- QC-12 Key uniqueness ------------------------------------------------------

test_that("check_key_uniqueness() returns PASS for unique key column", {
  cfg <- base_config(list(key_columns = "id"))
  res <- check_key_uniqueness(make_df(), cfg)
  expect_equal(res[[1]]$status, "PASS")
})

test_that("check_key_uniqueness() returns FAIL when duplicate key exists", {
  df       <- make_df()
  df$id[2] <- "ID001"
  cfg <- base_config(list(key_columns = "id"))
  res <- check_key_uniqueness(df, cfg)
  expect_equal(res[[1]]$status, "FAIL")
})

test_that("check_key_uniqueness() returns empty list when no key_columns configured", {
  res <- check_key_uniqueness(make_df(), base_config())
  expect_equal(length(res), 0)
})

# -- QC-13 Pattern -------------------------------------------------------------

test_that("check_pattern() returns PASS when all values match pattern", {
  cfg <- base_config(list(column_rules = list(
    country_code = list(pattern = "^[A-Z]{2}$")
  )))
  res <- check_pattern(make_df(), cfg)
  cc  <- Filter(\(r) r$column == "country_code", res)
  expect_equal(cc[[1]]$status, "PASS")
})

test_that("check_pattern() returns FAIL when values violate pattern", {
  df                   <- make_df()
  df$country_code[1]   <- "gb"
  cfg <- base_config(list(column_rules = list(
    country_code = list(pattern = "^[A-Z]{2}$")
  )))
  res <- check_pattern(df, cfg)
  cc  <- Filter(\(r) r$column == "country_code", res)
  expect_equal(cc[[1]]$status, "FAIL")
})

# -- QC-14 Minimum row count ---------------------------------------------------

test_that("check_min_row_count() returns PASS when disabled (min_row_count=0)", {
  res <- check_min_row_count(make_df(), base_config())
  expect_equal(res[[1]]$status, "PASS")
})

test_that("check_min_row_count() returns FAIL when row count below threshold", {
  cfg <- base_config(list(rules = list(
    max_missing_rate = 0.05, max_non_numeric_rate = 0.01,
    min_row_count = 1000,
    max_row_count_change_pct = 0.10, max_numeric_mean_shift_pct = 0.20,
    max_missing_rate_change_pp = 2.0, max_non_numeric_rate_change_pp = 1.0
  )))
  res <- check_min_row_count(make_df(), cfg)
  expect_equal(res[[1]]$status, "FAIL")
})

# -- SC-01 / SC-02 Schema contract ---------------------------------------------

test_that("check_schema_contract() returns PASS when columns match exactly", {
  cfg <- base_config(list(
    expected_columns = c("id", "name", "country_code",
                         "account_status", "account_balance", "created_date")
  ))
  res <- check_schema_contract(make_df(), cfg)
  expect_true(all(vapply(res, \(r) r$status == "PASS", logical(1))))
})

test_that("check_schema_contract() returns FAIL for extra column (SC-01)", {
  df           <- make_df()
  df$extra_col <- "x"
  cfg <- base_config(list(
    expected_columns = c("id", "name", "country_code",
                         "account_status", "account_balance", "created_date")
  ))
  res  <- check_schema_contract(df, cfg)
  sc01 <- Filter(\(r) r$check_id == "SC-01", res)
  expect_true(any(vapply(sc01, \(r) r$status == "FAIL", logical(1))))
})

test_that("check_schema_contract() returns FAIL for missing column (SC-02)", {
  df  <- make_df()[, -3]
  cfg <- base_config(list(
    expected_columns = c("id", "name", "country_code",
                         "account_status", "account_balance", "created_date")
  ))
  res  <- check_schema_contract(df, cfg)
  sc02 <- Filter(\(r) r$check_id == "SC-02", res)
  expect_true(any(vapply(sc02, \(r) r$status == "FAIL", logical(1))))
})

test_that("check_schema_contract() returns empty list when expected_columns not set", {
  res <- check_schema_contract(make_df(), base_config())
  expect_equal(length(res), 0)
})

---
title: "Drift Report"
format:
  html:
    embed-resources: true
    theme: flatly
    toc: true
    toc-location: left
params:
  rds_path: ""
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE, message = FALSE, warning = FALSE)
library(kableExtra)
.p    <- readRDS(params$rds_path)
drift <- .p$drift
```

## `r drift$dataset_name`

| | Previous | Current |
|---|---|---|
| **Snapshot** | #`r drift$snap_prev$id` | #`r drift$snap_curr$id` |
| **Timestamp** | `r drift$snap_prev$run_timestamp` | `r drift$snap_curr$run_timestamp` |
| **File** | `r drift$snap_prev$file_name` | `r drift$snap_curr$file_name` |

---

## Table-Level Drift

```{r table-drift}
td <- drift$table_drift
td$Exceeds <- ifelse(
  td$Exceeds == "***",
  cell_spec("EXCEEDS", color = "red", bold = TRUE),
  ""
)
kbl(td, escape = FALSE) |>
  kable_styling(bootstrap_options = c("striped", "hover"), full_width = FALSE)
```

---

## Schema Drift

```{r schema-drift}
sc <- drift$schema_changes
if (nrow(sc) == 0) {
  cat("No schema changes detected.")
} else {
  sc$Status <- cell_spec(
    sc$Status,
    color = ifelse(sc$Status == "TYPE CHANGED", "orange",
            ifelse(sc$Status == "NEW COLUMN",   "blue", "red")),
    bold = TRUE
  )
  kbl(sc, escape = FALSE) |>
    kable_styling(bootstrap_options = c("striped", "hover"), full_width = FALSE)
}
```

---

## Per-Column Drift

### Missing Rate Changes

```{r missing-rate}
mr <- drift$missing_rate_changes
if (nrow(mr) > 0) {
  out <- data.frame(
    Column    = mr$Column,
    Previous  = sprintf("%.1f%%", mr$missing_rate_prev * 100),
    Current   = sprintf("%.1f%%", mr$missing_rate_curr * 100),
    Change_pp = sprintf("%+.2f", mr$missing_rate_change_pp),
    Exceeds   = ifelse(mr$missing_rate_exceeds,
                       cell_spec("YES", color = "red", bold = TRUE), ""),
    stringsAsFactors = FALSE, check.names = FALSE
  )
  kbl(out, escape = FALSE) |>
    kable_styling(bootstrap_options = c("striped", "hover", "condensed"),
                  full_width = FALSE)
} else {
  cat("No missing rate changes.")
}
```

### Non-Numeric Rate Changes

```{r non-numeric-rate}
nn <- drift$non_numeric_changes
if (nrow(nn) > 0) {
  out <- data.frame(
    Column    = nn$Column,
    Previous  = sprintf("%.2f%%", nn$non_numeric_rate_prev * 100),
    Current   = sprintf("%.2f%%", nn$non_numeric_rate_curr * 100),
    Change_pp = sprintf("%+.2f", nn$non_numeric_rate_change_pp),
    Exceeds   = ifelse(nn$non_numeric_rate_exceeds,
                       cell_spec("YES", color = "red", bold = TRUE), ""),
    stringsAsFactors = FALSE, check.names = FALSE
  )
  kbl(out, escape = FALSE) |>
    kable_styling(bootstrap_options = c("striped", "hover", "condensed"),
                  full_width = FALSE)
} else {
  cat("No non-numeric rate changes.")
}
```

### Numeric Mean Shifts

```{r numeric-mean}
ms <- drift$mean_shifts
if (nrow(ms) > 0) {
  out <- data.frame(
    Column    = ms$Column,
    Previous  = sprintf("%.4g", ms$numeric_mean_prev),
    Current   = sprintf("%.4g", ms$numeric_mean_curr),
    Shift_pct = sprintf("%+.1f%%", ms$numeric_mean_shift_pct * 100),
    Exceeds   = ifelse(ms$numeric_mean_exceeds,
                       cell_spec("YES", color = "red", bold = TRUE), ""),
    stringsAsFactors = FALSE, check.names = FALSE
  )
  kbl(out, escape = FALSE) |>
    kable_styling(bootstrap_options = c("striped", "hover", "condensed"),
                  full_width = FALSE)
} else {
  cat("No numeric columns to compare.")
}
```

### Distinct Count Changes

```{r distinct-count}
dc <- drift$distinct_changes
if (nrow(dc) > 0) {
  out <- data.frame(
    Column     = dc$Column,
    Previous   = dc$distinct_count_prev,
    Current    = dc$distinct_count_curr,
    Change     = sprintf("%+d", dc$distinct_count_change),
    Change_pct = ifelse(!is.na(dc$distinct_count_change_pct),
                        sprintf("%+.1f%%", dc$distinct_count_change_pct * 100),
                        ""),
    stringsAsFactors = FALSE
  )
  kbl(out, escape = FALSE) |>
    kable_styling(bootstrap_options = c("striped", "hover", "condensed"),
                  full_width = FALSE)
} else {
  cat("No distinct count changes.")
}
```

---
title: "dqcheckr — Software Specification"
author: "Mick Mioduszewski"
date: "2026-06-08"
version: "0.2.1"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{dqcheckr — Software Specification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)
```

**Version**: 0.2.1
**Author**: Mick Mioduszewski
**Date**: 2026-06-08

---

## Part A — Purpose and Intent

### Why this exists

The analytics organisation ingests external datasets from multiple suppliers on a recurring basis. Verifying that each new delivery is complete, structurally correct, and consistent with prior deliveries is currently slow and largely manual. Analysts spend time on repetitive visual inspection instead of higher-value work, and quality problems risk being missed or caught late.

### What we are building

`dqcheckr` is a purpose-built R package that automates this verification process. For each dataset arrival it produces a single, plain-English HTML report that tells data management staff:

1. Whether the file is structurally sound (columns, types, row count as expected).
2. Whether data values are within acceptable bounds and free of obvious errors.
3. How this delivery compares to the previous one — what has changed, what has drifted.
4. How this delivery compares to a defined schema contract, if one exists.

Results are also stored in a local SQLite database so that quality trends can be tracked over time across many deliveries.

### Design principles

- **Owned and evolvable.** The tool is first-party R code, not a thin wrapper around a third-party API. Every check is directly readable and modifiable.
- **Plain English outputs.** Reports are written for data management staff, not engineers. No R code, no package names, no jargon.
- **Fail loudly.** Errors in configuration, custom checks, or parsing stop the run with a clear message. Nothing is silently skipped.
- **Non-destructive.** The tool reads files and writes reports and database records. It never modifies, moves, or deletes source files.

---

## Part B — Business-Level Description

### B.1 How it is used

A data officer runs a single command for each arriving dataset:

```{r}
run_dq_check("customer_accounts", config_dir = "config")
```

The tool finds the two most recent files in the dataset's folder, checks the newer one, compares it to the older one, writes a timestamped HTML report to a reports folder, and prints a one-line summary to the console. The report opens automatically in the browser.

Files can also be identified explicitly by naming a current and a previous file, rather than relying on folder modification times.

Scale: the tool is designed for up to 10 dataset arrivals per day across all dataset types.

### B.2 What is checked

**Data completeness**

- Any column with too many missing values is flagged.
- Any column that is entirely empty is an error.
- If the file has fewer rows than the configured minimum, it is an error.

**Data validity**

- Columns expected to contain only specific values (e.g. country codes, status flags) are checked against a permitted list. Unexpected values are an error.
- Numeric columns expected to fall within a range are checked. Out-of-range values are an error.
- Columns expected to match a format (e.g. a postcode pattern) are checked by regular expression. Violations are an error.

**Data integrity**

- Columns designated as keys (e.g. a unique ID column) are checked for duplicates. Duplicates are an error.
- Rows that are entirely identical to another row in the same file are flagged as a warning.

**Data consistency**

- The inferred data type of each column is recorded. A type change between deliveries is flagged.
- Non-numeric values in a column that should be numeric are flagged (warning if any present; error if too many).

**Schema contract** *(when an expected column list is configured)*

- Columns present in the file but not in the expected list are an error.
- Columns in the expected list but absent from the file are an error.

**Custom checks** *(when a custom checks file is configured for a dataset)*

- Organisation-specific rules written in R are run against the file. Results appear alongside the standard checks.

### B.3 What is compared to the previous delivery

- **Row count**: a large change (increase or decrease) is flagged.
- **Schema**: new columns, dropped columns, and type changes are flagged.
- **Missing rates**: columns where the proportion of missing values has increased significantly are flagged.
- **Numeric distributions**: columns where the average value has shifted significantly are flagged.
- **Non-numeric contamination**: columns where the proportion of bad (non-numeric) values has increased are flagged.
- **Distinct values**: new values that have appeared and values that have disappeared in categorical columns are listed.
- **Column order**: a change in the order of columns is a warning for CSV files and an error for fixed-width files (where column order determines parsing).

### B.4 What is produced

**HTML report** — one file per run, self-contained (no internet required to view). Contains:

1. A header showing dataset name, file name, run time, and an overall status badge (green = all clear, amber = warnings, red = failures).
2. A file summary (row count, column count, file size, format, encoding).
3. A quality checks table listing every check, its status, what was observed, the threshold, and a plain-English explanation. Failures and warnings appear first.
4. A custom checks table (if configured).
5. A version comparison section showing what changed since the previous delivery, with a plain-English narrative summary at the top.
6. A historical trend section showing row count and check outcomes across the last 10 runs (appears once at least two runs have been recorded).
7. An appendix with a full column-level statistics table.

**SQLite database** — a single database shared across all datasets. Records summary statistics for every run, enabling long-term trend queries without reading the source files again.

---

## Part C — Technical Specification

### C.1 Package structure

```
dqcheckr/
├── R/
│   ├── run_check.R          # main entry point: run_dq_check()
│   ├── ingest.R             # read CSV / FWF; trim whitespace
│   ├── snapshot.R           # SQLite write + read_recent_snapshots()
│   ├── drift.R              # compare_snapshots(), list_snapshots(), drift helpers
│   ├── checks_generic.R     # QC-01..15 and SC-01..02
│   ├── checks_custom.R      # source and call custom_checks(df)
│   ├── compare.R            # CP-01..08
│   ├── report.R             # quarto::quarto_render() wrapper
│   ├── utils.R              # dq_result(), load_config(), infer_col_type(), helpers
│   └── dqcheckr-package.R   # package-level documentation
├── inst/
│   ├── templates/
│   │   ├── report.qmd       # main HTML report template
│   │   └── drift_report.qmd # snapshot drift report template
│   ├── config/
│   │   ├── dqcheckr.yml
│   │   └── example_dataset.yml
│   └── demonstrations/
│       ├── demo.R           # named-file mode demo (CSV + FWF)
│       ├── demo2.R          # folder-scan mode demo with version comparison
│       ├── README.md
│       ├── config/          # global + starwars_csv + starwars_fwf configs
│       ├── config2/         # global + starwars_folder configs
│       ├── custom2/         # starwars_custom.R (human-specific checks)
│       ├── data/            # starwars.csv, starwars.fwf
│       └── data2/           # starwars_v1.csv (original), starwars_v2.csv (perturbed)
├── tests/testthat/
│   ├── test-utils.R
│   ├── test-checks.R
│   ├── test-compare.R
│   ├── test-ingest.R
│   ├── test-snapshot.R
│   ├── test-drift.R
│   └── test-integration.R
├── vignettes/
│   ├── dqcheckr.Rmd         # user guide
│   └── specification.Rmd    # this document
├── DESCRIPTION
└── NAMESPACE
```

### C.2 Configuration

**Global config — `dqcheckr.yml`**

```yaml
snapshot_db: "data/snapshots.sqlite"
report_output_dir: "reports/"

default_rules:
  # Single-snapshot thresholds
  max_missing_rate: 0.05            # FAIL if column missing rate exceeds this
  max_non_numeric_rate: 0.01        # FAIL if non-numeric rate in numeric column exceeds this
  min_row_count: 0                  # FAIL if row count below this (0 = disabled)
  # max_row_count: 1000000          # optional FAIL if row count exceeds this
  # max_file_size_mb: 500           # optional FAIL if file size exceeds this
  # max_z_score: 3.0                # optional QC-15 outlier detection (Z-score)
  # iqr_fence_multiplier: 1.5       # optional QC-15 outlier detection (IQR fence)

  # Version comparison thresholds
  max_row_count_change_pct: 0.10
  max_numeric_mean_shift_pct: 0.20
  max_missing_rate_change_pp: 2.0   # percentage points
  max_non_numeric_rate_change_pp: 1.0
  missing_rate_change_severity: warn  # "warn" or "fail" — severity for CP-03 breaches

  # Schema flags
  flag_new_columns: true            # set false to suppress new-column warnings in CP-02
  flag_dropped_columns: true        # set false to suppress dropped-column warnings in CP-02
  flag_type_changes: true           # set false to suppress type-change warnings in CP-02
  flag_column_order_change: true    # set false to skip CP-08 entirely
  column_order_severity: warn       # "warn" or "fail" — overrides format default for CP-08
```

**Per-dataset config — `<dataset_name>.yml`**

```yaml
dataset_name: "customer_accounts"
folder: "data/incoming/customer_accounts/"
format: csv                 # csv | fwf
encoding: "UTF-8"
delimiter: ","              # csv only
fwf_widths: ~               # fwf only
fwf_col_names: ~            # fwf only
fwf_skip: 0                 # fwf only — header rows to skip (default 0)

# Optional: explicit file paths override folder-scan version detection
# current_file: "data/incoming/customer_accounts/20260410.csv"
# previous_file: "data/incoming/customer_accounts/20260401.csv"

# Optional: schema contract
expected_columns:
  - id
  - name
  - country_code
  - account_status
  - account_balance
  - created_date

# Optional: key columns that must be unique
key_columns:
  - id

# Optional rule overrides (merged over global defaults)
rule_overrides:
  max_missing_rate: 0.02
  max_row_count_change_pct: 0.05
  max_non_numeric_rate: 0.005
  min_row_count: 1000

# Optional per-column rules
column_rules:
  country_code:
    allowed_values: ["GB", "US", "DE", "FR"]
    pattern: "^[A-Z]{2}$"
  account_status:
    allowed_values: ["ACTIVE", "CLOSED", "SUSPENDED"]
  account_balance:
    min_value: 0
    max_value: 1000000

# Optional custom checks
custom_checks_file: "custom/customer_accounts_checks.R"
```

### C.3 File version detection

Two modes:

1. **Folder scan (default)**: the two most recently modified files in `folder` are used. Newest = current, second-newest = previous.
2. **Explicit naming**: if `current_file` (and optionally `previous_file`) are set in the dataset config, those paths are used directly. Useful when file naming conventions do not reflect delivery order, or when re-running a check for a specific pair of files.

If only a current file is available (either via folder scan or explicit config), the run proceeds in single-file mode: version comparison and vs-previous schema checks are skipped.

### C.4 Ingestion and whitespace trimming

1. File is read entirely as character columns via `readr::read_delim()` or `readr::read_fwf()`.
2. Leading and trailing whitespace is trimmed from every value immediately after reading.
3. The trimmed data frame is used for all subsequent checks and comparisons. The raw file is never modified.

### C.5 Type inference

Applied to each column of the trimmed data frame. Rules applied in order:

| Priority | Condition | Inferred type |
|----------|-----------|---------------|
| 1 | All non-null/non-empty values parse as a known date format | `date` |
| 2 | ≥ 90% of non-null/non-empty values coerce to numeric | `numeric` |
| 3 | Otherwise | `character` |
| — | All values null or empty | `unknown` |

Known date formats tested: `%Y-%m-%d`, `%d/%m/%Y`, `%m/%d/%Y`, `%Y%m%d`, `%d-%m-%Y`.

The 90% threshold means a column with up to 10% bad (non-numeric) values is still classified as numeric, making those bad values visible to QC-11.

### C.6 Quality checks

Each check returns one or more `dq_result` objects:

```{r}
dq_result(check_id, check_name, column = NA, status, observed, threshold = NA, message)
# status: "PASS" | "WARN" | "FAIL" | "INFO"
```

**Single-snapshot checks (QC series)**

| ID | Name | Applies to | FAIL condition | WARN condition |
|----|------|------------|----------------|----------------|
| QC-01 | Missing rate | Every column | `missing_rate > max_missing_rate` | — |
| QC-02 | Empty column | Every column | Column is 100% empty | — |
| QC-03 | Duplicate rows | Table | — | Any fully-duplicate rows exist |
| QC-04 | Row count | Table | — | — (INFO) |
| QC-05 | Column count | Table | — | — (INFO) |
| QC-06 | Inferred type | Every column | — | — (INFO) |
| QC-07 | Numeric stats | Numeric columns | — | — (INFO: min/max/mean/SD) |
| QC-08 | Distinct value count | Character columns | — | — (INFO) |
| QC-09 | Allowed values | Configured columns | Any value outside `allowed_values` | — |
| QC-10 | Numeric bounds | Configured columns | Any value outside `min_value`/`max_value` | — |
| QC-11 | Non-numeric values | Numeric columns | `non_numeric_rate > max_non_numeric_rate` | `non_numeric_rate > warn_non_numeric_rate` |
| QC-12 | Key uniqueness | Configured `key_columns` (single or composite) | Any duplicate value in key | — |
| QC-13 | Pattern / regex | Configured columns | Any value not matching `pattern` | — |
| QC-14 | Row count bounds | Table | `row_count < min_row_count` (if > 0); `row_count > max_row_count` (if set); file size > `max_file_size_mb` (if set) | — |
| QC-15 | Outlier detection | Numeric columns | Value exceeds Z-score (`max_z_score`) or IQR fence (`iqr_fence_multiplier`) threshold | — (skipped if neither threshold is configured) |

**Schema contract checks (SC series)** — only when `expected_columns` is configured

| ID | Name | FAIL condition |
|----|------|----------------|
| SC-01 | Unexpected column | Column in file but not in `expected_columns` |
| SC-02 | Missing expected column | Column in `expected_columns` but absent from file |

**Version comparison checks (CP series)** — skipped in single-file mode

| ID | Name | Flag condition | Severity |
|----|------|----------------|----------|
| CP-01 | Row count change | `abs(pct_change) > max_row_count_change_pct` (WARN if previous has 0 rows) | WARN |
| CP-02a | New columns | Columns in current not in previous (suppressible via `flag_new_columns: false`) | WARN |
| CP-02b | Dropped columns | Columns in previous not in current (suppressible via `flag_dropped_columns: false`) | WARN |
| CP-02c | Column type changes | Type of a common column changed (suppressible via `flag_type_changes: false`) | WARN |
| CP-03 | Missing rate change | Increase > `max_missing_rate_change_pp` pp per column; severity controlled by `missing_rate_change_severity` | WARN (default) / FAIL |
| CP-04 | Numeric mean shift | Shift > `max_numeric_mean_shift_pct` per column | WARN |
| CP-05 | New distinct values | Values in current not in previous (categorical columns) | INFO |
| CP-06 | Dropped distinct values | Values in previous not in current (categorical columns) | INFO |
| CP-07 | Non-numeric rate change | Increase > `max_non_numeric_rate_change_pp` pp per column; always emits a result | WARN |
| CP-08 | Column order change | Column order differs from previous (suppressible via `flag_column_order_change: false`); severity overridable via `column_order_severity` | WARN (CSV) / FAIL (FWF) |

> CP-02 is split into three separate results (CP-02a, CP-02b, CP-02c) so each type of schema change can be independently read in the report and suppressed via config. The `new_cols_vs_previous` and `missing_cols_vs_previous` columns in the `snapshots` table are derived directly from the CP-02a/b results — they are not computed separately.

### C.7 Custom checks

- `custom_checks_file` in dataset config points to an `.R` file.
- Sourced into an isolated environment whose parent is `baseenv()` (not the global env).
- Must define `custom_checks(df)` returning a list of `dq_result` objects.
- `dq_result()` is explicitly injected and available without qualification.
- All other dqcheckr exports (e.g. `resolve_col_type()`) require the `dqcheckr::` prefix.
- Receives the trimmed data frame.
- Errors (missing file, missing function, runtime error) stop the run.
- Results appear in the "Custom Checks" report section and are written to `column_snapshots` (see C.8).

### C.8 SQLite schema

Database and tables created automatically on first run via `init_snapshot_db()`.

**Table: `snapshots`**

| Column | Type | Notes |
|--------|------|-------|
| `id` | INTEGER PK | Auto-increment |
| `dataset_name` | TEXT | |
| `run_timestamp` | TEXT | `YYYY-MM-DD HH:MM:SS` |
| `file_name` | TEXT | Basename of current file |
| `row_count` | INTEGER | |
| `col_count` | INTEGER | |
| `check_pass_count` | INTEGER | |
| `check_warn_count` | INTEGER | |
| `check_fail_count` | INTEGER | |
| `check_info_count` | INTEGER | |
| `overall_status` | TEXT | `FAIL` > `WARN` > `PASS` > `INFO` |
| `new_cols_vs_previous` | TEXT | Comma-separated; NULL if none or no previous |
| `missing_cols_vs_previous` | TEXT | Comma-separated; NULL if none or no previous |
| `new_cols_vs_schema` | TEXT | Comma-separated; NULL if none or `expected_columns` not set |
| `missing_cols_vs_schema` | TEXT | Comma-separated; NULL if none or `expected_columns` not set |
| `comparison_mode` | TEXT | `"comparison"` or `"single"` (single-file runs have no previous) |
| `render_status` | TEXT | `"success"` or `"failed"` (updated if Quarto render fails) |
| `type_changed_cols_vs_previous` | TEXT | Comma-separated `col (prev->curr)` pairs; NULL if none |

**Table: `column_snapshots`**

| Column | Type | Notes |
|--------|------|-------|
| `id` | INTEGER PK | Auto-increment |
| `snapshot_id` | INTEGER FK | → `snapshots.id` |
| `column_name` | TEXT | Column name, or check_id for custom checks |
| `dq_check` | TEXT | Stat name or custom check_id |
| `value` | TEXT | Always stored as text |
| `threshold` | TEXT | NULL if no threshold applies |
| `severity_on_breach` | TEXT | `FAIL`, `WARN`, or NULL if informational |

Rows written per column type:

| `dq_check` | Written for | `threshold` | `severity_on_breach` |
|------------|-------------|-------------|----------------------|
| `inferred_type` | All columns | NULL | NULL |
| `missing_count` | All columns | NULL | NULL |
| `missing_rate` | All columns | `max_missing_rate` | `FAIL` |
| `distinct_count` | All columns | NULL | NULL |
| `numeric_parseable_mean` | Numeric only | NULL | NULL |
| `numeric_sd` | Numeric only | NULL | NULL |
| `numeric_min` | Numeric only | NULL | NULL |
| `numeric_max` | Numeric only | NULL | NULL |
| `non_numeric_count` | Numeric only | NULL | NULL |
| `non_numeric_rate` | Numeric only | `max_non_numeric_rate` | `FAIL` |

Custom check rows: one row per `dq_result` with a non-NA column value. `dq_check` = `check_id`, `value` = `observed`, `threshold` = `threshold`, `severity_on_breach` = status if WARN/FAIL, else NULL.

### C.9 Main entry points

**Standard run:**

```{r}
run_dq_check(
  dataset_name,        # matches <dataset_name>.yml in config_dir
  config_dir = ".",    # contains dqcheckr.yml and <dataset_name>.yml
  open_report = TRUE   # open in browser when interactive
)
# Returns invisibly: list(status, report_path, snapshot_id)
```

Report filename: `<dataset_name>_<YYYYMMDD_HHMMSS>.html`

**Snapshot drift analysis** (compare any two historical snapshots):

```{r}
compare_snapshots(
  dataset_name,
  snapshot_id_prev = NULL,   # defaults to second-most-recent
  snapshot_id_curr = NULL,   # defaults to most-recent
  db_path          = NULL,   # defaults to snapshot_db from dqcheckr.yml
  config_dir       = ".",
  report           = TRUE,   # render HTML drift report
  open_report      = interactive()
)
# Returns invisibly: named list with dataset_name, snap_prev, snap_curr,
#   table_drift, schema_changes, missing_rate_changes, non_numeric_changes,
#   mean_shifts, distinct_changes

list_snapshots(dataset_name = NULL, db_path)
# Returns a data frame of snapshot records; dataset_name = NULL returns all datasets
```

Console output:
```
[dqcheckr] customer_accounts: FAIL - 0 warning(s), 2 failure(s). Report: reports/customer_accounts_20260410_143022.html
```

### C.10 Error handling

| Situation | Behaviour |
|-----------|-----------|
| Folder does not exist | Stop; no output written |
| No files in folder | Stop; no output written |
| Explicit file path not found | Stop; no output written |
| Only one file available | Single-file mode; comparison and vs-previous schema skipped |
| File cannot be parsed | Stop with filename and parse error |
| SC-01 / SC-02 violations | FAIL recorded; run continues; report and snapshot written |
| Custom checks file missing | Stop with message |
| `custom_checks()` not defined | Stop with message |
| `custom_checks()` runtime error | Stop with R error detail |
| SQLite write fails | Warning emitted; HTML report still written |
| Report render fails | Warning emitted; snapshot marked `render_status = "failed"` |
| `snapshot_db` path missing | Created automatically by `init_snapshot_db()` |

### C.11 Dependencies

| Package | Role |
|---------|------|
| `readr` | CSV / FWF ingestion |
| `DBI` + `RSQLite` | Snapshot store |
| `quarto` | Report rendering (Quarto CLI required separately) |
| `knitr` | Template execution during Quarto render |
| `kableExtra` | Styled HTML tables in report templates |
| `ggplot2` + `gridExtra` | Trend charts in report templates |
| `yaml` | Config parsing |
| `dplyr` | Pipe operator (`%>%`) used in report templates |
| `tidyr` | Data reshaping for trend charts in report templates |
| `rlang` | Structured error conditions |

All on CRAN. `pointblank` and `dataCompareR` were evaluated and not used — all logic is implemented natively. `rmarkdown` is listed in `Suggests` for vignette compilation only; it is no longer used for report rendering.

---

## Part D — Unresolved and Deferred Issues

### D.1 Needs design before implementation

| # | Issue | Detail |
|---|-------|--------|
| D-01 | **Explicit file naming — partial spec** | The dataset config allows `current_file` and `previous_file` to override folder-scan detection, but the behaviour when one is set and the other is not has not been fully specified. Implemented as: `current_file` set but `previous_file` absent runs in single-file mode. |
| ~~D-02~~ | ~~**CP-08 column order — WARN vs FAIL severity**~~ | Resolved in v0.2.0: `column_order_severity` config key overrides the format-based default for CP-08. |
| ~~D-03~~ | ~~**QC-03 duplicate rows — no key scope**~~ | Resolved in v0.2.0: QC-12 now supports composite keys via a `key_columns` list. QC-03 retains full-row duplicate detection. |

### D.2 Backlog — agreed

| # | Feature |
|---|---------|
| B-01 | **Cross-column consistency rules** — conditional rules between columns (e.g. if `status = CLOSED` then `close_date` must not be null) |
| B-02 | **Date column recency check** — flag if the most recent value in a date column is older than N days |
| ~~B-03~~ | ~~**Multi-column key uniqueness**~~ | Resolved in v0.2.0: QC-12 now accepts a `key_columns` list for composite keys. |
| B-04 | **`post_run_hook` parameter** — user-supplied function for email/Slack/S3 delivery after a run |
| B-05 | **`batch_run(config_dir)`** — run all configured datasets in sequence |
| B-06 | **Additional input formats** — Parquet (`arrow`), Excel, database tables via `DBI` |
| ~~B-07~~ | ~~**CP-03 severity**~~ | Resolved in v0.2.0: `missing_rate_change_severity` config key controls whether CP-03 breaches are WARN or FAIL. |
| B-08 | **Report themes** — organisation branding via a `theme` parameter |

### D.3 Known limitations

| # | Limitation |
|---|------------|
| L-01 | ~~Type inference uses a fixed 90% threshold~~ — resolved in v0.1.1; configurable via `type_inference_threshold` in `rule_overrides` |
| L-02 | Date formats tested are hardcoded to five common patterns; exotic formats will be misclassified as character |
| L-03 | The report appendix reads from in-memory `col_stats`, not from SQLite. If the SQLite write fails but the report succeeds, the appendix and the database can show slightly different values in edge cases. Report render failures are non-fatal from v0.2.0: the snapshot is preserved and marked `render_status = "failed"`. |
| L-04 | Performance scales linearly with rows but super-linearly with columns. Benchmarked on macOS / aarch64 / R 4.6.0 with 1 M-row synthetic data: 10 cols → 47.6 s (21,031 rows/s); 20 cols → 91.9 s (10,881 rows/s); 49 cols → 138.9 s (7,200 rows/s). Very wide files (100+ columns) will be substantially slower. |

---
title: "Data Quality Report"
format:
  html:
    embed-resources: true
    theme: flatly
    toc: false
params:
  rds_path: ""
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE, message = FALSE, warning = FALSE,
                      fig.path = file.path(tempdir(), "dqcheckr_figs", ""))
library(kableExtra)
library(ggplot2)
library(dplyr)

p <- readRDS(params$rds_path)

`%||%` <- function(a, b) if (!is.null(a)) a else b

status_colour <- function(s) {
  switch(s, FAIL = "red", WARN = "orange", PASS = "green", INFO = "grey", "grey")
}

badge_html <- function(s) {
  col <- switch(s, FAIL = "#d9534f", WARN = "#f0ad4e", PASS = "#5cb85c", "#777")
  sprintf('<span style="background:%s;color:#fff;padding:4px 10px;border-radius:4px;font-weight:bold;">%s</span>',
          col, s)
}

results_to_df <- function(results) {
  if (length(results) == 0) return(data.frame())
  do.call(rbind, lapply(results, function(r) {
    data.frame(
      `Check ID`   = r$check_id,
      `Check Name` = r$check_name,
      Column       = ifelse(is.na(r$column), "-", r$column),
      Status       = r$status,
      Observed     = r$observed,
      Threshold    = ifelse(is.na(r$threshold), "-", r$threshold),
      Message      = r$message,
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  }))
}

status_order <- c(FAIL = 1, WARN = 2, PASS = 3, INFO = 4)

sort_results_df <- function(df) {
  if (nrow(df) == 0) return(df)
  df[order(status_order[df$Status]), ]
}

colour_status_col <- function(df) {
  df$Status <- kableExtra::cell_spec(
    df$Status,
    color = sapply(df$Status, status_colour),
    bold  = TRUE
  )
  df
}
```

## 1. Header

**Dataset**: `r p$dataset_name`
**File**: `r p$file_name`
**Run time**: `r p$run_timestamp`
**Overall status**: `r knitr::raw_html(badge_html(p$overall_status))`

---

## 2. File summary

```{r file-summary}
fsize <- tryCatch(file.info(p$file_path)$size, error = function(e) NA)
fsize_str <- if (!is.na(fsize)) sprintf("%.1f KB", fsize / 1024) else "N/A"

summary_df <- data.frame(
  Item  = c("Row count", "Column count", "File size", "Format", "Encoding"),
  Value = c(
    as.character(nrow(p$df)),
    as.character(ncol(p$df)),
    fsize_str,
    toupper(p$config$format %||% "CSV"),
    p$config$encoding %||% "UTF-8"
  ),
  stringsAsFactors = FALSE
)

kableExtra::kbl(summary_df, col.names = c("Item", "Value")) |>
  kableExtra::kable_styling(full_width = FALSE)
```

---

## 3. Quality checks

```{r qc-table}
qc_df <- results_to_df(p$qc_results)
if (nrow(qc_df) > 0) {
  qc_df <- sort_results_df(qc_df)
  qc_df <- colour_status_col(qc_df)
  kableExtra::kbl(qc_df, escape = FALSE) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "hover"))
} else {
  cat("No quality check results.")
}
```

---

```{r custom-section, results='asis'}
if (length(p$custom_results) > 0) {
  cat("## Custom checks\n\n")
  cc_df <- results_to_df(p$custom_results)
  cc_df <- sort_results_df(cc_df)
  cc_df <- colour_status_col(cc_df)
  print(kableExtra::kbl(cc_df, escape = FALSE) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "hover")))
  cat("\n\n---\n\n")
}
```

```{r comparison-section, results='asis'}
if (length(p$cp_results) > 0) {
  cat("## Version comparison\n\n")

  # C-05: show all FAIL/WARN messages as a bullet list
  bad_results <- Filter(\(r) r$status %in% c("FAIL", "WARN"), p$cp_results)
  if (length(bad_results) > 0) {
    cat("**Issues found:**\n\n")
    for (r in bad_results) cat(sprintf("- %s\n", r$message))
    cat("\n")
  } else {
    cat("**Summary**: All comparison checks passed.\n\n")
  }

  cp_df <- results_to_df(p$cp_results)
  cp_df <- sort_results_df(cp_df)
  cp_df <- colour_status_col(cp_df)
  print(kableExtra::kbl(cp_df, escape = FALSE) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "hover")))
  cat("\n\n---\n\n")
}
```

```{r trend-header, results='asis'}
hist <- p$snapshot_history
show_trend <- !is.null(hist) && nrow(hist) >= 2
if (show_trend) {
  hist$run_timestamp <- as.POSIXct(hist$run_timestamp,
                                    format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
  hist <- hist[order(hist$run_timestamp), ]
  cat("## Historical trend\n\n")
}
```

```{r trend-plot, fig.width=10, fig.height=4}
if (show_trend) {
  p1 <- ggplot(hist, aes(x = run_timestamp, y = row_count)) +
    geom_line() + geom_point() +
    labs(title = "Row count over last runs", x = "Run time", y = "Row count") +
    theme_minimal()

  hist_long <- tidyr::pivot_longer(
    hist[, c("run_timestamp", "check_pass_count",
             "check_warn_count", "check_fail_count")],
    cols = -run_timestamp, names_to = "type", values_to = "count"
  )
  hist_long$type <- gsub("check_|_count", "", hist_long$type)

  p2 <- ggplot(hist_long, aes(x = run_timestamp, y = count, fill = type)) +
    geom_col(position = "stack") +
    scale_fill_manual(values = c(pass = "#5cb85c", warn = "#f0ad4e", fail = "#d9534f")) +
    labs(title = "Check outcomes over last runs", x = "Run time", y = "Count",
         fill = "Status") +
    theme_minimal()

  gridExtra::grid.arrange(p1, p2, ncol = 2)
}
```

```{r trend-footer, results='asis'}
if (show_trend) cat("\n\n---\n\n")
```

## Appendix - Column statistics

Per-column statistics computed for this run.

```{r appendix}
if (!is.null(p$col_stats)) {
  col_stats <- p$col_stats
  names(col_stats) <- c("Column", "Stat", "Value", "Threshold", "Severity")
  kableExtra::kbl(col_stats) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "condensed"),
                              full_width = TRUE)
}
```

---
title: "Getting started with dqcheckr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with dqcheckr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)
```

`dqcheckr` automates the verification of recurring external dataset deliveries.
For each new file arrival it runs a battery of quality checks, compares the
file to the previous delivery, writes a self-contained HTML report, and records
summary statistics in a local SQLite database so that quality trends can be
tracked over time.

## How it works

```
Your data file                  Your YAML config
      │                                │
      ▼                                ▼
  read_dataset()            load_config()
  (all columns as           (global defaults +
   character, trimmed)       per-dataset rules)
      │                                │
      └──────────────┬─────────────────┘
                     ▼
          run_qc_checks()        ← single-snapshot checks (QC / SC series)
          run_comparison_checks()← vs previous delivery   (CP series)
          run_custom_checks()    ← your own rules
                     │
                     ▼
          write_snapshot()       ← SQLite: run summary + per-column stats
          render_report()        ← self-contained HTML
                     │
                     ▼
          run_dq_check()         ← the one function you call
```

A data officer runs a single command for each arriving dataset:

```{r one-call}
run_dq_check("customer_accounts", config_dir = "path/to/configs")
```

This prints a one-line console summary, writes an HTML report, and returns
`list(status, report_path, snapshot_id)` invisibly.

---

## Installation

`dqcheckr` is a first-party package — install it directly from source:

```{r install}
devtools::install("path/to/dqr/dqcheckr")
library(dqcheckr)
```

All dependencies are on CRAN: `readr`, `DBI`, `RSQLite`, `rmarkdown`,
`knitr`, `kableExtra`, `ggplot2`, `gridExtra`, `yaml`, `dplyr`, `tidyr`,
`rlang`.

---

## Configuration

Two YAML files control every run.

### Global config — `dqcheckr.yml`

Placed in your config directory. Sets default thresholds used by all datasets
unless a dataset overrides them.

```yaml
snapshot_db:       "data/snapshots.sqlite"
report_output_dir: "reports/"

default_rules:
  # Proportion of non-empty values that must parse as numeric for a column to
  # be classified as numeric (affects type inference across many checks)
  type_inference_threshold: 0.90

  # FAIL if any column's missing-value rate exceeds this (0-1)
  max_missing_rate: 0.05

  # FAIL if non-numeric values in a numeric column exceed this rate
  max_non_numeric_rate: 0.01

  # FAIL if row count drops below this (0 = disabled)
  min_row_count: 0
  # max_row_count: 1000000       # optional FAIL if row count exceeds this
  # max_file_size_mb: 500        # optional FAIL if file size exceeds this

  # Outlier detection (QC-15): both thresholds are optional; omit to skip
  # max_z_score: 3.0
  # iqr_fence_multiplier: 1.5

  # Version-comparison thresholds (used when a previous file is available)
  max_row_count_change_pct:       0.10
  max_numeric_mean_shift_pct:     0.20
  max_missing_rate_change_pp:     2.0    # percentage points
  max_non_numeric_rate_change_pp: 1.0
  missing_rate_change_severity:   warn   # set to "fail" to escalate CP-03 breaches to FAIL

  flag_new_columns:         true   # set false to suppress new-column warnings in CP-02
  flag_dropped_columns:     true   # set false to suppress dropped-column warnings in CP-02
  flag_type_changes:        true   # set false to suppress type-change warnings in CP-02
  flag_column_order_change: true   # set false to skip CP-08 entirely
  column_order_severity:    warn   # set to "fail" to escalate CP-08 breaches to FAIL for CSV
```

### Per-dataset config — `<dataset_name>.yml`

One file per dataset. Only the keys you specify are needed — everything else
inherits from the global defaults.

```yaml
dataset_name: "customer_accounts"   # for human reference only — the functional
                                     # identifier is the YAML filename, which must
                                     # match the dataset_name argument to run_dq_check()

# --- File location ---
# Option A: folder scan (two most recently modified files are used)
folder: "data/incoming/customer_accounts/"

# Option B: explicit paths (set current_file only → single-file mode,
#           no version comparison)
# current_file:  "data/incoming/customer_accounts/20260410.csv"
# previous_file: "data/incoming/customer_accounts/20260401.csv"

format:    csv
encoding:  "UTF-8"
delimiter: ","

# --- Schema contract (optional) ---
# SC-01 flags extra columns; SC-02 flags missing columns.
expected_columns:
  - id
  - name
  - country_code
  - account_status
  - account_balance
  - created_date

# --- Key columns (optional) ---
# QC-12 checks each for duplicate values.
key_columns:
  - id

# --- Rule overrides (optional) ---
# These replace the matching keys in default_rules for this dataset only.
rule_overrides:
  max_missing_rate:        0.02
  max_row_count_change_pct: 0.05
  min_row_count:           1000

# --- Per-column rules (optional) ---
column_rules:
  country_code:
    allowed_values: ["GB", "US", "DE", "FR"]   # QC-09
    pattern: "^[A-Z]{2}$"                       # QC-13
  account_status:
    allowed_values: ["ACTIVE", "CLOSED", "SUSPENDED"]
  account_balance:
    min_value: 0                                 # QC-10
    max_value: 1000000

# --- Custom checks (optional) ---
# custom_checks_file: "custom/customer_accounts_checks.R"

# --- Output paths (optional, override global) ---
# snapshot_db:       "data/snapshots.sqlite"
# report_output_dir: "reports/"
```

### What is required vs optional

The only fields the package requires to run are the file location (`folder`
or `current_file`) and the `format`. Every other config key is optional —
the corresponding checks are simply skipped when the key is absent.

| Config key | Skipped / default when absent |
|---|---|
| `key_columns` | QC-12 (key uniqueness) skipped |
| `expected_columns` | SC-01 and SC-02 (schema contract) skipped |
| `column_rules.allowed_values` | QC-09 (allowed values) skipped |
| `column_rules.min_value` / `max_value` | QC-10 (numeric bounds) skipped |
| `column_rules.pattern` | QC-13 (pattern / regex) skipped |
| `max_row_count` | QC-14 maximum row count sub-check skipped |
| `max_file_size_mb` | QC-14 file size sub-check skipped |
| `max_z_score` / `iqr_fence_multiplier` | QC-15 (outlier detection) skipped unless at least one is set |
| `custom_checks_file` | All custom checks skipped |
| `previous_file` / second file in folder | All CP-01–CP-08 (version comparison) skipped |
| `min_row_count` set to 0 | QC-14 minimum row count sub-check skipped |
| `type_inference_threshold` | Defaults to 0.90 (affects QC-06, QC-07, QC-08, QC-11, CP-02, CP-04, CP-05, CP-06, CP-07) |
| `missing_rate_change_severity` | Defaults to `"warn"`; set to `"fail"` to escalate CP-03 breaches |
| `flag_new_columns` | Defaults to `true`; set to `false` to suppress new-column warnings in CP-02 (still tracked in SQLite) |
| `flag_dropped_columns` | Defaults to `true`; set to `false` to suppress dropped-column warnings in CP-02 (still tracked in SQLite) |
| `flag_type_changes` | Defaults to `true`; set to `false` to suppress type-change warnings in CP-02 |
| `flag_column_order_change` | Defaults to `true`; set to `false` to skip CP-08 entirely |
| `column_order_severity` | Defaults to `"warn"` for CSV / `"fail"` for FWF; override with `"fail"` to force FAIL on CSV |

A minimal dataset config that runs basic checks with no column-level rules:

```yaml
dataset_name: "my_dataset"           # for reference; must match the YAML filename
current_file: "data/incoming/my_dataset.csv"
format: csv
```

### Fixed-width files

For fixed-width files, set `format: fwf` and supply column widths and names.
If the file has a header row, set `fwf_skip: 1` to skip it.

```yaml
format: fwf
fwf_skip: 1
fwf_widths:    [10, 20, 3, 10, 12, 11]
fwf_col_names: [id, name, country_code, account_status, account_balance, created_date]
```

---

## The quality checks

Each check returns one or more `dq_result` objects with four key fields:
`check_id`, `status` (`PASS` / `WARN` / `FAIL` / `INFO`), `observed`, and
`message`. The overall run status is the worst single status across all checks.

### Single-snapshot checks (QC series)

These run on the current file alone.

| ID | Name | Triggers on | Severity |
|----|------|-------------|----------|
| QC-01 | Missing rate | Column missing rate > `max_missing_rate` | FAIL |
| QC-02 | Empty column | Column is 100% empty | FAIL |
| QC-03 | Duplicate rows | Any fully-identical rows | WARN |
| QC-04 | Row count | Always | INFO |
| QC-05 | Column count | Always | INFO |
| QC-06 | Inferred type | Always (date / numeric / character / unknown) | INFO |
| QC-07 | Numeric stats | Numeric columns (min, max, mean, SD) | INFO |
| QC-08 | Distinct count | Character columns | INFO |
| QC-09 | Allowed values | Values outside `allowed_values` list | FAIL |
| QC-10 | Numeric bounds | Values outside `min_value` / `max_value` | FAIL |
| QC-11 | Non-numeric values | Non-numeric values in a numeric column | WARN / FAIL |
| QC-12 | Key uniqueness | Duplicate values in `key_columns` (single or composite) | FAIL |
| QC-13 | Pattern / regex | Values not matching `pattern` | FAIL |
| QC-14 | Row count bounds | Row count below `min_row_count`; above `max_row_count`; file size above `max_file_size_mb` | FAIL |
| QC-15 | Outlier detection | Numeric values outside Z-score (`max_z_score`) or IQR fence (`iqr_fence_multiplier`) | FAIL |

### Schema contract checks (SC series)

Only run when `expected_columns` is set.

| ID | Name | Triggers on | Severity |
|----|------|-------------|----------|
| SC-01 | Unexpected column | Column in file not in `expected_columns` | FAIL |
| SC-02 | Missing expected column | Column in `expected_columns` absent from file | FAIL |

### Version comparison checks (CP series)

Only run when a previous file is available. Skipped automatically in
single-file mode.

| ID | Name | Triggers on | Severity |
|----|------|-------------|----------|
| CP-01 | Row count change | Change > `max_row_count_change_pct` | WARN |
| CP-02 | Schema diff | Columns added, dropped, or type-changed | WARN |
| CP-03 | Missing rate change | Increase > `max_missing_rate_change_pp` pp | WARN |
| CP-04 | Numeric mean shift | Shift > `max_numeric_mean_shift_pct` | WARN |
| CP-05 | New distinct values | Values in current not seen in previous | INFO |
| CP-06 | Dropped distinct values | Values in previous absent from current | INFO |
| CP-07 | Non-numeric rate change | Increase > `max_non_numeric_rate_change_pp` pp | WARN |
| CP-08 | Column order change | Column order differs from previous | WARN (CSV) / FAIL (FWF) |

---

## Type inference

`infer_col_type()` classifies each column after whitespace trimming. Rules
are applied in priority order:

1. **date** — all non-empty values parse as a known date format
   (`%Y-%m-%d`, `%d/%m/%Y`, `%m/%d/%Y`, `%Y%m%d`, `%d-%m-%Y`)
2. **numeric** — ≥ `type_inference_threshold` of non-empty values coerce to numeric
3. **character** — everything else
4. **unknown** — all values are empty or `NA`

The threshold defaults to 90%, meaning a column with up to 10% bad values is
still classified as numeric, making those bad values visible to QC-11
(non-numeric contamination). It is configurable per dataset via
`type_inference_threshold` in `default_rules` or `rule_overrides`:

```yaml
# raise threshold — stricter, fewer columns classified as numeric
rule_overrides:
  type_inference_threshold: 0.95

# lower threshold — more lenient, columns with more noise still classified numeric
rule_overrides:
  type_inference_threshold: 0.80
```

Changing the threshold affects QC-06, QC-07, QC-08, QC-11, CP-02, CP-04,
CP-05, CP-06, and CP-07. Note that changing it between runs may cause CP-02
to report a spurious type change on columns that sit close to the old and new
thresholds, even if the underlying data did not change.

```{r type-inference}
infer_col_type(c("2024-01-15", "2024-06-30"))  # "date"
infer_col_type(c("100", "200", "N/A"))          # "character" (only 67% numeric)
infer_col_type(c("100", "200", "N/A", rep("50", 17)))  # "numeric" (95% numeric)
infer_col_type(c(NA, "", NA))                   # "unknown"
# with a custom threshold:
infer_col_type(c(rep("1", 17), "a", "b", "c"), threshold = 0.80)  # "numeric" (85%)
```

---

## Running a check

```{r run-check}
result <- run_dq_check(
  dataset_name = "customer_accounts",
  config_dir   = "path/to/configs",   # contains dqcheckr.yml + customer_accounts.yml
  open_report  = TRUE                 # open HTML in browser when interactive
)

# Console output (always printed):
# [dqcheckr] customer_accounts: FAIL - 1 warning(s), 2 failure(s). Report: reports/...html

result$status       # "PASS", "WARN", or "FAIL"
result$report_path  # full path to the HTML file
result$snapshot_id  # integer row ID in snapshots table
```

The HTML report contains:

1. **Header** — dataset name, file name, run time, overall status badge
2. **File summary** — row count, column count, file size, format, encoding
3. **Quality checks table** — every check result, failures first
4. **Custom checks table** — if configured
5. **Version comparison** — what changed since the previous delivery
6. **Historical trend** — row count and check outcomes across the last 10 runs
   (appears once at least two runs have been recorded)
7. **Appendix** — full column-level statistics table

---

## Calling individual checks

Every check function is exported and can be called directly, without going
through the full pipeline. All checks share the same signature —
`(df, config)` — and return a list of `dq_result` objects.

```{r individual-checks}
library(dqcheckr)

# Build a minimal config programmatically
cfg <- list(
  rules = list(
    max_missing_rate     = 0.05,
    max_non_numeric_rate = 0.01,
    min_row_count        = 0
  ),
  column_rules     = list(
    status = list(allowed_values = c("ACTIVE", "CLOSED"))
  ),
  key_columns      = "id",
  expected_columns = NULL
)

df <- read.csv("data/customer_accounts.csv", colClasses = "character")
df[] <- lapply(df, trimws)

# Run any individual check
check_missing_rate(df, cfg)       # QC-01
check_empty_column(df, cfg)       # QC-02
check_duplicate_rows(df, cfg)     # QC-03
check_row_count(df, cfg)          # QC-04  (INFO)
check_col_count(df, cfg)          # QC-05  (INFO)
check_inferred_types(df, cfg)     # QC-06  (INFO)
check_numeric_stats(df, cfg)      # QC-07  (INFO)
check_distinct_counts(df, cfg)    # QC-08  (INFO)
check_allowed_values(df, cfg)     # QC-09
check_numeric_bounds(df, cfg)     # QC-10
check_non_numeric(df, cfg)        # QC-11
check_key_uniqueness(df, cfg)     # QC-12
check_pattern(df, cfg)            # QC-13
check_min_row_count(df, cfg)      # QC-14
check_outliers(df, cfg)           # QC-15
check_schema_contract(df, cfg)    # SC-01 / SC-02

# Inspect a result
r <- check_missing_rate(df, cfg)[[1]]
r$status    # "PASS" or "FAIL"
r$observed  # e.g. "3.2% missing (4 of 125)"

# Run all single-snapshot checks at once
all_results <- run_qc_checks(df, cfg)

# Overall status across any result list
overall_status(all_results)  # "FAIL" > "WARN" > "PASS" > "INFO"
```

---

## Custom checks

Create a plain `.R` file that defines `custom_checks(df)`. It receives the
trimmed data frame and must return a list of `dq_result` objects.

```{r custom-file}
# File: custom/customer_accounts_checks.R

custom_checks <- function(df) {
  results <- list()

  # Rule: ACTIVE accounts must not have a zero balance
  active_zero <- df[df$account_status == "ACTIVE" &
                    !is.na(df$account_balance) &
                    df$account_balance == "0", ]
  n <- nrow(active_zero)
  results <- c(results, list(dq_result(
    check_id   = "CUST-01",
    check_name = "No zero-balance active accounts",
    column     = "account_balance",   # enables per-column storage in SQLite
    status     = if (n > 0) "FAIL" else "PASS",
    observed   = sprintf("%d ACTIVE account(s) with balance 0", n),
    message    = if (n > 0)
      sprintf("%d ACTIVE account(s) have a zero balance.", n)
    else
      "No ACTIVE accounts have a zero balance."
  )))

  results
}
```

Point to it in the dataset config:

```yaml
custom_checks_file: "custom/customer_accounts_checks.R"
```

The file is sourced in an isolated environment (not the global session) whose
parent is `baseenv()`. `dq_result()` is explicitly available without
qualification. All other dqcheckr functions (e.g. `resolve_col_type()`,
`infer_col_type()`) must be called with the package prefix:
`dqcheckr::resolve_col_type()`. Standard base R functions are fully available.
Any error — missing file, undefined function, runtime failure — stops the run
with a clear message.

**Tip**: always set `column =` when your check is scoped to a specific column.
Results with `column = NA` (the default) appear in the HTML report and
contribute to the overall status, but are not written to the `column_snapshots`
SQLite table and therefore do not appear in per-column trend queries.

---

## The snapshot database

Every run writes two rows to a shared SQLite database: one row in `snapshots`
(run-level summary) and one row per column-stat in `column_snapshots`.

```{r snapshot-query}
library(DBI)
library(RSQLite)

con <- dbConnect(SQLite(), "data/snapshots.sqlite")

# Recent runs for one dataset
dbGetQuery(con,
  "SELECT id, file_name, overall_status, check_fail_count, run_timestamp
   FROM snapshots
   WHERE dataset_name = 'customer_accounts'
   ORDER BY id DESC
   LIMIT 10")

# Column-level stats for the most recent run
dbGetQuery(con,
  "SELECT column_name, dq_check, value, threshold
   FROM column_snapshots
   WHERE snapshot_id = (
     SELECT MAX(id) FROM snapshots WHERE dataset_name = 'customer_accounts'
   )
   ORDER BY column_name, dq_check")

dbDisconnect(con)
```

Because all datasets share one database, you can query across them:

```{r cross-dataset}
dbGetQuery(con,
  "SELECT dataset_name, COUNT(*) AS runs,
          SUM(check_fail_count) AS total_failures
   FROM snapshots
   GROUP BY dataset_name")
```

The `output/` directory (and database file) are created automatically on the
first run if they do not exist.

---

## Worked example — Star Wars dataset

The following uses `dplyr::starwars` exported to CSV and fixed-width files.

```{r starwars-setup}
# Config directory contains:
#   dqcheckr.yml         — global thresholds
#   starwars_csv.yml     — CSV dataset config
#   starwars_fwf.yml     — FWF dataset config

# Run checks on both formats
result_csv <- run_dq_check("starwars_csv", config_dir = "config", open_report = TRUE)
result_fwf <- run_dq_check("starwars_fwf", config_dir = "config", open_report = TRUE)
```

The Star Wars config sets `max_missing_rate: 0.60` because `vehicles` (87%)
and `starships` (77%) are naturally sparse — most characters have none. Both
columns FAIL, which is the correct finding: they are genuinely incomplete and
a consumer should be aware before using them.

Expected console output:

```
[dqcheckr] starwars_csv: FAIL - 0 warning(s), 2 failure(s). Report: output/reports/starwars_csv_....html
[dqcheckr] starwars_fwf: FAIL - 0 warning(s), 2 failure(s). Report: output/reports/starwars_fwf_....html
```

---

## Error handling

| Situation | Behaviour |
|-----------|-----------|
| Config file not found | Stop immediately; no output written |
| Data file not found | Stop immediately; no output written |
| Only one file available | Single-file mode; CP checks skipped |
| File cannot be parsed | Stop with filename and parse error |
| SC-01 / SC-02 violations | FAIL recorded; run continues; report written |
| Custom checks file missing | Stop with message |
| `custom_checks()` not defined | Stop with message |
| `custom_checks()` runtime error | Stop with R error detail |
| SQLite write fails | Warning emitted; HTML report still written |

The package never modifies, moves, or deletes source files.

---

## Design principles

**Owned and evolvable.** Every check is plain R — readable, modifiable, and
not hidden behind a third-party API.

**Plain English outputs.** Reports are written for data management staff, not
engineers. No R code, no package names, no jargon.

**Fail loudly.** Configuration errors, missing files, and custom-check
failures all stop the run with a clear message. Nothing is silently skipped.

**Non-destructive.** The package reads files and writes reports and database
records. Source files are never touched.

---
title: "dqcheckr — Software Specification"
author: "Mick Mioduszewski"
date: "2026-05-10"
version: "0.1.1"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Software Specification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)
```

**Version**: 0.1.1
**Author**: Mick Mioduszewski
**Date**: 2026-05-10

---

## Part A — Purpose and Intent

### Why this exists

The analytics organisation ingests external datasets from multiple suppliers on a recurring basis. Verifying that each new delivery is complete, structurally correct, and consistent with prior deliveries is currently slow and largely manual. Analysts spend time on repetitive visual inspection instead of higher-value work, and quality problems risk being missed or caught late.

### What we are building

`dqcheckr` is a purpose-built R package that automates this verification process. For each dataset arrival it produces a single, plain-English HTML report that tells data management staff:

1. Whether the file is structurally sound (columns, types, row count as expected).
2. Whether data values are within acceptable bounds and free of obvious errors.
3. How this delivery compares to the previous one — what has changed, what has drifted.
4. How this delivery compares to a defined schema contract, if one exists.

Results are also stored in a local SQLite database so that quality trends can be tracked over time across many deliveries.

### Design principles

- **Owned and evolvable.** The tool is first-party R code, not a thin wrapper around a third-party API. Every check is directly readable and modifiable.
- **Plain English outputs.** Reports are written for data management staff, not engineers. No R code, no package names, no jargon.
- **Fail loudly.** Errors in configuration, custom checks, or parsing stop the run with a clear message. Nothing is silently skipped.
- **Non-destructive.** The tool reads files and writes reports and database records. It never modifies, moves, or deletes source files.

---

## Part B — Business-Level Description

### B.1 How it is used

A data officer runs a single command for each arriving dataset:

```{r}
run_dq_check("customer_accounts", config_dir = "config")
```

The tool finds the two most recent files in the dataset's folder, checks the newer one, compares it to the older one, writes a timestamped HTML report to a reports folder, and prints a one-line summary to the console. The report opens automatically in the browser.

Files can also be identified explicitly by naming a current and a previous file, rather than relying on folder modification times.

Scale: the tool is designed for up to 10 dataset arrivals per day across all dataset types.

### B.2 What is checked

**Data completeness**

- Any column with too many missing values is flagged.
- Any column that is entirely empty is an error.
- If the file has fewer rows than the configured minimum, it is an error.

**Data validity**

- Columns expected to contain only specific values (e.g. country codes, status flags) are checked against a permitted list. Unexpected values are an error.
- Numeric columns expected to fall within a range are checked. Out-of-range values are an error.
- Columns expected to match a format (e.g. a postcode pattern) are checked by regular expression. Violations are an error.

**Data integrity**

- Columns designated as keys (e.g. a unique ID column) are checked for duplicates. Duplicates are an error.
- Rows that are entirely identical to another row in the same file are flagged as a warning.

**Data consistency**

- The inferred data type of each column is recorded. A type change between deliveries is flagged.
- Non-numeric values in a column that should be numeric are flagged (warning if any present; error if too many).

**Schema contract** *(when an expected column list is configured)*

- Columns present in the file but not in the expected list are an error.
- Columns in the expected list but absent from the file are an error.

**Custom checks** *(when a custom checks file is configured for a dataset)*

- Organisation-specific rules written in R are run against the file. Results appear alongside the standard checks.

### B.3 What is compared to the previous delivery

- **Row count**: a large change (increase or decrease) is flagged.
- **Schema**: new columns, dropped columns, and type changes are flagged.
- **Missing rates**: columns where the proportion of missing values has increased significantly are flagged.
- **Numeric distributions**: columns where the average value has shifted significantly are flagged.
- **Non-numeric contamination**: columns where the proportion of bad (non-numeric) values has increased are flagged.
- **Distinct values**: new values that have appeared and values that have disappeared in categorical columns are listed.
- **Column order**: a change in the order of columns is a warning for CSV files and an error for fixed-width files (where column order determines parsing).

### B.4 What is produced

**HTML report** — one file per run, self-contained (no internet required to view). Contains:

1. A header showing dataset name, file name, run time, and an overall status badge (green = all clear, amber = warnings, red = failures).
2. A file summary (row count, column count, file size, format, encoding).
3. A quality checks table listing every check, its status, what was observed, the threshold, and a plain-English explanation. Failures and warnings appear first.
4. A custom checks table (if configured).
5. A version comparison section showing what changed since the previous delivery, with a plain-English narrative summary at the top.
6. A historical trend section showing row count and check outcomes across the last 10 runs (appears once at least two runs have been recorded).
7. An appendix with a full column-level statistics table.

**SQLite database** — a single database shared across all datasets. Records summary statistics for every run, enabling long-term trend queries without reading the source files again.

---

## Part C — Technical Specification

### C.1 Package structure

```
dqcheckr/
├── R/
│   ├── run_check.R          # main entry point: run_dq_check()
│   ├── ingest.R             # read CSV / FWF; trim whitespace
│   ├── snapshot.R           # SQLite read/write
│   ├── checks_generic.R     # QC-01..14 and SC-01..02
│   ├── checks_custom.R      # source and call custom_checks(df)
│   ├── compare.R            # CP-01..08
│   ├── report.R             # rmarkdown::render() wrapper
│   ├── utils.R              # dq_result(), load_config(), infer_col_type(), helpers
│   └── dqcheckr-package.R   # package-level documentation
├── inst/
│   ├── templates/report.Rmd
│   ├── config/
│   │   ├── dqcheckr.yml
│   │   └── example_dataset.yml
│   └── demonstrations/
│       ├── demo.R           # named-file mode demo (CSV + FWF)
│       ├── demo2.R          # folder-scan mode demo with version comparison
│       ├── README.md
│       ├── config/          # global + starwars_csv + starwars_fwf configs
│       ├── config2/         # global + starwars_folder configs
│       ├── custom2/         # starwars_custom.R (human-specific checks)
│       ├── data/            # starwars.csv, starwars.fwf
│       └── data2/           # starwars_v1.csv (original), starwars_v2.csv (perturbed)
├── tests/testthat/
│   ├── test-utils.R
│   ├── test-checks.R
│   ├── test-compare.R
│   ├── test-ingest.R
│   ├── test-snapshot.R
│   └── test-integration.R
├── vignettes/
│   ├── dqcheckr.Rmd         # user guide
│   └── specification.Rmd    # this document
├── DESCRIPTION
└── NAMESPACE
```

### C.2 Configuration

**Global config — `dqcheckr.yml`**

```yaml
snapshot_db: "data/snapshots.sqlite"
report_output_dir: "reports/"

default_rules:
  # Single-snapshot thresholds
  max_missing_rate: 0.05            # FAIL if column missing rate exceeds this
  max_non_numeric_rate: 0.01        # FAIL if non-numeric rate in numeric column exceeds this
  min_row_count: 0                  # FAIL if row count below this (0 = disabled)

  # Version comparison thresholds
  max_row_count_change_pct: 0.10
  max_numeric_mean_shift_pct: 0.20
  max_missing_rate_change_pp: 2.0   # percentage points
  max_non_numeric_rate_change_pp: 1.0

  # Schema flags
  flag_new_columns: true            # set false to suppress new-column warnings in CP-02
  flag_dropped_columns: true        # set false to suppress dropped-column warnings in CP-02
  flag_type_changes: true           # set false to suppress type-change warnings in CP-02
  flag_column_order_change: true    # set false to skip CP-08 entirely (WARN CSV / FAIL FWF)
```

**Per-dataset config — `<dataset_name>.yml`**

```yaml
dataset_name: "customer_accounts"
folder: "data/incoming/customer_accounts/"
format: csv                 # csv | fwf
encoding: "UTF-8"
delimiter: ","              # csv only
fwf_widths: ~               # fwf only
fwf_col_names: ~            # fwf only
fwf_skip: 0                 # fwf only — header rows to skip (default 0)

# Optional: explicit file paths override folder-scan version detection
# current_file: "data/incoming/customer_accounts/20260410.csv"
# previous_file: "data/incoming/customer_accounts/20260401.csv"

# Optional: schema contract
expected_columns:
  - id
  - name
  - country_code
  - account_status
  - account_balance
  - created_date

# Optional: key columns that must be unique
key_columns:
  - id

# Optional rule overrides (merged over global defaults)
rule_overrides:
  max_missing_rate: 0.02
  max_row_count_change_pct: 0.05
  max_non_numeric_rate: 0.005
  min_row_count: 1000

# Optional per-column rules
column_rules:
  country_code:
    allowed_values: ["GB", "US", "DE", "FR"]
    pattern: "^[A-Z]{2}$"
  account_status:
    allowed_values: ["ACTIVE", "CLOSED", "SUSPENDED"]
  account_balance:
    min_value: 0
    max_value: 1000000

# Optional custom checks
custom_checks_file: "custom/customer_accounts_checks.R"
```

### C.3 File version detection

Two modes:

1. **Folder scan (default)**: the two most recently modified files in `folder` are used. Newest = current, second-newest = previous.
2. **Explicit naming**: if `current_file` (and optionally `previous_file`) are set in the dataset config, those paths are used directly. Useful when file naming conventions do not reflect delivery order, or when re-running a check for a specific pair of files.

If only a current file is available (either via folder scan or explicit config), the run proceeds in single-file mode: version comparison and vs-previous schema checks are skipped.

### C.4 Ingestion and whitespace trimming

1. File is read entirely as character columns via `readr::read_delim()` or `readr::read_fwf()`.
2. Leading and trailing whitespace is trimmed from every value immediately after reading.
3. The trimmed data frame is used for all subsequent checks and comparisons. The raw file is never modified.

### C.5 Type inference

Applied to each column of the trimmed data frame. Rules applied in order:

| Priority | Condition | Inferred type |
|----------|-----------|---------------|
| 1 | All non-null/non-empty values parse as a known date format | `date` |
| 2 | ≥ 90% of non-null/non-empty values coerce to numeric | `numeric` |
| 3 | Otherwise | `character` |
| — | All values null or empty | `unknown` |

Known date formats tested: `%Y-%m-%d`, `%d/%m/%Y`, `%m/%d/%Y`, `%Y%m%d`, `%d-%m-%Y`.

The 90% threshold means a column with up to 10% bad (non-numeric) values is still classified as numeric, making those bad values visible to QC-11.

### C.6 Quality checks

Each check returns one or more `dq_result` objects:

```{r}
dq_result(check_id, check_name, column = NA, status, observed, threshold = NA, message)
# status: "PASS" | "WARN" | "FAIL" | "INFO"
```

**Single-snapshot checks (QC series)**

| ID | Name | Applies to | FAIL condition | WARN condition |
|----|------|------------|----------------|----------------|
| QC-01 | Missing rate | Every column | `missing_rate > max_missing_rate` | — |
| QC-02 | Empty column | Every column | Column is 100% empty | — |
| QC-03 | Duplicate rows | Table | — | Any fully-duplicate rows exist |
| QC-04 | Row count | Table | — | — (INFO) |
| QC-05 | Column count | Table | — | — (INFO) |
| QC-06 | Inferred type | Every column | — | — (INFO) |
| QC-07 | Numeric stats | Numeric columns | — | — (INFO: min/max/mean/SD) |
| QC-08 | Distinct value count | Character columns | — | — (INFO) |
| QC-09 | Allowed values | Configured columns | Any value outside `allowed_values` | — |
| QC-10 | Numeric bounds | Configured columns | Any value outside `min_value`/`max_value` | — |
| QC-11 | Non-numeric values | Numeric columns | `non_numeric_rate > max_non_numeric_rate` | Any non-numeric values present |
| QC-12 | Key uniqueness | Configured `key_columns` | Any duplicate value in key column | — |
| QC-13 | Pattern / regex | Configured columns | Any value not matching `pattern` | — |
| QC-14 | Minimum row count | Table | `row_count < min_row_count` (if > 0) | — |

**Schema contract checks (SC series)** — only when `expected_columns` is configured

| ID | Name | FAIL condition |
|----|------|----------------|
| SC-01 | Unexpected column | Column in file but not in `expected_columns` |
| SC-02 | Missing expected column | Column in `expected_columns` but absent from file |

**Version comparison checks (CP series)** — skipped in single-file mode

| ID | Name | Flag condition | Severity |
|----|------|----------------|----------|
| CP-01 | Row count change | `abs(pct_change) > max_row_count_change_pct` | WARN |
| CP-02 | Schema diff | Columns added, dropped, or type-changed (each independently suppressible via `flag_new_columns`, `flag_dropped_columns`, `flag_type_changes`) | WARN |
| CP-03 | Missing rate change | Increase > `max_missing_rate_change_pp` pp per column | WARN |
| CP-04 | Numeric mean shift | Shift > `max_numeric_mean_shift_pct` per column | WARN |
| CP-05 | New distinct values | Values in current not in previous (categorical columns) | INFO |
| CP-06 | Dropped distinct values | Values in previous not in current (categorical columns) | INFO |
| CP-07 | Non-numeric rate change | Increase > `max_non_numeric_rate_change_pp` pp per column | WARN |
| CP-08 | Column order change | Column order differs from previous (suppressible via `flag_column_order_change: false`) | WARN (CSV) / FAIL (FWF) |

> CP-02 is the single source of truth for schema differences. The `new_cols_vs_previous` and `missing_cols_vs_previous` columns in the `snapshots` table are derived directly from the CP-02 result — they are not computed separately.

### C.7 Custom checks

- `custom_checks_file` in dataset config points to an `.R` file.
- Sourced into an isolated environment (not the global env).
- Must define `custom_checks(df)` returning a list of `dq_result` objects.
- Receives the trimmed data frame.
- Errors (missing file, missing function, runtime error) stop the run.
- Results appear in the "Custom Checks" report section and are written to `column_snapshots` (see C.8).

### C.8 SQLite schema

Database and tables created automatically on first run via `init_snapshot_db()`.

**Table: `snapshots`**

| Column | Type | Notes |
|--------|------|-------|
| `id` | INTEGER PK | Auto-increment |
| `dataset_name` | TEXT | |
| `run_timestamp` | TEXT | `YYYY-MM-DD HH:MM:SS` |
| `file_name` | TEXT | Basename of current file |
| `row_count` | INTEGER | |
| `col_count` | INTEGER | |
| `check_pass_count` | INTEGER | |
| `check_warn_count` | INTEGER | |
| `check_fail_count` | INTEGER | |
| `check_info_count` | INTEGER | |
| `overall_status` | TEXT | `FAIL` > `WARN` > `PASS` > `INFO` |
| `new_cols_vs_previous` | TEXT | Comma-separated; NULL if none or no previous |
| `missing_cols_vs_previous` | TEXT | Comma-separated; NULL if none or no previous |
| `new_cols_vs_schema` | TEXT | Comma-separated; NULL if none or `expected_columns` not set |
| `missing_cols_vs_schema` | TEXT | Comma-separated; NULL if none or `expected_columns` not set |

**Table: `column_snapshots`**

| Column | Type | Notes |
|--------|------|-------|
| `id` | INTEGER PK | Auto-increment |
| `snapshot_id` | INTEGER FK | → `snapshots.id` |
| `column_name` | TEXT | Column name, or check_id for custom checks |
| `dq_check` | TEXT | Stat name or custom check_id |
| `value` | TEXT | Always stored as text |
| `threshold` | TEXT | NULL if no threshold applies |
| `severity_on_breach` | TEXT | `FAIL`, `WARN`, or NULL if informational |

Rows written per column type:

| `dq_check` | Written for | `threshold` | `severity_on_breach` |
|------------|-------------|-------------|----------------------|
| `inferred_type` | All columns | NULL | NULL |
| `missing_count` | All columns | NULL | NULL |
| `missing_rate` | All columns | `max_missing_rate` | `FAIL` |
| `distinct_count` | All columns | NULL | NULL |
| `numeric_mean` | Numeric only | NULL | NULL |
| `numeric_sd` | Numeric only | NULL | NULL |
| `numeric_min` | Numeric only | NULL | NULL |
| `numeric_max` | Numeric only | NULL | NULL |
| `non_numeric_count` | Numeric only | NULL | NULL |
| `non_numeric_rate` | Numeric only | `max_non_numeric_rate` | `FAIL` |

Custom check rows: one row per `dq_result` with a non-NA column value. `dq_check` = `check_id`, `value` = `observed`, `threshold` = `threshold`, `severity_on_breach` = status if WARN/FAIL, else NULL.

### C.9 Main entry point

```{r}
run_dq_check(
  dataset_name,        # matches <dataset_name>.yml in config_dir
  config_dir = ".",    # contains dqcheckr.yml and <dataset_name>.yml
  open_report = TRUE   # open in browser when interactive
)
# Returns invisibly: list(status, report_path, snapshot_id)
```

Report filename: `<dataset_name>_<YYYYMMDD_HHMMSS>.html`

Console output:
```
[dqcheckr] customer_accounts: FAIL - 0 warning(s), 2 failure(s). Report: reports/customer_accounts_20260410_143022.html
```

### C.10 Error handling

| Situation | Behaviour |
|-----------|-----------|
| Folder does not exist | Stop; no output written |
| No files in folder | Stop; no output written |
| Explicit file path not found | Stop; no output written |
| Only one file available | Single-file mode; comparison and vs-previous schema skipped |
| File cannot be parsed | Stop with filename and parse error |
| SC-01 / SC-02 violations | FAIL recorded; run continues; report and snapshot written |
| Custom checks file missing | Stop with message |
| `custom_checks()` not defined | Stop with message |
| `custom_checks()` runtime error | Stop with R error detail |
| SQLite write fails | Warning emitted; HTML report still written |
| Report render fails | Stop; snapshot already written |
| `snapshot_db` path missing | Created automatically by `init_snapshot_db()` |

### C.11 Dependencies

| Package | Role |
|---------|------|
| `readr` | CSV / FWF ingestion |
| `DBI` + `RSQLite` | Snapshot store |
| `rmarkdown` + `knitr` | Report rendering |
| `kableExtra` | Styled HTML tables in report |
| `ggplot2` + `gridExtra` | Trend charts in report |
| `yaml` | Config parsing |
| `dplyr` | Data manipulation |
| `tidyr` | Data reshaping for trend charts |
| `rlang` | Structured error conditions |

All on CRAN. `pointblank` and `dataCompareR` were evaluated and not used — all logic is implemented natively.

---

## Part D — Unresolved and Deferred Issues

### D.1 Needs design before implementation

| # | Issue | Detail |
|---|-------|--------|
| D-01 | **Explicit file naming — partial spec** | The dataset config allows `current_file` and `previous_file` to override folder-scan detection, but the behaviour when one is set and the other is not has not been fully specified. Implemented as: `current_file` set but `previous_file` absent runs in single-file mode. |
| D-02 | **CP-08 column order — WARN vs FAIL severity** | CP-08 is WARN for CSV and FAIL for FWF. There is currently no way to override the severity per dataset. A `column_order_change_severity` per-dataset override may be needed. |
| D-03 | **QC-03 duplicate rows — no key scope** | QC-03 checks for fully-identical rows only. Multi-column key uniqueness (two rows with the same `id` but different `amount`) is not yet implemented. |

### D.2 Backlog — agreed

| # | Feature |
|---|---------|
| B-01 | **Cross-column consistency rules** — conditional rules between columns (e.g. if `status = CLOSED` then `close_date` must not be null) |
| B-02 | **Date column recency check** — flag if the most recent value in a date column is older than N days |
| B-03 | **Multi-column key uniqueness** — QC-12 currently checks single columns only |
| B-04 | **`post_run_hook` parameter** — user-supplied function for email/Slack/S3 delivery after a run |
| B-05 | **`batch_run(config_dir)`** — run all configured datasets in sequence |
| B-06 | **Additional input formats** — Parquet (`arrow`), Excel, database tables via `DBI` |
| B-07 | **CP-03 severity** — make WARN vs FAIL configurable (currently always WARN) |
| B-08 | **Report themes** — organisation branding via a `theme` parameter |

### D.3 Known limitations

| # | Limitation |
|---|------------|
| L-01 | ~~Type inference uses a fixed 90% threshold~~ — resolved in v0.1.1; configurable via `type_inference_threshold` in `rule_overrides` |
| L-02 | Date formats tested are hardcoded to five common patterns; exotic formats will be misclassified as character |
| L-03 | The report appendix reads from in-memory `col_stats`, not from SQLite. If the SQLite write fails but the report succeeds, the appendix and the database can show slightly different values in edge cases |
| L-04 | Performance target (< 60 seconds for 1 M rows) has not been benchmarked; very wide files (many columns) may be slower than expected |

---
title: "Data Quality Report"
output:
  html_document:
    self_contained: true
    theme: flatly
    toc: true
    toc_float: true
params:
  dataset_name: ""
  file_name: ""
  file_path: ""
  run_timestamp: ""
  df: NULL
  qc_results: NULL
  cp_results: NULL
  custom_results: NULL
  snapshot_history: NULL
  config: NULL
  col_stats: NULL
  overall_status: ""
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE, message = FALSE, warning = FALSE,
                      fig.path = file.path(tempdir(), "dqcheckr_figs", ""))
library(kableExtra)
library(ggplot2)
library(dplyr)

`%||%` <- function(a, b) if (!is.null(a)) a else b

status_colour <- function(s) {
  switch(s,
    FAIL = "red",
    WARN = "orange",
    PASS = "green",
    INFO = "grey",
    "grey"
  )
}

badge_html <- function(s) {
  col <- switch(s, FAIL = "#d9534f", WARN = "#f0ad4e", PASS = "#5cb85c", "#777")
  sprintf('<span style="background:%s;color:#fff;padding:4px 10px;border-radius:4px;font-weight:bold;">%s</span>', col, s)
}

results_to_df <- function(results) {
  if (length(results) == 0) return(data.frame())
  do.call(rbind, lapply(results, function(r) {
    data.frame(
      `Check ID`   = r$check_id,
      `Check Name` = r$check_name,
      Column       = ifelse(is.na(r$column), "-", r$column),
      Status       = r$status,
      Observed     = r$observed,
      Threshold    = ifelse(is.na(r$threshold), "-", r$threshold),
      Message      = r$message,
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
  }))
}

status_order <- c(FAIL = 1, WARN = 2, PASS = 3, INFO = 4)

sort_results_df <- function(df) {
  if (nrow(df) == 0) return(df)
  df[order(status_order[df$Status]), ]
}

colour_status_col <- function(df) {
  df$Status <- kableExtra::cell_spec(
    df$Status,
    color = sapply(df$Status, status_colour),
    bold  = TRUE
  )
  df
}
```

## 1. Header

**Dataset**: `r params$dataset_name`
**File**: `r params$file_name`
**Run time**: `r params$run_timestamp`
**Overall status**: `r knitr::raw_html(badge_html(params$overall_status))`

---

## 2. File summary

```{r file-summary}
fsize <- tryCatch(file.info(params$file_path)$size, error = function(e) NA)
fsize_str <- if (!is.na(fsize)) {
  sprintf("%.1f KB", fsize / 1024)
} else {
  "N/A"
}

summary_df <- data.frame(
  Item  = c("Row count", "Column count", "File size", "Format", "Encoding"),
  Value = c(
    as.character(nrow(params$df)),
    as.character(ncol(params$df)),
    fsize_str,
    toupper(params$config$format %||% "CSV"),
    params$config$encoding %||% "UTF-8"
  ),
  stringsAsFactors = FALSE
)

kableExtra::kbl(summary_df, col.names = c("Item", "Value")) |>
  kableExtra::kable_styling(full_width = FALSE)
```

---

## 3. Quality checks

```{r qc-table}
qc_df <- results_to_df(params$qc_results)
if (nrow(qc_df) > 0) {
  qc_df <- sort_results_df(qc_df)
  qc_df <- colour_status_col(qc_df)
  kableExtra::kbl(qc_df, escape = FALSE) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "hover"))
} else {
  cat("No quality check results.")
}
```

---

```{r custom-section, results='asis'}
if (length(params$custom_results) > 0) {
  cat("## Custom checks\n\n")
  cc_df <- results_to_df(params$custom_results)
  cc_df <- sort_results_df(cc_df)
  cc_df <- colour_status_col(cc_df)
  print(kableExtra::kbl(cc_df, escape = FALSE) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "hover")))
  cat("\n\n---\n\n")
}
```

```{r comparison-section, results='asis'}
if (length(params$cp_results) > 0) {
  cat("## Version comparison\n\n")

  top_result <- params$cp_results[[which.min(
    sapply(params$cp_results, \(r) status_order[r$status])
  )]]
  cat(sprintf("**Summary**: %s\n\n", top_result$message))

  cp_df <- results_to_df(params$cp_results)
  cp_df <- sort_results_df(cp_df)
  cp_df <- colour_status_col(cp_df)
  print(kableExtra::kbl(cp_df, escape = FALSE) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "hover")))
  cat("\n\n---\n\n")
}
```

```{r trend-header, results='asis'}
hist <- params$snapshot_history
show_trend <- !is.null(hist) && nrow(hist) >= 2
if (show_trend) {
  hist$run_timestamp <- as.POSIXct(hist$run_timestamp)
  hist <- hist[order(hist$run_timestamp), ]
  cat("## Historical trend\n\n")
}
```

```{r trend-plot, fig.width=10, fig.height=4}
if (show_trend) {
  p1 <- ggplot(hist, aes(x = run_timestamp, y = row_count)) +
    geom_line() + geom_point() +
    labs(title = "Row count over last runs", x = "Run time", y = "Row count") +
    theme_minimal()

  hist_long <- tidyr::pivot_longer(
    hist[, c("run_timestamp", "check_pass_count",
             "check_warn_count", "check_fail_count")],
    cols = -run_timestamp,
    names_to  = "type",
    values_to = "count"
  )
  hist_long$type <- gsub("check_|_count", "", hist_long$type)

  p2 <- ggplot(hist_long, aes(x = run_timestamp, y = count, fill = type)) +
    geom_col(position = "stack") +
    scale_fill_manual(values = c(pass = "#5cb85c", warn = "#f0ad4e", fail = "#d9534f")) +
    labs(title = "Check outcomes over last runs", x = "Run time", y = "Count",
         fill = "Status") +
    theme_minimal()

  gridExtra::grid.arrange(p1, p2, ncol = 2)
}
```

```{r trend-footer, results='asis'}
if (show_trend) cat("\n\n---\n\n")
```

## Appendix - Column statistics

Per-column statistics computed for this run. These correspond to the rows
written to the `column_snapshots` table in the SQLite snapshot database
(`column_name` -> Column, `dq_check` -> Stat, `value` -> Value,
`threshold` -> Threshold, `severity_on_breach` -> Severity).

```{r appendix}
if (!is.null(params$col_stats)) {
  col_stats <- params$col_stats
  names(col_stats) <- c("Column", "Stat", "Value", "Threshold", "Severity")
  kableExtra::kbl(col_stats) |>
    kableExtra::kable_styling(bootstrap_options = c("striped", "condensed"),
                              full_width = TRUE)
}
```

---
title: "Getting started with dqcheckr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with dqcheckr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)
```

`dqcheckr` automates the verification of recurring external dataset deliveries.
For each new file arrival it runs a battery of quality checks, compares the
file to the previous delivery, writes a self-contained HTML report, and records
summary statistics in a local SQLite database so that quality trends can be
tracked over time.

## How it works

```
Your data file                  Your YAML config
      │                                │
      ▼                                ▼
  read_dataset()            load_config()
  (all columns as           (global defaults +
   character, trimmed)       per-dataset rules)
      │                                │
      └──────────────┬─────────────────┘
                     ▼
          run_qc_checks()        ← single-snapshot checks (QC / SC series)
          run_comparison_checks()← vs previous delivery   (CP series)
          run_custom_checks()    ← your own rules
                     │
                     ▼
          write_snapshot()       ← SQLite: run summary + per-column stats
          render_report()        ← self-contained HTML
                     │
                     ▼
          run_dq_check()         ← the one function you call
```

A data officer runs a single command for each arriving dataset:

```{r one-call}
run_dq_check("customer_accounts", config_dir = "path/to/configs")
```

This prints a one-line console summary, writes an HTML report, and returns
`list(status, report_path, snapshot_id)` invisibly.

---

## Installation

`dqcheckr` is a first-party package — install it directly from source:

```{r install}
devtools::install("path/to/dqr/dqcheckr")
library(dqcheckr)
```

All dependencies are on CRAN: `readr`, `DBI`, `RSQLite`, `rmarkdown`,
`knitr`, `kableExtra`, `ggplot2`, `gridExtra`, `yaml`, `dplyr`, `tidyr`,
`rlang`.

---

## Configuration

Two YAML files control every run.

### Global config — `dqcheckr.yml`

Placed in your config directory. Sets default thresholds used by all datasets
unless a dataset overrides them.

```yaml
snapshot_db:       "data/snapshots.sqlite"
report_output_dir: "reports/"

default_rules:
  # Proportion of non-empty values that must parse as numeric for a column to
  # be classified as numeric (affects type inference across many checks)
  type_inference_threshold: 0.90

  # FAIL if any column's missing-value rate exceeds this (0-1)
  max_missing_rate: 0.05

  # FAIL if non-numeric values in a numeric column exceed this rate
  max_non_numeric_rate: 0.01

  # FAIL if row count drops below this (0 = disabled)
  min_row_count: 0

  # Version-comparison thresholds (used when a previous file is available)
  max_row_count_change_pct:       0.10
  max_numeric_mean_shift_pct:     0.20
  max_missing_rate_change_pp:     2.0    # percentage points
  max_non_numeric_rate_change_pp: 1.0

  flag_new_columns:         true   # set false to suppress new-column warnings in CP-02
  flag_dropped_columns:     true   # set false to suppress dropped-column warnings in CP-02
  flag_type_changes:        true   # set false to suppress type-change warnings in CP-02
  flag_column_order_change: true   # set false to skip CP-08 entirely (WARN CSV / FAIL FWF)
```

### Per-dataset config — `<dataset_name>.yml`

One file per dataset. Only the keys you specify are needed — everything else
inherits from the global defaults.

```yaml
dataset_name: "customer_accounts"   # for human reference only — the functional
                                     # identifier is the YAML filename, which must
                                     # match the dataset_name argument to run_dq_check()

# --- File location ---
# Option A: folder scan (two most recently modified files are used)
folder: "data/incoming/customer_accounts/"

# Option B: explicit paths (set current_file only → single-file mode,
#           no version comparison)
# current_file:  "data/incoming/customer_accounts/20260410.csv"
# previous_file: "data/incoming/customer_accounts/20260401.csv"

format:    csv
encoding:  "UTF-8"
delimiter: ","

# --- Schema contract (optional) ---
# SC-01 flags extra columns; SC-02 flags missing columns.
expected_columns:
  - id
  - name
  - country_code
  - account_status
  - account_balance
  - created_date

# --- Key columns (optional) ---
# QC-12 checks each for duplicate values.
key_columns:
  - id

# --- Rule overrides (optional) ---
# These replace the matching keys in default_rules for this dataset only.
rule_overrides:
  max_missing_rate:        0.02
  max_row_count_change_pct: 0.05
  min_row_count:           1000

# --- Per-column rules (optional) ---
column_rules:
  country_code:
    allowed_values: ["GB", "US", "DE", "FR"]   # QC-09
    pattern: "^[A-Z]{2}$"                       # QC-13
  account_status:
    allowed_values: ["ACTIVE", "CLOSED", "SUSPENDED"]
  account_balance:
    min_value: 0                                 # QC-10
    max_value: 1000000

# --- Custom checks (optional) ---
# custom_checks_file: "custom/customer_accounts_checks.R"

# --- Output paths (optional, override global) ---
# snapshot_db:       "data/snapshots.sqlite"
# report_output_dir: "reports/"
```

### What is required vs optional

The only fields the package requires to run are the file location (`folder`
or `current_file`) and the `format`. Every other config key is optional —
the corresponding checks are simply skipped when the key is absent.

| Config key | Skipped / default when absent |
|---|---|
| `key_columns` | QC-12 (key uniqueness) skipped |
| `expected_columns` | SC-01 and SC-02 (schema contract) skipped |
| `column_rules.allowed_values` | QC-09 (allowed values) skipped |
| `column_rules.min_value` / `max_value` | QC-10 (numeric bounds) skipped |
| `column_rules.pattern` | QC-13 (pattern / regex) skipped |
| `custom_checks_file` | All custom checks skipped |
| `previous_file` / second file in folder | All CP-01–CP-08 (version comparison) skipped |
| `min_row_count` set to 0 | QC-14 (minimum row count) skipped |
| `type_inference_threshold` | Defaults to 0.90 (affects QC-06, QC-07, QC-08, QC-11, CP-02, CP-04, CP-05, CP-06, CP-07) |
| `flag_new_columns` | Defaults to `true`; set to `false` to suppress new-column warnings in CP-02 (still tracked in SQLite) |
| `flag_dropped_columns` | Defaults to `true`; set to `false` to suppress dropped-column warnings in CP-02 (still tracked in SQLite) |
| `flag_type_changes` | Defaults to `true`; set to `false` to suppress type-change warnings in CP-02 |
| `flag_column_order_change` | Defaults to `true`; set to `false` to skip CP-08 entirely |

A minimal dataset config that runs basic checks with no column-level rules:

```yaml
dataset_name: "my_dataset"           # for reference; must match the YAML filename
current_file: "data/incoming/my_dataset.csv"
format: csv
```

### Fixed-width files

For fixed-width files, set `format: fwf` and supply column widths and names.
If the file has a header row, set `fwf_skip: 1` to skip it.

```yaml
format: fwf
fwf_skip: 1
fwf_widths:    [10, 20, 3, 10, 12, 11]
fwf_col_names: [id, name, country_code, account_status, account_balance, created_date]
```

---

## The quality checks

Each check returns one or more `dq_result` objects with four key fields:
`check_id`, `status` (`PASS` / `WARN` / `FAIL` / `INFO`), `observed`, and
`message`. The overall run status is the worst single status across all checks.

### Single-snapshot checks (QC series)

These run on the current file alone.

| ID | Name | Triggers on | Severity |
|----|------|-------------|----------|
| QC-01 | Missing rate | Column missing rate > `max_missing_rate` | FAIL |
| QC-02 | Empty column | Column is 100% empty | FAIL |
| QC-03 | Duplicate rows | Any fully-identical rows | WARN |
| QC-04 | Row count | Always | INFO |
| QC-05 | Column count | Always | INFO |
| QC-06 | Inferred type | Always (date / numeric / character / unknown) | INFO |
| QC-07 | Numeric stats | Numeric columns (min, max, mean, SD) | INFO |
| QC-08 | Distinct count | Character columns | INFO |
| QC-09 | Allowed values | Values outside `allowed_values` list | FAIL |
| QC-10 | Numeric bounds | Values outside `min_value` / `max_value` | FAIL |
| QC-11 | Non-numeric values | Non-numeric values in a numeric column | WARN / FAIL |
| QC-12 | Key uniqueness | Duplicate values in a `key_columns` column | FAIL |
| QC-13 | Pattern / regex | Values not matching `pattern` | FAIL |
| QC-14 | Minimum row count | Row count below `min_row_count` | FAIL |

### Schema contract checks (SC series)

Only run when `expected_columns` is set.

| ID | Name | Triggers on | Severity |
|----|------|-------------|----------|
| SC-01 | Unexpected column | Column in file not in `expected_columns` | FAIL |
| SC-02 | Missing expected column | Column in `expected_columns` absent from file | FAIL |

### Version comparison checks (CP series)

Only run when a previous file is available. Skipped automatically in
single-file mode.

| ID | Name | Triggers on | Severity |
|----|------|-------------|----------|
| CP-01 | Row count change | Change > `max_row_count_change_pct` | WARN |
| CP-02 | Schema diff | Columns added, dropped, or type-changed | WARN |
| CP-03 | Missing rate change | Increase > `max_missing_rate_change_pp` pp | WARN |
| CP-04 | Numeric mean shift | Shift > `max_numeric_mean_shift_pct` | WARN |
| CP-05 | New distinct values | Values in current not seen in previous | INFO |
| CP-06 | Dropped distinct values | Values in previous absent from current | INFO |
| CP-07 | Non-numeric rate change | Increase > `max_non_numeric_rate_change_pp` pp | WARN |
| CP-08 | Column order change | Column order differs from previous | WARN (CSV) / FAIL (FWF) |

---

## Type inference

`infer_col_type()` classifies each column after whitespace trimming. Rules
are applied in priority order:

1. **date** — all non-empty values parse as a known date format
   (`%Y-%m-%d`, `%d/%m/%Y`, `%m/%d/%Y`, `%Y%m%d`, `%d-%m-%Y`)
2. **numeric** — ≥ `type_inference_threshold` of non-empty values coerce to numeric
3. **character** — everything else
4. **unknown** — all values are empty or `NA`

The threshold defaults to 90%, meaning a column with up to 10% bad values is
still classified as numeric, making those bad values visible to QC-11
(non-numeric contamination). It is configurable per dataset via
`type_inference_threshold` in `default_rules` or `rule_overrides`:

```yaml
# raise threshold — stricter, fewer columns classified as numeric
rule_overrides:
  type_inference_threshold: 0.95

# lower threshold — more lenient, columns with more noise still classified numeric
rule_overrides:
  type_inference_threshold: 0.80
```

Changing the threshold affects QC-06, QC-07, QC-08, QC-11, CP-02, CP-04,
CP-05, CP-06, and CP-07. Note that changing it between runs may cause CP-02
to report a spurious type change on columns that sit close to the old and new
thresholds, even if the underlying data did not change.

```{r type-inference}
infer_col_type(c("2024-01-15", "2024-06-30"))  # "date"
infer_col_type(c("100", "200", "N/A"))          # "character" (only 67% numeric)
infer_col_type(c("100", "200", "N/A", rep("50", 17)))  # "numeric" (95% numeric)
infer_col_type(c(NA, "", NA))                   # "unknown"
# with a custom threshold:
infer_col_type(c(rep("1", 17), "a", "b", "c"), threshold = 0.80)  # "numeric" (85%)
```

---

## Running a check

```{r run-check}
result <- run_dq_check(
  dataset_name = "customer_accounts",
  config_dir   = "path/to/configs",   # contains dqcheckr.yml + customer_accounts.yml
  open_report  = TRUE                 # open HTML in browser when interactive
)

# Console output (always printed):
# [dqcheckr] customer_accounts: FAIL - 1 warning(s), 2 failure(s). Report: reports/...html

result$status       # "PASS", "WARN", or "FAIL"
result$report_path  # full path to the HTML file
result$snapshot_id  # integer row ID in snapshots table
```

The HTML report contains:

1. **Header** — dataset name, file name, run time, overall status badge
2. **File summary** — row count, column count, file size, format, encoding
3. **Quality checks table** — every check result, failures first
4. **Custom checks table** — if configured
5. **Version comparison** — what changed since the previous delivery
6. **Historical trend** — row count and check outcomes across the last 10 runs
   (appears once at least two runs have been recorded)
7. **Appendix** — full column-level statistics table

---

## Calling individual checks

You can call any check function directly, without going through the full
pipeline. Every check takes `(df, config)` and returns a list of `dq_result`
objects.

```{r individual-checks}
library(dqcheckr)

# Build a minimal config programmatically
cfg <- list(
  rules = list(
    max_missing_rate     = 0.05,
    max_non_numeric_rate = 0.01,
    min_row_count        = 0
  ),
  column_rules     = list(
    status = list(allowed_values = c("ACTIVE", "CLOSED"))
  ),
  key_columns      = "id",
  expected_columns = NULL
)

df <- read.csv("data/customer_accounts.csv", colClasses = "character")
df[] <- lapply(df, trimws)

# Run a single check
results <- check_missing_rate(df, cfg)

# Inspect
results[[1]]$status    # "PASS" or "FAIL"
results[[1]]$observed  # e.g. "3.2% missing (4 of 125)"

# Run all single-snapshot checks at once
all_results <- run_qc_checks(df, cfg)

# Overall status
overall_status(all_results)  # "FAIL" > "WARN" > "PASS" > "INFO"
```

---

## Custom checks

Create a plain `.R` file that defines `custom_checks(df)`. It receives the
trimmed data frame and must return a list of `dq_result` objects.

```{r custom-file}
# File: custom/customer_accounts_checks.R

custom_checks <- function(df) {
  results <- list()

  # Rule: ACTIVE accounts must not have a zero balance
  active_zero <- df[df$account_status == "ACTIVE" &
                    !is.na(df$account_balance) &
                    df$account_balance == "0", ]
  n <- nrow(active_zero)
  results <- c(results, list(dq_result(
    check_id   = "CUST-01",
    check_name = "No zero-balance active accounts",
    column     = "account_balance",   # enables per-column storage in SQLite
    status     = if (n > 0) "FAIL" else "PASS",
    observed   = sprintf("%d ACTIVE account(s) with balance 0", n),
    message    = if (n > 0)
      sprintf("%d ACTIVE account(s) have a zero balance.", n)
    else
      "No ACTIVE accounts have a zero balance."
  )))

  results
}
```

Point to it in the dataset config:

```yaml
custom_checks_file: "custom/customer_accounts_checks.R"
```

The file is sourced in an isolated environment (not the global session).
`dq_result()` is available inside it. Any error — missing file, undefined
function, runtime failure — stops the run with a clear message.

**Tip**: always set `column =` when your check is scoped to a specific column.
Results with `column = NA` (the default) appear in the HTML report and
contribute to the overall status, but are not written to the `column_snapshots`
SQLite table and therefore do not appear in per-column trend queries.

---

## The snapshot database

Every run writes two rows to a shared SQLite database: one row in `snapshots`
(run-level summary) and one row per column-stat in `column_snapshots`.

```{r snapshot-query}
library(DBI)
library(RSQLite)

con <- dbConnect(SQLite(), "data/snapshots.sqlite")

# Recent runs for one dataset
dbGetQuery(con,
  "SELECT id, file_name, overall_status, check_fail_count, run_timestamp
   FROM snapshots
   WHERE dataset_name = 'customer_accounts'
   ORDER BY id DESC
   LIMIT 10")

# Column-level stats for the most recent run
dbGetQuery(con,
  "SELECT column_name, dq_check, value, threshold
   FROM column_snapshots
   WHERE snapshot_id = (
     SELECT MAX(id) FROM snapshots WHERE dataset_name = 'customer_accounts'
   )
   ORDER BY column_name, dq_check")

dbDisconnect(con)
```

Because all datasets share one database, you can query across them:

```{r cross-dataset}
dbGetQuery(con,
  "SELECT dataset_name, COUNT(*) AS runs,
          SUM(check_fail_count) AS total_failures
   FROM snapshots
   GROUP BY dataset_name")
```

The `output/` directory (and database file) are created automatically on the
first run if they do not exist.

---

## Worked example — Star Wars dataset

The following uses `dplyr::starwars` exported to CSV and fixed-width files.

```{r starwars-setup}
# Config directory contains:
#   dqcheckr.yml         — global thresholds
#   starwars_csv.yml     — CSV dataset config
#   starwars_fwf.yml     — FWF dataset config

# Run checks on both formats
result_csv <- run_dq_check("starwars_csv", config_dir = "config", open_report = TRUE)
result_fwf <- run_dq_check("starwars_fwf", config_dir = "config", open_report = TRUE)
```

The Star Wars config sets `max_missing_rate: 0.60` because `vehicles` (87%)
and `starships` (77%) are naturally sparse — most characters have none. Both
columns FAIL, which is the correct finding: they are genuinely incomplete and
a consumer should be aware before using them.

Expected console output:

```
[dqcheckr] starwars_csv: FAIL - 0 warning(s), 2 failure(s). Report: output/reports/starwars_csv_....html
[dqcheckr] starwars_fwf: FAIL - 0 warning(s), 2 failure(s). Report: output/reports/starwars_fwf_....html
```

---

## Error handling

| Situation | Behaviour |
|-----------|-----------|
| Config file not found | Stop immediately; no output written |
| Data file not found | Stop immediately; no output written |
| Only one file available | Single-file mode; CP checks skipped |
| File cannot be parsed | Stop with filename and parse error |
| SC-01 / SC-02 violations | FAIL recorded; run continues; report written |
| Custom checks file missing | Stop with message |
| `custom_checks()` not defined | Stop with message |
| `custom_checks()` runtime error | Stop with R error detail |
| SQLite write fails | Warning emitted; HTML report still written |

The package never modifies, moves, or deletes source files.

---

## Design principles

**Owned and evolvable.** Every check is plain R — readable, modifiable, and
not hidden behind a third-party API.

**Plain English outputs.** Reports are written for data management staff, not
engineers. No R code, no package names, no jargon.

**Fail loudly.** Configuration errors, missing files, and custom-check
failures all stop the run with a clear message. Nothing is silently skipped.

**Non-destructive.** The package reads files and writes reports and database
records. Source files are never touched.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_col_count}
\alias{check_col_count}
\title{QC-05: Report column count}
\usage{
check_col_count(df, config)
}
\description{
QC-05: Report column count
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_check.R
\name{run_dq_check}
\alias{run_dq_check}
\title{Run a full data quality check pipeline}
\usage{
run_dq_check(dataset_name, config_dir = ".", open_report = TRUE)
}
\arguments{
\item{dataset_name}{Character. Name of the dataset; must match a YAML config
file \code{<dataset_name>.yml} in \code{config_dir}.}

\item{config_dir}{Character. Path to the directory containing
\code{dqcheckr.yml} and the dataset YAML file. Defaults to \code{"."}.}

\item{open_report}{Logical. Whether to open the HTML report in the browser
after rendering (only takes effect in interactive sessions).}
}
\value{
Invisibly, a named list with:
  \describe{
    \item{status}{Overall status string: \code{"PASS"}, \code{"WARN"},
      \code{"FAIL"}, or \code{"INFO"}.}
    \item{report_path}{Absolute path to the rendered HTML report.}
    \item{snapshot_id}{Integer row ID of the snapshot written to SQLite,
      or \code{NULL} if the write failed.}
  }
}
\description{
Orchestrates the complete dqcheckr pipeline: loads configuration, detects
files, runs QC and comparison checks, writes a snapshot to SQLite, and
renders an HTML report.
}
\examples{
\donttest{
tmp <- gsub("\\\\\\\\", "/", tempdir())
dat <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
writeLines(c(
  paste0('snapshot_db: "',       tmp, '/snap.sqlite"'),
  paste0('report_output_dir: "', tmp, '"'),
  'default_rules:',
  '  max_missing_rate: 0.60',
  '  min_row_count: 80'
), file.path(tmp, "dqcheckr.yml"))
writeLines(c(
  'dataset_name: "starwars_csv"',
  paste0('current_file: "', dat, '"'),
  'format: csv',
  'encoding: "UTF-8"',
  'delimiter: ","'
), file.path(tmp, "starwars_csv.yml"))
result <- run_dq_check("starwars_csv", config_dir = tmp, open_report = FALSE)
result$status
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{infer_col_type}
\alias{infer_col_type}
\title{Infer the logical type of a character column}
\usage{
infer_col_type(x, threshold = 0.9)
}
\arguments{
\item{x}{Character vector to classify (as read from a CSV or FWF file).}

\item{threshold}{Numeric. Minimum proportion of non-empty values that must
parse as numeric for the column to be classified as \code{"numeric"}.
Defaults to \code{0.90}. Configurable via \code{type_inference_threshold}
in \code{rule_overrides}.}
}
\value{
A single character string: \code{"date"}, \code{"numeric"},
  \code{"character"}, or \code{"unknown"}.
}
\description{
Classifies a character vector as \code{"date"}, \code{"numeric"},
\code{"character"}, or \code{"unknown"} by applying rules in priority order.
}
\examples{
infer_col_type(c("2024-01-01", "2024-06-15"))   # "date"
infer_col_type(c("1.5", "2.0", "3.1"))          # "numeric"
infer_col_type(c("high", "low", "medium"))       # "character"
infer_col_type(c(NA, "", NA))                    # "unknown"
infer_col_type(c(rep("1", 17), "a", "b", "c"), threshold = 0.80)  # "numeric"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ingest.R
\name{detect_files}
\alias{detect_files}
\title{Detect current and previous dataset files}
\usage{
detect_files(config)
}
\arguments{
\item{config}{Named list. Merged configuration as returned by
\code{\link{load_config}}.}
}
\value{
A named list with elements \code{current} (character path) and
  \code{previous} (character path or \code{NULL}).
}
\description{
Resolves the current and previous file paths from the configuration. If
\code{current_file} is set explicitly, it is used directly. Otherwise the
two most recently modified files in \code{folder} are used.
}
\examples{
cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
cfg <- load_config("starwars_csv", config_dir = cfg_dir)
cfg$current_file <- system.file("demonstrations/data/starwars.csv",
                                 package = "dqcheckr")
files <- detect_files(cfg)
files$current

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{run_qc_checks}
\alias{run_qc_checks}
\title{Run all generic quality checks on a dataset}
\usage{
run_qc_checks(df, config)
}
\arguments{
\item{df}{A data frame with all columns as character vectors (as returned by
\code{\link{read_dataset}}).}

\item{config}{Named list. Merged configuration as returned by
\code{\link{load_config}}.}
}
\value{
A list of \code{\link{dq_result}} objects.
}
\description{
Runs the full QC check suite (QC-01 to QC-14, SC-01, SC-02) against a
single data frame snapshot.
}
\examples{
cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
cfg     <- load_config("starwars_csv", config_dir = cfg_dir)
path    <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
df      <- read_dataset(path, cfg)
results <- run_qc_checks(df, cfg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{run_comparison_checks}
\alias{run_comparison_checks}
\title{Run all version comparison checks between two dataset snapshots}
\usage{
run_comparison_checks(df_current, df_previous, config)
}
\arguments{
\item{df_current}{A data frame. The current delivery.}

\item{df_previous}{A data frame. The previous delivery.}

\item{config}{Named list. Merged configuration as returned by
\code{\link{load_config}}.}
}
\value{
A list of \code{\link{dq_result}} objects. The list carries
  attributes \code{new_cols} and \code{dropped_cols} (character vectors)
  for use by the snapshot writer.
}
\description{
Runs CP-01 to CP-08 comparing a current delivery against the previous one.
}
\examples{
cfg_dir   <- system.file("demonstrations/config", package = "dqcheckr")
cfg       <- load_config("starwars_csv", config_dir = cfg_dir)
curr_path <- system.file("demonstrations/data2/starwars_v2.csv", package = "dqcheckr")
prev_path <- system.file("demonstrations/data2/starwars_v1.csv", package = "dqcheckr")
curr      <- read_dataset(curr_path, cfg)
prev      <- read_dataset(prev_path, cfg)
results   <- run_comparison_checks(curr, prev, cfg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_missing_rate}
\alias{check_missing_rate}
\title{QC-01: Check missing rate per column}
\usage{
check_missing_rate(df, config)
}
\arguments{
\item{df}{A data frame with all columns as character vectors.}

\item{config}{Named list as returned by \code{\link{load_config}}.}
}
\value{
A list of \code{\link{dq_result}} objects, one per column.
}
\description{
Returns a \code{\link{dq_result}} per column flagging columns whose
proportion of missing or empty values exceeds \code{max_missing_rate}.
}
\examples{
cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
cfg  <- load_config("starwars_csv", config_dir = cfg_dir)
path <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
df   <- read_dataset(path, cfg)
check_missing_rate(df, cfg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{overall_status}
\alias{overall_status}
\title{Compute the worst status across a list of dq_result objects}
\usage{
overall_status(results)
}
\arguments{
\item{results}{A list of \code{\link{dq_result}} objects.}
}
\value{
A single character string: \code{"FAIL"}, \code{"WARN"},
  \code{"PASS"}, or \code{"INFO"}.
}
\description{
Returns the single worst status in precedence order:
\code{"FAIL"} > \code{"WARN"} > \code{"PASS"} > \code{"INFO"}.
}
\examples{
r1 <- dq_result("QC-01", "test", status = "PASS", observed = "ok", message = "ok")
r2 <- dq_result("QC-02", "test", status = "WARN", observed = "ok", message = "ok")
overall_status(list(r1, r2))  # "WARN"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_numeric_bounds}
\alias{check_numeric_bounds}
\title{QC-10: Check for out-of-range numeric values}
\usage{
check_numeric_bounds(df, config)
}
\description{
QC-10: Check for out-of-range numeric values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_schema_contract}
\alias{check_schema_contract}
\title{SC-01/SC-02: Check columns against expected schema contract}
\usage{
check_schema_contract(df, config)
}
\description{
SC-01/SC-02: Check columns against expected schema contract
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_row_count}
\alias{check_row_count}
\title{QC-04: Report row count}
\usage{
check_row_count(df, config)
}
\description{
QC-04: Report row count
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare_row_count}
\alias{compare_row_count}
\title{CP-01: Compare row count between deliveries}
\usage{
compare_row_count(df_current, df_previous, config)
}
\description{
CP-01: Compare row count between deliveries
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_numeric_stats}
\alias{check_numeric_stats}
\title{QC-07: Report numeric summary statistics}
\usage{
check_numeric_stats(df, config)
}
\description{
QC-07: Report numeric summary statistics
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{write_snapshot}
\alias{write_snapshot}
\title{Write a run snapshot to the SQLite database}
\usage{
write_snapshot(
  db_path,
  dataset_name,
  file_name,
  df,
  qc_results,
  cp_results,
  custom_results,
  config,
  col_stats = NULL
)
}
\description{
Write a run snapshot to the SQLite database
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{compute_col_stats}
\alias{compute_col_stats}
\title{Compute per-column statistics for snapshot storage}
\usage{
compute_col_stats(df, config, qc_results)
}
\description{
Compute per-column statistics for snapshot storage
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ingest.R
\name{read_dataset}
\alias{read_dataset}
\title{Read a dataset file into a data frame}
\usage{
read_dataset(path, config)
}
\arguments{
\item{path}{Character. Path to the file to read.}

\item{config}{Named list. Merged configuration as returned by
\code{\link{load_config}}. Must include \code{format} (\code{"csv"} or
\code{"fwf"}). For FWF files, \code{fwf_widths} is required and
\code{fwf_col_names} and \code{fwf_skip} are optional.}
}
\value{
A data frame with all columns as character vectors.
}
\description{
Reads a CSV or fixed-width file, coercing all columns to character and
trimming whitespace. Encoding and delimiter are taken from \code{config}.
}
\examples{
cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
cfg  <- load_config("starwars_csv", config_dir = cfg_dir)
path <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
df   <- read_dataset(path, cfg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare_missing_rate}
\alias{compare_missing_rate}
\title{CP-03: Compare per-column missing rate between deliveries}
\usage{
compare_missing_rate(df_current, df_previous, config)
}
\description{
CP-03: Compare per-column missing rate between deliveries
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_duplicate_rows}
\alias{check_duplicate_rows}
\title{QC-03: Check for fully-duplicate rows}
\usage{
check_duplicate_rows(df, config)
}
\description{
QC-03: Check for fully-duplicate rows
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_key_uniqueness}
\alias{check_key_uniqueness}
\title{QC-12: Check uniqueness of key columns}
\usage{
check_key_uniqueness(df, config)
}
\description{
QC-12: Check uniqueness of key columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_empty_column}
\alias{check_empty_column}
\title{QC-02: Check for entirely empty columns}
\usage{
check_empty_column(df, config)
}
\description{
QC-02: Check for entirely empty columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{dq_result}
\alias{dq_result}
\title{Construct a data quality result object}
\usage{
dq_result(
  check_id,
  check_name,
  column = NA_character_,
  status,
  observed,
  threshold = NA_character_,
  message
)
}
\arguments{
\item{check_id}{Character. Short identifier for the check (e.g. \code{"QC-01"}).}

\item{check_name}{Character. Human-readable name of the check.}

\item{column}{Character. Column the check applies to, or \code{NA_character_}
for row-level or file-level checks.}

\item{status}{Character. One of \code{"PASS"}, \code{"WARN"}, \code{"FAIL"},
or \code{"INFO"}.}

\item{observed}{Character. What was observed (e.g. \code{"5.2\% missing"}).}

\item{threshold}{Character. The configured threshold, or \code{NA_character_}
if not applicable.}

\item{message}{Character. Human-readable description of the result.}
}
\value{
A named list with seven elements: \code{check_id}, \code{check_name},
  \code{column}, \code{status}, \code{observed}, \code{threshold},
  \code{message}.
}
\description{
Creates the atomic result unit returned by every check function.
}
\examples{
dq_result("QC-01", "Missing rate", column = "age",
          status = "PASS", observed = "0\% missing",
          message = "No missing values.")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{.missing_vals}
\alias{.missing_vals}
\title{Test for missing or empty values}
\usage{
.missing_vals(x)
}
\description{
Test for missing or empty values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_inferred_types}
\alias{check_inferred_types}
\title{QC-06: Report inferred column types}
\usage{
check_inferred_types(df, config)
}
\description{
QC-06: Report inferred column types
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{.missing_rate_vec}
\alias{.missing_rate_vec}
\title{Compute missing rate for a vector}
\usage{
.missing_rate_vec(x)
}
\description{
Compute missing rate for a vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_pattern}
\alias{check_pattern}
\title{QC-13: Check values against a regex pattern}
\usage{
check_pattern(df, config)
}
\description{
QC-13: Check values against a regex pattern
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{load_config}
\alias{load_config}
\title{Load and merge dataset configuration}
\usage{
load_config(dataset_name, config_dir)
}
\arguments{
\item{dataset_name}{Character. Dataset name; must match
\code{<dataset_name>.yml} in \code{config_dir}.}

\item{config_dir}{Character. Path to the directory containing both YAML
files.}
}
\value{
A named list representing the merged configuration.
}
\description{
Reads the global \code{dqcheckr.yml} and the dataset-specific YAML, merging
\code{rule_overrides} from the dataset config on top of \code{default_rules}
from the global config.
}
\examples{
cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
cfg <- load_config("starwars_csv", config_dir = cfg_dir)
cfg$format

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dqcheckr-package.R
\docType{package}
\name{dqcheckr-package}
\alias{dqcheckr}
\alias{dqcheckr-package}
\title{dqcheckr: Automated Data Quality Checks for Recurring Dataset Deliveries}
\description{
Automates quality verification of recurring external dataset deliveries.
For each new file arrival, it runs single-snapshot quality checks (QC-01 to
QC-14, SC-01/SC-02), compares the file to the previous delivery (CP-01 to
CP-08), writes a self-contained 'HTML' report, and records summary statistics
in a local 'SQLite' database for long-term trend tracking. Supports 'CSV' and
fixed-width formats. Custom organisation-specific checks can be supplied as
plain R files.
}
\details{
The main entry point is \code{\link{run_dq_check}}. Configuration is driven
by two 'YAML' files: a global \code{dqcheckr.yml} and a per-dataset
\code{<dataset_name>.yml}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/mickmioduszewski/dqcheckr}
  \item Report bugs at \url{https://github.com/mickmioduszewski/dqcheckr/issues}
}

}
\author{
\strong{Maintainer}: Mick Mioduszewski \email{mick@mioduszewski.net}

Authors:
\itemize{
  \item Mick Mioduszewski \email{mick@mioduszewski.net}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{read_recent_snapshots}
\alias{read_recent_snapshots}
\title{Read recent snapshot history from the SQLite database}
\usage{
read_recent_snapshots(db_path, dataset_name, n = 10)
}
\arguments{
\item{db_path}{Character. Path to the SQLite database file.}

\item{dataset_name}{Character. Dataset name to filter on.}

\item{n}{Integer. Maximum number of records to return. Defaults to 10.}
}
\value{
A data frame with one row per run and columns including
  \code{id}, \code{run_timestamp}, \code{file_name}, \code{row_count},
  \code{overall_status}, \code{check_pass_count}, \code{check_warn_count},
  \code{check_fail_count}. Returns an empty data frame if the database does
  not exist or contains no records for the dataset.
}
\description{
Retrieves the \code{n} most recent run records for a given dataset from the
snapshot database, ordered newest-first.
}
\examples{
history <- read_recent_snapshots(tempfile(fileext = ".sqlite"), "starwars_csv")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{init_snapshot_db}
\alias{init_snapshot_db}
\title{Initialise the SQLite snapshot database}
\usage{
init_snapshot_db(db_path)
}
\description{
Initialise the SQLite snapshot database
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare_dropped_values}
\alias{compare_dropped_values}
\title{CP-06: Detect dropped distinct values in character columns}
\usage{
compare_dropped_values(df_current, df_previous, config)
}
\description{
CP-06: Detect dropped distinct values in character columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{render_report}
\alias{render_report}
\title{Render the HTML data quality report}
\usage{
render_report(
  dataset_name,
  file_name,
  file_path,
  df,
  qc_results,
  cp_results,
  custom_results,
  snapshot_history,
  config,
  col_stats = NULL,
  output_dir,
  open_report = TRUE
)
}
\description{
Render the HTML data quality report
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_custom.R
\name{run_custom_checks}
\alias{run_custom_checks}
\title{Run organisation-specific custom checks}
\usage{
run_custom_checks(df, config)
}
\arguments{
\item{df}{A data frame. The current delivery.}

\item{config}{Named list. Merged configuration as returned by
\code{\link{load_config}}.}
}
\value{
A list of \code{\link{dq_result}} objects (may be empty).
}
\description{
Sources the R file specified by \code{config$custom_checks_file}, which must
define a function \code{custom_checks(df)} returning a list of
\code{\link{dq_result}} objects. Returns an empty list if
\code{custom_checks_file} is not set in the config.
}
\examples{
cfg_dir <- system.file("demonstrations/config", package = "dqcheckr")
cfg     <- load_config("starwars_csv", config_dir = cfg_dir)
path    <- system.file("demonstrations/data/starwars.csv", package = "dqcheckr")
df      <- read_dataset(path, cfg)
results <- run_custom_checks(df, cfg)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_allowed_values}
\alias{check_allowed_values}
\title{QC-09: Check for values outside the allowed set}
\usage{
check_allowed_values(df, config)
}
\description{
QC-09: Check for values outside the allowed set
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_non_numeric}
\alias{check_non_numeric}
\title{QC-11: Check non-numeric rate in numeric columns}
\usage{
check_non_numeric(df, config)
}
\description{
QC-11: Check non-numeric rate in numeric columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks_generic.R
\name{check_distinct_counts}
\alias{check_distinct_counts}
\title{QC-08: Report distinct value counts for character columns}
\usage{
check_distinct_counts(df, config)
}
\description{
QC-08: Report distinct value counts for character columns
}
\keyword{internal}
